//
// Flame screensaver
// ----------------------------
// (c) 1996 Lucian Wischik
// Original code (c) 1995 Brian Stone
// and (c) 1993 JCAB of Inconexi
//
//
// TImage is a class to represent a DIBSection. Call
// TImage::Reset(newx,newy) to create a new DIBsection. Call TImage::Clear()
// to remove the current bitmap. To change the palette (so you can
// do things like colour-cycling) first write your new palette into
// TImage.logpal.Entries[] and then call ChangePalette(). This is
// what I have done in my procedure TFlames::RenderColours().
// TFlames.Step is the routine in which the raw bits of the bitmap
// are manipulated. It returns 'TRUE' if it wants to be repainted
//

#define STRICT
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <mmsystem.h>
#include <commctrl.h>
#include <stdlib.h>
#pragma hdrstop

#include "../../lib/scrplus.h"
#include "flame.rh"



// Note: I've put the following #defines just so this example compiles
// even if you don't have the latest headers
#if (WINVER < 0x0400)
#define WM_NOTIFY 0x004E
typedef struct tagNMHDR
{ HWND  hwndFrom;
  UINT  idFrom;
  UINT  code;         // NM_ code
} NMHDR;
typedef NMHDR FAR * LPNMHDR;
#define PSN_FIRST (0U-200U)
#define PSN_APPLY (PSN_FIRST-2)
#define PSN_RESET (PSN_FIRST-3)
#endif


BYTE defpal[64*3] =
{
        0,   0,   0,   0,   1,   1,   0,   4,   5,   0,   7,   9,
	     0,   8,  11,   0,   9,  12,  15,   6,   8,  25,   4,   4,
	  	 33,   3,   3,  40,   2,   2,  48,   2,   2,  55,   1,   1,
	  	 63,   0,   0,  63,   0,   0,  63,   3,   0,  63,   7,   0,
	  	 63,  10,   0,  63,  13,   0,  63,  16,   0,  63,  20,   0,
	  	 63,  23,   0,  63,  26,   0,  63,  29,   0,  63,  33,   0,
	  	 63,  36,   0,  63,  39,   0,  63,  39,   0,  63,  40,   0,
	  	 63,  40,   0,  63,  41,   0,  63,  42,   0,  63,  42,   0,
	  	 63,  43,   0,  63,  44,   0,  63,  44,   0,  63,  45,   0,
	  	 63,  45,   0,  63,  46,   0,  63,  47,   0,  63,  47,   0,
	  	 63,  48,   0,  63,  49,   0,  63,  49,   0,  63,  50,   0,
	  	 63,  51,   0,  63,  51,   0,  63,  52,   0,  63,  53,   0,
	  	 63,  53,   0,  63,  54,   0,  63,  55,   0,  63,  55,   0,
	  	 63,  56,   0,  63,  57,   0,  63,  57,   0,  63,  58,   0,
	  	 63,  58,   0,  63,  59,   0,  63,  60,   0,  63,  60,   0,
       63,  61,   0,  63,  62,   0,  63,  62,   0,  63,  63,  63,
};

HINSTANCE hInstance;

typedef struct {
  WORD Version;
  WORD NumberOfEntries;
  PALETTEENTRY Entries[256];
} TMyPalette;

class TImage { public:
  TImage();
  ~TImage();
  void DeleteBitmap();
  void ResetBitmap(int x,int y);
  void ChangePalette();
  //
  BOOL dibpresent;
  BITMAPINFOHEADER bi;
  RGBQUAD cols[256];
  union { LPDWORD lpdw; LPVOID lpv; LPBYTE lpb; } data; // Just for ease of access
  HDC hdc;
  HGDIOBJ holdbitmap;
  unsigned int cx,cy,bpp,numbytes,numlongs;
  HPALETTE hpal;
  TMyPalette logpal;
};

TImage::TImage() {dibpresent=FALSE;}
TImage::~TImage() {DeleteBitmap(); }

void TImage::DeleteBitmap()
{ if (!dibpresent) return;
  HGDIOBJ hbm=SelectObject(hdc,holdbitmap);
  DeleteObject(hbm);
  DeleteDC(hdc);
  DeleteObject(hpal);
  dibpresent=FALSE;
}

void TImage::ResetBitmap(int newx,int newy)
{ DeleteBitmap();
  bi.biSize=sizeof(BITMAPINFOHEADER);
  bi.biPlanes=1;
  bi.biBitCount=8;
  bi.biCompression=BI_RGB;
  bi.biSizeImage=0;
  bi.biClrUsed=0;
  bi.biClrImportant=0;
  bi.biWidth=newx;
  bi.biHeight=newy;
  logpal.Version=0x300;
  logpal.NumberOfEntries=256;
  HDC hscreen=GetDC(HWND_DESKTOP);
  GetSystemPaletteEntries(hscreen,0,256,logpal.Entries);
  ReleaseDC(0,hscreen);
  ChangePalette();
  hpal=CreatePalette((LOGPALETTE *)&logpal);
  hdc=CreateCompatibleDC(NULL);
  HGDIOBJ hbm=CreateDIBSection(hdc,(BITMAPINFO *)&bi,DIB_PAL_COLORS,&data.lpv,NULL,0);
  bi.biSizeImage=bi.biWidth*bi.biHeight;
  holdbitmap=SelectObject(hdc,hbm);
  SetDIBColorTable(hdc,0,256,cols);
  cx=bi.biWidth; cy=bi.biHeight;
  numbytes=bi.biSizeImage;
  numlongs=numbytes/4;
  FillMemory(data.lpv,numbytes,0);
  dibpresent=TRUE;
}

void TImage::ChangePalette()
{ if (!dibpresent) return;
  for (int i=0; i<256; i++)
  { cols[i].rgbRed  =logpal.Entries[i].peRed;
    cols[i].rgbGreen=logpal.Entries[i].peGreen;
    cols[i].rgbBlue =logpal.Entries[i].peBlue;
    cols[i].rgbReserved=0;
    logpal.Entries[i].peFlags=0;
  }
  AnimatePalette(hpal,0,256,logpal.Entries);
  SetDIBColorTable(hdc,0,256,cols);
}

#define SCRCHANGE_FLAME 2
#define SCRCHANGE_RESET 3

class TFlames { public:
  BOOL ignorechanges;  // for dialog
  BYTE GenColFirst, markGenColFirst;
  BYTE GenColLast,  markGenColLast;
  int  MaxJagg,     markMaxJagg;
  int  Fade,        markFade;
  int  Speed,       markSpeed;
  int  Diag,        markDiag;
  int  Jerk,        markJerk;
  int  Cut,         markCut;
  int  ScreenX,     markScreenX;
  int  ScreenY,     markScreenY;
  BOOL Performance, markPerformance;
  int  MinCoarse,   markMinCoarse;
  int  MaxCoarse,   markMaxCoarse;

  void ReadRegistry();
  void WriteRegistry();
  void NotifyChanges(HWND hwnd,BOOL isapply,WPARAM code=SCRCHANGE_FLAME);
  void Mark();
  BOOL Reset();
  void SetControls(HWND);
  BOOL GetControls(HWND);

  void Change(HWND,TFlames *);

  HWND hsrcwnd;
  char perfstring[30];
  int jagger;
  BYTE colfirst, collast;
  int startwidth,startheight;
  int margin;
  unsigned int nframes;
  BOOL isstarted;
  DWORD randnum;
  int orientation;
  unsigned int period;
  unsigned int lasttick;
  unsigned int cx,cy;
  unsigned int scrx,scry,scrleft,scrtop;
  HPALETTE hmainpal;
  TImage *bits;
  unsigned int step;

  TFlames();
  ~TFlames();
  void Start(HWND);
  void Stop();
  BOOL Step(HWND);
  void RenderColours(PALETTEENTRY *);
  void Paint(HWND,PAINTSTRUCT &);
  DWORD fastrand(const int start,const int len);
};


TFlames::TFlames()
{ isstarted=FALSE; hmainpal=NULL; hsrcwnd=NULL;
  bits=new TImage(); 

  GenColFirst=5; GenColLast=110;
  MaxJagg=10;
  Fade=3;
  Speed=3;
  Diag=2;
  Jerk=2;
  ScreenX=320; ScreenY=320;
  Cut=10;
  Performance=FALSE;
  MinCoarse=1; MaxCoarse=20;

  strcpy(perfstring,"Performance");
  colfirst=0; collast=253;
  randnum=GetTickCount();
  orientation=-1; // bottom is zero
  margin=1;
  nframes=0;
  jagger=0;
}

TFlames::~TFlames()
{ if (bits!=NULL) delete bits; bits=NULL;
}

inline DWORD TFlames::fastrand(const int first,const int last)
{ randnum=1664525*randnum+1013904223;
  int len=last-first;
  //unsigned int v=(randnum&&0x00FF0000)>>16;
  //v=(v*len)/256;
  unsigned int v=((double)rand())/((double)RAND_MAX)*((double)len);
  return v+first;
}


void TFlames::RenderColours(PALETTEENTRY *p)
{ //p[254].peRed=255;  p[254].peGreen=255;  p[254].peBlue=80;
  for (int j=0; j<64; j++)
  { p[j].peRed=(BYTE)(defpal[j*3]*4);
	 p[j].peGreen=(BYTE)(defpal[j*3+1]*4);
	 p[j].peBlue=(BYTE)(defpal[j*3+2]*4);
  }
  for (int i=64; i<120; i++)
  { p[i].peRed=(BYTE)(defpal[63*3]*4-(i-64));
    p[i].peGreen=(BYTE)(defpal[63*3+1]*4-(i-64));
    p[i].peBlue=(BYTE)(defpal[63*3+2]*4);
  }
}

void TFlames::Start(HWND hwnd)
{ bits->ResetBitmap(ScreenX,ScreenY);
  RenderColours(bits->logpal.Entries); bits->ChangePalette();
  hmainpal=bits->hpal;
  cx=bits->cx; cy=bits->cy;
  isstarted=TRUE;
  RECT rc; GetWindowRect(hwnd,&rc);
  scrx=rc.right-rc.left; scry=rc.bottom-rc.top;
  scrleft=(scrx-cx)/2; scrtop=scry-cy;
  step=0;
  lasttick=timeGetTime();
}


void TFlames::Stop()
{ if (bits!=NULL) delete bits; bits=NULL;
  isstarted=FALSE;
}



BOOL TFlames::Step(HWND)
{ unsigned int start=timeGetTime();
  step++;
  // First step is to draw in the bottom six rows
  // with random colours.
  // middleness is to give it a pointy look: 'frac' ranges
  // from 0.5 to 1.5, but is capped at 1.0
  int middlex=cx/2;
  int coarsity=fastrand(MinCoarse,MaxCoarse);
  int maxext=cx/20+5;
  int margin=4;
  BYTE *topbaseleft; // the top of the three bottom rows
  if (orientation==-1) topbaseleft=bits->data.lpb+6*cx-1;
  else topbaseleft=bits->data.lpb+(cy-6)*cx;
  // We draw blocks every coursity, and these blocks
  // may be width 1 to maxext.
  for (int i=margin; i<bits->cx-margin; i+=coarsity)
  { double frac;
    if (i<middlex) frac=((double)i)/((double)middlex)+0.5;
    else frac=2.5-((double)i)/((double)middlex);
    if (frac>1.0) frac=1.0;
	 DWORD rn=fastrand(frac*((double)GenColFirst),frac*((double)GenColLast));
    BYTE col=(BYTE)rn;
    DWORD blockwidth=fastrand(1,maxext);
    if (blockwidth>coarsity) blockwidth=coarsity;
    if (i+blockwidth>cx-margin) blockwidth=cx-margin-i;
    for (int row=0; row<6; row++)
    { for (int k=i; k<blockwidth+i; k++)
      { *(topbaseleft+orientation*row*cx+orientation*k)=col;
      }
    }
  }


  // Now we're going to do the averaging trick:
  // . T .
  // . Pa.  new val for T = av. of old vals of P
  // PbPcPd
  // this is done in a sweep starting at the top.
  // At the same time, we're going to shift the
  // next frame up three slices
  // It wraps around. So what.
  // NB. Our 'left','right' assume an orientation of +1
  // (ie. 0 is top left). If orientation is -1, then
  // this will be reversed
  if (orientation==-1)
  { BYTE *pasttopright=bits->data.lpb+bits->numbytes;
    BYTE *botleft=bits->data.lpb;
    jagger++; if (jagger>=MaxJagg) jagger=0;
    BYTE *linewrite=pasttopright;
    for (int y=cy-1; y>=3; y--)
	 { BYTE *write=linewrite-cx+(jagger+Diag*y)%MaxJagg;
      BYTE *reada=write-cx;
      BYTE *readb=reada-cx;
      BYTE *readc=readb-1;
      BYTE *readd=readb+1;
      BYTE *endline=linewrite;
      while (write<endline)
      { BYTE avcol=(BYTE)(((*reada+*readb+*readc+*readd)>>2));
        if (avcol<Fade) *write=0; else *write=(BYTE)(avcol-Fade);
        write+=MaxJagg;
		  reada+=MaxJagg; readb+=MaxJagg; readc+=MaxJagg; readd+=MaxJagg;
      }
      linewrite-=cx;
    }
    BYTE *shiftwrite=pasttopright-4;
    BYTE *shiftread=shiftwrite-Speed*cx;
    BYTE *shiftlast=botleft+Speed*cx;
    while (shiftwrite>shiftlast)
    { *((LPDWORD)shiftwrite)=*((LPDWORD)shiftread);
      shiftwrite-=4; shiftread-=4;
	 }

  }
  unsigned int end=timeGetTime();
  period=end-start;

  nframes++;
  if ((nframes%Jerk)==0) return TRUE; else return FALSE;
  // Only redraw once every couple of frames
}

void TFlames::Change(HWND hwnd,TFlames *f)
{ GenColFirst=f->GenColFirst;
  GenColLast =f->GenColLast;
  MaxJagg    =f->MaxJagg;
  Fade       =f->Fade;
  Speed      =f->Speed;
  Diag       =f->Diag;
  Jerk       =f->Jerk;
  if (Cut!=f->Cut)
  { Cut      =f->Cut;
    InvalidateRect(hwnd,NULL,TRUE);
  }
  if (ScreenX!=f->ScreenX || ScreenY!=f->ScreenY)
  { ScreenX  =f->ScreenX;
    ScreenY  =f->ScreenY;
    Start(hwnd);
    InvalidateRect(hwnd,NULL,TRUE);
  }
  Performance=f->Performance;
  MinCoarse=f->MinCoarse;
  MaxCoarse=f->MaxCoarse;

}


void TFlames::Paint(HWND,PAINTSTRUCT &ps)
{
  unsigned int start=timeGetTime();
  if (Performance)
  { COLORREF oldcol=SetTextColor(bits->hdc,0x00808080);
    COLORREF oldbk=SetBkColor(bits->hdc,0);
    TextOut(bits->hdc,0,0,perfstring,strlen(perfstring));
	 SetTextColor(bits->hdc,oldcol);
    SetBkColor(bits->hdc,oldbk);
  }
  BitBlt(ps.hdc,scrleft,scrtop+Cut,cx,cy-Cut,bits->hdc,0,0,SRCCOPY);
  unsigned int end=timeGetTime();
  unsigned int len=end-start;
  unsigned int thistick=timeGetTime();
  unsigned int tickperiod=thistick-lasttick;
  lasttick=thistick;
  wsprintf(perfstring,"%lu+%lu=%lu",period,len,tickperiod);
}



LRESULT CALLBACK FlameWindowProc(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam)
{ TFlames *f=(TFlames *)GetWindowLong(hwnd,GWL_USERDATA);
  if (f==NULL && msg!=WM_CREATE) return DefWindowProc(hwnd,msg,wParam,lParam);
  switch (msg)
  { case WM_CREATE:
    { TFlames *f=new TFlames();
		f->ReadRegistry();
      SetWindowLong(hwnd,GWL_USERDATA,(LONG)f);
      f->Start(hwnd);
    } break;
    case WM_ERASEBKGND:
    { HDC hdc=(HDC) wParam; RECT rc; GetClientRect(hwnd,&rc);
      FillRect(hdc,&rc,(HBRUSH)GetStockObject(BLACK_BRUSH));
    } break;
    case WM_PAINT:
    { PAINTSTRUCT ps;
		BeginPaint(hwnd,&ps);
      HDC hdc=ps.hdc;
      if (f->hmainpal!=NULL)
      { SelectPalette(hdc,f->hmainpal,FALSE);
        RealizePalette(hdc);
      }
      f->Paint(hwnd,ps);
      EndPaint(hwnd,&ps);
    } break;
    case WM_PALETTECHANGED:
	 { if (f->hmainpal!=NULL && (HWND)wParam==hwnd)
      { SendMessage(hwnd,WM_QUERYNEWPALETTE,0,0);
      }
    } break;
    case WM_QUERYNEWPALETTE:
    { if (f->hmainpal!=NULL)
      { HDC hdc=GetDC(hwnd);
        SelectPalette(hdc,f->hmainpal,FALSE);
        int ncolsmapped=RealizePalette(hdc);
        ReleaseDC(hwnd,hdc);
		  if (ncolsmapped>0)
        { InvalidateRect(hwnd,NULL,TRUE); return TRUE;
        }
        else
        { return FALSE;
        }
      }
    }
    case SCRM_TIMER:
    { if (f->isstarted)
      { if (f->Step(hwnd))
        { InvalidateRect(hwnd,NULL,FALSE);
          UpdateWindow(hwnd);
        }
      }
      return 1;
	 }
    case SCRM_CHANGE:
    { if ((wParam&SCRCHANGE_GETCODE)==SCRCHANGE_FLAME)
      { f->Change(hwnd,(TFlames *)lParam);
      }
      if ((wParam&SCRCHANGE_GETCODE)==SCRCHANGE_RESET)
      { f->Start(hwnd); InvalidateRect(hwnd,NULL,TRUE);
      }
    } break;
    case WM_DESTROY:
	 { if (f->isstarted) f->Stop();
      SetWindowLong(hwnd,GWL_USERDATA,0);
      delete f;
    } break;
  }
  return DefScreenSaverProc(hwnd,msg,wParam,lParam);
}



BOOL WINAPI FlameDialogProc(HWND hwnd, UINT msg,WPARAM wParam,LPARAM lParam)
{ TFlames *f=(TFlames *)GetWindowLong(hwnd,DWL_USER);
  if (f==NULL && msg!=WM_INITDIALOG) return FALSE;
  switch (msg)
  { case WM_INITDIALOG:
    { TFlames *f=new TFlames();
      f->hsrcwnd=hwnd;
      f->ReadRegistry();
      f->Mark();
      SetWindowLong(hwnd,DWL_USER,(LONG)f);
		f->SetControls(hwnd);
      return TRUE;
    }
    case WM_DESTROY:
    { SetWindowLong(hwnd,DWL_USER,0);
      delete f;
      return TRUE;
    }
    case WM_COMMAND:
    { HWND hctrlwnd=(HWND) lParam;
		if (hctrlwnd==NULL) return FALSE;
      int id=LOWORD(wParam);
      if (id==IDOK)
      { f->Mark();
        f->WriteRegistry();
        f->NotifyChanges(f->hsrcwnd,TRUE);
        f->ignorechanges=TRUE;
        EndDialog(hwnd,IDOK);
        return TRUE;
      }
      if (id==IDCANCEL)
      { BOOL anychanges=f->Reset();
		  if (anychanges) f->NotifyChanges(f->hsrcwnd,TRUE);
        f->ignorechanges=TRUE;
        EndDialog(hwnd,IDCANCEL);
        return TRUE;
      }
      if (id==ID_RESET) { f->NotifyChanges(f->hsrcwnd,FALSE,SCRCHANGE_RESET); return TRUE; }
		BOOL ischanged=f->GetControls(hwnd);
      if (ischanged) f->NotifyChanges(f->hsrcwnd,FALSE);
		return TRUE;
	 }
	 case (WM_NOTIFY):
    { LPNMHDR nmh=(LPNMHDR) lParam;
		UINT code=nmh->code;
		switch (code) {
        case (PSN_APPLY):
		  { f->Mark();
          f->WriteRegistry();
			 f->NotifyChanges(f->hsrcwnd,TRUE);
			 f->ignorechanges=TRUE;
			 return TRUE;
		  }
		  case (PSN_RESET):
        { // Revert back to old values
          BOOL anychanges=f->Reset();
          if (anychanges) f->NotifyChanges(f->hsrcwnd,TRUE);
          f->ignorechanges=TRUE;
          return TRUE;
        }
        default: return FALSE;
      }
    }

  }
  return FALSE;
}

void SetDlgItemRng(HWND hwnd,UINT id,int min,int max,int)
{ SendDlgItemMessage(hwnd,id,UDM_SETRANGE,0,MAKELONG(max,min));
}

void TFlames::SetControls(HWND hwnd)
{ ignorechanges=TRUE;
  SetDlgItemInt(hwnd,ID_GENCOLFIRST,GenColFirst,TRUE);
  SetDlgItemRng(hwnd,UD_GENCOLFIRST,0,254,2);
  SetDlgItemInt(hwnd,ID_GENCOLLAST, GenColLast, TRUE);
  SetDlgItemRng(hwnd,UD_GENCOLLAST, 0,254,2);
  SetDlgItemInt(hwnd,ID_MAXJAGG,    MaxJagg,    TRUE);
  SetDlgItemRng(hwnd,UD_MAXJAGG,    0,50,1);
  SetDlgItemInt(hwnd,ID_FADE,       Fade,       TRUE);
  SetDlgItemRng(hwnd,UD_FADE,       0,50,1);
  SetDlgItemInt(hwnd,ID_SPEED,      Speed,      TRUE);
  SetDlgItemRng(hwnd,UD_SPEED,      0,50,1);
  SetDlgItemInt(hwnd,ID_DIAG,       Diag,       TRUE);
  SetDlgItemRng(hwnd,UD_DIAG,       0,50,1);
  SetDlgItemInt(hwnd,ID_JERK,       Jerk,       TRUE);
  SetDlgItemRng(hwnd,UD_JERK,       0,50,1);
  SetDlgItemInt(hwnd,ID_CUT,        Cut,        TRUE);
  SetDlgItemRng(hwnd,UD_CUT,        0,50,1);
  SetDlgItemInt(hwnd,ID_SCREENX,    ScreenX,    TRUE);
  SetDlgItemRng(hwnd,UD_SCREENX,    0,800,16);
  SetDlgItemInt(hwnd,ID_SCREENY,    ScreenY,    TRUE);
  SetDlgItemRng(hwnd,UD_SCREENY,    0,800,16);
  CheckDlgButton(hwnd,ID_PERFORMANCE,Performance);
  SetDlgItemInt(hwnd,ID_MINCOARSE,  MinCoarse,  TRUE);
  SetDlgItemRng(hwnd,UD_MINCOARSE,  1,300,5);
  SetDlgItemInt(hwnd,ID_MAXCOARSE,  MaxCoarse,  TRUE);
  SetDlgItemRng(hwnd,UD_MAXCOARSE,  1,300,5);
  ScrCreatePreview(GetDlgItem(hwnd,ID_MONITOR));
  ignorechanges=FALSE;
}

BOOL TFlames::GetControls(HWND hwnd)
{ if (ignorechanges) return FALSE;
  BOOL ischanged=FALSE;
  BOOL succ;
  BYTE newGenColFirst=(BYTE)GetDlgItemInt(hwnd,ID_GENCOLFIRST,&succ,TRUE);
  if (succ)
  { if (newGenColFirst!=GenColFirst) ischanged=TRUE; GenColFirst=newGenColFirst;
  }
  BYTE newGenColLast =(BYTE)GetDlgItemInt(hwnd,ID_GENCOLLAST, &succ,TRUE);
  if (succ)
  { if (newGenColLast !=GenColLast ) ischanged=TRUE; GenColLast =newGenColLast;
  }
  int  newMaxJagg    =GetDlgItemInt(hwnd,ID_MAXJAGG,    &succ,TRUE);
  if (succ && newMaxJagg>=1)
  { if (newMaxJagg    !=MaxJagg    ) ischanged=TRUE; MaxJagg    =newMaxJagg;
  }
  int  newFade       =GetDlgItemInt(hwnd,ID_FADE,       &succ,TRUE);
  if (succ)
  { if (newFade       !=Fade       ) ischanged=TRUE; Fade       =newFade;
  }
  int  newSpeed      =GetDlgItemInt(hwnd,ID_SPEED,      &succ,TRUE);
  if (succ)
  { if (newSpeed      !=Speed      ) ischanged=TRUE; Speed      =newSpeed;
  }
  int  newDiag       =GetDlgItemInt(hwnd,ID_DIAG,       &succ,TRUE);
  if (succ)
  { if (newDiag       !=Diag       ) ischanged=TRUE; Diag       =newDiag;
  }
  int  newJerk       =GetDlgItemInt(hwnd,ID_JERK,       &succ,TRUE);
  if (succ && newJerk>=1)
  { if (newJerk       !=Jerk       ) ischanged=TRUE; Jerk       =newJerk;
  }
  int  newCut        =GetDlgItemInt(hwnd,ID_CUT,        &succ,TRUE);
  if (succ)
  { if (newCut        !=Cut        ) ischanged=TRUE; Cut        =newCut;
  }
  int  newScreenX    =GetDlgItemInt(hwnd,ID_SCREENX,    &succ,TRUE);
  if (succ && newScreenX>=6)
  { if (newScreenX    !=ScreenX    ) ischanged=TRUE; ScreenX    =newScreenX;
  }
  int  newScreenY    =GetDlgItemInt(hwnd,ID_SCREENY,    &succ,TRUE);
  if (succ && newScreenY>=6)
  { if (newScreenY    !=ScreenY    ) ischanged=TRUE; ScreenY    =newScreenY;
  }
  BOOL newPerformance=IsDlgButtonChecked(hwnd,ID_PERFORMANCE);
  if (newPerformance!=Performance) ischanged=TRUE; Performance=newPerformance;
  int  newMinCoarse  =GetDlgItemInt(hwnd,ID_MINCOARSE,  &succ,TRUE);
  if (succ)
  { if (newMinCoarse<1) newMinCoarse=1;
    if (newMinCoarse !=MinCoarse  ) ischanged=TRUE; MinCoarse  =newMinCoarse;
  }
  int  newMaxCoarse  =GetDlgItemInt(hwnd,ID_MAXCOARSE,  &succ,TRUE);
  if (succ && newMaxCoarse>=MinCoarse)
  { if (newMaxCoarse  !=MaxCoarse  ) ischanged=TRUE; MaxCoarse  =newMaxCoarse;
  }

  return ischanged;
}

BOOL TFlames::Reset()
{ BOOL ischanged=FALSE;
  if (markGenColFirst!=GenColFirst) ischanged=TRUE; GenColFirst=markGenColFirst;
  if (markGenColLast !=GenColLast ) ischanged=TRUE; GenColLast =markGenColLast;
  if (markMaxJagg    !=MaxJagg    ) ischanged=TRUE; MaxJagg    =markMaxJagg;
  if (markFade       !=Fade       ) ischanged=TRUE; Fade       =markFade;
  if (markSpeed      !=Speed      ) ischanged=TRUE; Speed      =markSpeed;
  if (markDiag       !=Diag       ) ischanged=TRUE; Diag       =markDiag;
  if (markJerk       !=Jerk       ) ischanged=TRUE; Jerk       =markJerk;
  if (markCut        !=Cut        ) ischanged=TRUE; Cut        =markCut;
  if (markScreenX    !=ScreenX    ) ischanged=TRUE; ScreenX    =markScreenX;
  if (markScreenY    !=ScreenY    ) ischanged=TRUE; ScreenY    =markScreenY;
  if (markPerformance!=Performance) ischanged=TRUE; Performance=markPerformance;
  if (markMinCoarse  !=MinCoarse  ) ischanged=TRUE; MinCoarse  =markMinCoarse;
  if (markMaxCoarse  !=MaxCoarse  ) ischanged=TRUE; MaxCoarse  =markMaxCoarse;
  return ischanged;
}

void TFlames::Mark()
{ markGenColFirst=GenColFirst;
  markGenColLast =GenColLast;
  markMaxJagg    =MaxJagg;
  markFade       =Fade;
  markSpeed      =Speed;
  markDiag       =Diag;
  markJerk       =Jerk;
  markCut        =Cut;
  markScreenX    =ScreenX;
  markScreenY    =ScreenY;
  markPerformance=Performance;
  markMinCoarse  =MinCoarse;
  markMaxCoarse  =MaxCoarse;
}

void TFlames::NotifyChanges(HWND hsrcwnd, BOOL isapply, WPARAM defcode)
{ WPARAM code;
  if (isapply) code=defcode | SCRCHANGE_FROMAPPLY   | SCRCHANGE_TOBOTH;
  else         code=defcode | SCRCHANGE_FROMCONTROL | SCRCHANGE_TOBOTH;
  TScrChangeRouter cr;
  cr.hsrcwnd=hsrcwnd; cr.lParam=(LPARAM)this; cr.code=code;
  HWND hdlg=GetParent(hsrcwnd);
  SendMessage(hdlg,SCRM_CHANGE,0,(LPARAM)&cr);
}

void TFlames::WriteRegistry()
{ LONG res; HKEY skey; DWORD val; DWORD disp;
  res=RegCreateKeyEx(HKEY_CURRENT_USER,"Software\\Lu\\Flame",0,NULL,
                     REG_OPTION_NON_VOLATILE,KEY_ALL_ACCESS,NULL,&skey,&disp);
  if (res!=ERROR_SUCCESS) return;
  // 1. GenColFirst
  val=GenColFirst;
  RegSetValueEx(skey,"GenColFirst",0,REG_DWORD,(CONST BYTE*)&val,sizeof(val));
  // 2. GenColLast
  val=GenColLast;
  RegSetValueEx(skey,"GenColLast" ,0,REG_DWORD,(CONST BYTE*)&val,sizeof(val));
  // 3. MaxJagg
  val=MaxJagg;
  RegSetValueEx(skey,"MaxJagg"    ,0,REG_DWORD,(CONST BYTE*)&val,sizeof(val));
  // 4. Fade
  val=Fade;
  RegSetValueEx(skey,"Fade"       ,0,REG_DWORD,(CONST BYTE*)&val,sizeof(val));
  // 5. Speed
  val=Speed;
  RegSetValueEx(skey,"Speed"      ,0,REG_DWORD,(CONST BYTE*)&val,sizeof(val));
  // 6. Diag
  val=Diag;
  RegSetValueEx(skey,"Diag"       ,0,REG_DWORD,(CONST BYTE*)&val,sizeof(val));
  // 7. Jerk
  val=Jerk;
  RegSetValueEx(skey,"Jerk"       ,0,REG_DWORD,(CONST BYTE*)&val,sizeof(val));
  // 8. Cut
  val=Cut;
  RegSetValueEx(skey,"Cut"        ,0,REG_DWORD,(CONST BYTE*)&val,sizeof(val));
  // 9. ScreenX
  val=ScreenX;
  RegSetValueEx(skey,"ScreenX"    ,0,REG_DWORD,(CONST BYTE*)&val,sizeof(val));
  // 10. ScreenY
  val=ScreenY;
  RegSetValueEx(skey,"ScreenY"    ,0,REG_DWORD,(CONST BYTE*)&val,sizeof(val));
  // 11. Performance
  val=Performance;
  RegSetValueEx(skey,"Performance",0,REG_DWORD,(CONST BYTE*)&val,sizeof(val));
  // 12. MinCoarse
  val=MinCoarse;
  RegSetValueEx(skey,"MinCoarse"  ,0,REG_DWORD,(CONST BYTE*)&val,sizeof(val));
  // 13. MaxCoarse
  val=MaxCoarse;
  RegSetValueEx(skey,"MaxCoarse"  ,0,REG_DWORD,(CONST BYTE*)&val,sizeof(val));
  RegCloseKey(skey);
}

void TFlames::ReadRegistry()
{ LONG res; HKEY skey; DWORD val,valsize,valtype;
  res=RegOpenKeyEx(HKEY_CURRENT_USER,"Software\\Lu\\Flame",0,KEY_ALL_ACCESS,&skey);
  if (res!=ERROR_SUCCESS) return;
  // 1. GenColFirst
  valsize=sizeof(val); res=RegQueryValueEx(skey,"GenColFirst",0,&valtype,(LPBYTE)&val,&valsize);
  if (res==ERROR_SUCCESS) GenColFirst=(BYTE)val;
  // 2. GenColLast
  valsize=sizeof(val); res=RegQueryValueEx(skey,"GenColLast" ,0,&valtype,(LPBYTE)&val,&valsize);
  if (res==ERROR_SUCCESS) GenColLast=(BYTE)val;
  // 3. MaxJagg
  valsize=sizeof(val); res=RegQueryValueEx(skey,"MaxJagg"    ,0,&valtype,(LPBYTE)&val,&valsize);
  if (res==ERROR_SUCCESS && val>0) MaxJagg=val;
  // 4. Fade
  valsize=sizeof(val); res=RegQueryValueEx(skey,"Fade"       ,0,&valtype,(LPBYTE)&val,&valsize);
  if (res==ERROR_SUCCESS) Fade=val;
  // 5. Speed
  valsize=sizeof(val); res=RegQueryValueEx(skey,"Speed"      ,0,&valtype,(LPBYTE)&val,&valsize);
  if (res==ERROR_SUCCESS) Speed=val;
  // 6. Diag
  valsize=sizeof(val); res=RegQueryValueEx(skey,"Diag"       ,0,&valtype,(LPBYTE)&val,&valsize);
  if (res==ERROR_SUCCESS) Diag=val;
  // 7. Jerk
  valsize=sizeof(val); res=RegQueryValueEx(skey,"Jerk"       ,0,&valtype,(LPBYTE)&val,&valsize);
  if (res==ERROR_SUCCESS && val>0) Jerk=val;
  // 8. Cut
  valsize=sizeof(val); res=RegQueryValueEx(skey,"Cut"        ,0,&valtype,(LPBYTE)&val,&valsize);
  if (res==ERROR_SUCCESS) Cut=val;
  // 9. ScreenX
  valsize=sizeof(val); res=RegQueryValueEx(skey,"ScreenX"    ,0,&valtype,(LPBYTE)&val,&valsize);
  if (res==ERROR_SUCCESS && val>=6) ScreenX=val;
  // 10. ScreenY
  valsize=sizeof(val); res=RegQueryValueEx(skey,"ScreenY"    ,0,&valtype,(LPBYTE)&val,&valsize);
  if (res==ERROR_SUCCESS && val>=6) ScreenY=val;
  // 11. Performance
  valsize=sizeof(val); res=RegQueryValueEx(skey,"Performance",0,&valtype,(LPBYTE)&val,&valsize);
  if (res==ERROR_SUCCESS) Performance=val;
  // 12. MinCoarse
  valsize=sizeof(val); res=RegQueryValueEx(skey,"MinCoarse"  ,0,&valtype,(LPBYTE)&val,&valsize);
  if (res==ERROR_SUCCESS) MinCoarse=val;
  // 13. MaxCoarse
  valsize=sizeof(val); res=RegQueryValueEx(skey,"MaxCoarse"  ,0,&valtype,(LPBYTE)&val,&valsize);
  if (res==ERROR_SUCCESS && val>=MinCoarse) MaxCoarse=val;
  RegCloseKey(skey);
}

#define FlameClassName "FlameClass"


BOOL CALLBACK TidyConfigProc(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam)
{ if (msg==WM_INITDIALOG)
  { ShowWindow(GetDlgItem(hwnd,IDCANCEL),SW_HIDE);
    ShowWindow(GetDlgItem(hwnd,IDOK),SW_HIDE);
    EnableWindow(GetDlgItem(hwnd,IDCANCEL),FALSE);
    EnableWindow(GetDlgItem(hwnd,IDOK),FALSE);
    BOOL ret=FlameDialogProc(hwnd,msg,wParam,lParam);
    TFlames *f=(TFlames *)GetWindowLong(hwnd,DWL_USER);
    if (f==NULL && msg!=WM_INITDIALOG) return ret;
    f->hsrcwnd=GetParent(f->hsrcwnd);
    return ret;
  }
  return FlameDialogProc(hwnd,msg,wParam,lParam);
}



HWND CALLBACK LuCreateSaver(BOOL,CREATESTRUCT &cs)
{ return CreateWindowEx(cs.dwExStyle,FlameClassName,"SaverWindow",cs.style,
								cs.x,cs.y,cs.cx,cs.cy,cs.hwndParent,NULL,
								hInstance,cs.lpCreateParams);
}

#if (WINVER < 0x0400)
int  CALLBACK LuConfig(HWND hwnd,CONFIGSHEETHEADER &,BOOL)
{ return DialogBox(hInstance,MAKEINTRESOURCE(DLG_ACTIVECONFIG),hwnd,FlameDialogProc);
}
#else
int  CALLBACK LuConfig(HWND hwnd,CONFIGSHEETHEADER &psh,BOOL isactive)
{
  if (isactive) return DialogBox(hInstance,MAKEINTRESOURCE(DLG_ACTIVECONFIG),hwnd,FlameDialogProc);
  // otherwise...
  LPPROPSHEETPAGE psp;
  psp=(LPPROPSHEETPAGE) &(psh.ppsp[1]);
  psp->dwSize=sizeof(*psp);
  psp->dwFlags=PSP_DEFAULT;
  psp->hInstance=hInstance; // This is where to load dialog box, icon or title string resource
  psp->pszTemplate=MAKEINTRESOURCE(DLG_CONFIG);
  psp->hIcon=NULL;  // ignored
  psp->pszTitle=NULL; // ignored
  psp->pfnDlgProc=TidyConfigProc;
  psp->lParam=NULL;
  // Set up the overall property sheet
  psh.nPages=2;
  psh.nStartPage=1;
  return PropertySheet(&psh);
}
#endif



int WINAPI WinMain(HINSTANCE h,HINSTANCE,LPSTR,int)
{
  hInstance=h;
  WNDCLASS wcs;
  wcs.style=CS_VREDRAW | CS_HREDRAW;
  wcs.lpfnWndProc=FlameWindowProc;
  wcs.cbClsExtra=0;
  wcs.cbWndExtra=0;
  wcs.hInstance=hInstance; // instance of the procedure
  wcs.hIcon=NULL;
  wcs.hCursor=NULL;
  wcs.hbrBackground=(HBRUSH)GetStockObject(NULL_BRUSH);
  wcs.lpszMenuName=NULL;
  wcs.lpszClassName=FlameClassName;
  RegisterClass(&wcs);
                      
  TScrPrefs *p=ScrCreatePrefs();
  p->AllowActiveConfig=TRUE;
  p->AllowMute=FALSE;
  p->IsTrace=FALSE;
  p->IsDebug=FALSE;
  ScrSetPrefs(p);        
  int res=ScrExecuteEx(hInstance,LuCreateSaver,LuConfig);
  return res;
}
