//
// GLSave screensaver
// -------------------
// (c) 2000 Lucian Wischik
//
// Note: you must link this against opengl32.lib
//


#define WIN32_LEAN_AND_MEAN
#define STRICT
#include <windows.h>
#include <stdlib.h>
#include <gl/gl.h>
#include <math.h>
#include <float.h>
#pragma hdrstop

#include "../../lib/scrplus.h"

HINSTANCE hInstance;


#ifndef PDF_GENERIC_ACCELERATED
#define PFD_GENERIC_ACCELERATED     0x00001000
#endif


// ChoosePixelFormatEx chooses a pixel format.
// It is more aggressive than the normal ChoosePixelFormat command:
// that is to say, it is more keen to get hadrware acceleration.
//
int ChoosePixelFormatEx(HDC hdc,int *p_bpp,int *p_depth,int *p_dbl,int *p_acc)
{ int wbpp; if (p_bpp==NULL) wbpp=-1; else wbpp=*p_bpp;
  int wdepth; if (p_depth==NULL) wdepth=16; else wdepth=*p_depth;
  int wdbl; if (p_dbl==NULL) wdbl=-1; else wdbl=*p_dbl;
  int wacc; if (p_acc==NULL) wacc=1; else wacc=*p_acc;
  PIXELFORMATDESCRIPTOR pfd; ZeroMemory(&pfd,sizeof(pfd)); pfd.nSize=sizeof(pfd); pfd.nVersion=1;
  int num=DescribePixelFormat(hdc,1,sizeof(pfd),&pfd);
  if (num==0) return 0;
  unsigned int maxqual=0; int maxindex=0;
  int max_bpp, max_depth, max_dbl, max_acc;
  for (int i=1; i<=num; i++)
  { ZeroMemory(&pfd,sizeof(pfd)); pfd.nSize=sizeof(pfd); pfd.nVersion=1;
    DescribePixelFormat(hdc,i,sizeof(pfd),&pfd);
    int bpp=pfd.cColorBits;
    int depth=pfd.cDepthBits;
    BOOL pal=(pfd.iPixelType==PFD_TYPE_COLORINDEX);
    BOOL mcd=((pfd.dwFlags & PFD_GENERIC_FORMAT) && (pfd.dwFlags & PFD_GENERIC_ACCELERATED));
    BOOL soft=((pfd.dwFlags & PFD_GENERIC_FORMAT) && !(pfd.dwFlags & PFD_GENERIC_ACCELERATED));
    BOOL icd=(!(pfd.dwFlags & PFD_GENERIC_FORMAT) && !(pfd.dwFlags & PFD_GENERIC_ACCELERATED));
    BOOL opengl=(pfd.dwFlags & PFD_SUPPORT_OPENGL);
    BOOL window=(pfd.dwFlags & PFD_DRAW_TO_WINDOW);
    BOOL bitmap=(pfd.dwFlags & PFD_DRAW_TO_BITMAP);
    BOOL dbuff=(pfd.dwFlags & PFD_DOUBLEBUFFER);
    //
    unsigned int q=0;
    if (opengl && window) q=q+0x8000;
    if (wdepth==-1 || (wdepth>0 && depth>0)) q=q+0x4000;
    if (wdbl==-1 || (wdbl==0 && !dbuff) || (wdbl==1 && dbuff)) q=q+0x2000;
    if (wacc==-1 || (wacc==0 && soft) || (wacc==1 && (mcd || icd))) q=q+0x1000;
    if (mcd || icd) q=q+0x0040; if (icd) q=q+0x0002;
    if (wbpp==-1 || (wbpp==bpp)) q=q+0x0800;
    if (bpp>=16) q=q+0x0020; if (bpp==16) q=q+0x0008;
    if (wdepth==-1 || (wdepth==depth)) q=q+0x0400;
    if (depth>=16) q=q+0x0010; if (depth==16) q=q+0x0004;
    if (!pal) q=q+0x0080;
    if (bitmap) q=q+0x0001;
    if (q>maxqual) {maxqual=q; maxindex=i;max_bpp=bpp; max_depth=depth; max_dbl=dbuff?1:0; max_acc=soft?0:1;}
  }
  if (maxindex==0) return maxindex;
  if (p_bpp!=NULL) *p_bpp=max_bpp;
  if (p_depth!=NULL) *p_depth=max_depth;
  if (p_dbl!=NULL) *p_dbl=max_dbl;
  if (p_acc!=NULL) *p_acc=max_acc;
  return maxindex;
}



void APIENTRY xglLookAt( GLdouble eyex, GLdouble eyey, GLdouble eyez,
                         GLdouble centerx, GLdouble centery, GLdouble centerz,
                         GLdouble upx, GLdouble upy, GLdouble upz )
{
   GLdouble m[16];
   GLdouble x[3], y[3], z[3];
   GLdouble mag;

   /* Make rotation matrix */

   /* Z vector */
   z[0] = eyex - centerx;
   z[1] = eyey - centery;
   z[2] = eyez - centerz;
   mag = sqrt( z[0]*z[0] + z[1]*z[1] + z[2]*z[2] );
   if (mag) {  /* mpichler, 19950515 */
      z[0] /= mag;
      z[1] /= mag;
      z[2] /= mag;
   }

   /* Y vector */
   y[0] = upx;
   y[1] = upy;
   y[2] = upz;

   /* X vector = Y cross Z */
   x[0] =  y[1]*z[2] - y[2]*z[1];
   x[1] = -y[0]*z[2] + y[2]*z[0];
   x[2] =  y[0]*z[1] - y[1]*z[0];

   /* Recompute Y = Z cross X */
   y[0] =  z[1]*x[2] - z[2]*x[1];
   y[1] = -z[0]*x[2] + z[2]*x[0];
   y[2] =  z[0]*x[1] - z[1]*x[0];

   /* mpichler, 19950515 */
   /* cross product gives area of parallelogram, which is < 1.0 for
    * non-perpendicular unit-length vectors; so normalize x, y here
    */

   mag = sqrt( x[0]*x[0] + x[1]*x[1] + x[2]*x[2] );
   if (mag) {
      x[0] /= mag;
      x[1] /= mag;
      x[2] /= mag;
   }

   mag = sqrt( y[0]*y[0] + y[1]*y[1] + y[2]*y[2] );
   if (mag) {
      y[0] /= mag;
      y[1] /= mag;
      y[2] /= mag;
   }

#define M(row,col)  m[col*4+row]
   M(0,0) = x[0];  M(0,1) = x[1];  M(0,2) = x[2];  M(0,3) = 0.0;
   M(1,0) = y[0];  M(1,1) = y[1];  M(1,2) = y[2];  M(1,3) = 0.0;
   M(2,0) = z[0];  M(2,1) = z[1];  M(2,2) = z[2];  M(2,3) = 0.0;
   M(3,0) = 0.0;   M(3,1) = 0.0;   M(3,2) = 0.0;   M(3,3) = 1.0;
#undef M
   glMultMatrixd( m );

   /* Translate Eye to Origin */
   glTranslated( -eyex, -eyey, -eyez );
}



class TGlData
{
public:
  TGlData(HWND hwnd);
  void Init();
  ~TGlData();
  void Timer();
  void Paint();
  UINT uTimer; BOOL HasInited;
  HGLRC hgl; HDC hdc; HWND hwnd;
  float RotX,RotY,RotZ;
};


TGlData::TGlData(HWND h)
{ RotX=0.0f; RotY=0.0f; RotZ=0.0f;
  hwnd=h; hdc=GetDC(hwnd);
  hgl=NULL;
}

void TGlData::Init()
{
#ifdef __BORLANDC__
  _control87(MCW_EM, MCW_EM); // reprogram the FPU so it doesn't raise exceptions
#endif
  int bpp=-1,depth=16,dbl=1,acc=1;
  int pixelformat=ChoosePixelFormatEx(hdc,&bpp,&depth,&dbl,&acc);
  if (pixelformat==0) return;
  PIXELFORMATDESCRIPTOR pfd; ZeroMemory(&pfd,sizeof(pfd));  pfd.nSize=sizeof(pfd); pfd.nVersion=1;
  DescribePixelFormat(hdc,pixelformat,sizeof(pfd),&pfd);
  BOOL res=SetPixelFormat(hdc,pixelformat,&pfd); if (!res) return;
  hgl=wglCreateContext(hdc);
  if (hgl==NULL) return;
  wglMakeCurrent(hdc,hgl);
  //
  glClearColor(0.0f, 0.0f, 0.2f, 0.0f);
  glClearDepth(1.0);
  glDepthFunc(GL_LEQUAL);  // Superimpose smaller Z values over larger ones
  glShadeModel(GL_SMOOTH); // Smooth shading
  glFrontFace(GL_CCW);
  glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  glAlphaFunc(GL_GEQUAL, 0.07f);
  glDisable(GL_ALPHA_TEST);
  glDisable(GL_BLEND);
  glDisable(GL_CULL_FACE);
  glEnable(GL_DEPTH_TEST); // Z buffering is go...
  //
  RECT rc; GetClientRect(hwnd,&rc); int width=rc.right, height=rc.bottom;
  double  znear = 1;     // Put the screen 1 unit in front of the "eye"
  double  zfar  = 50.0;  // It's good to have a small far/near ratio
  double  aspect = ((double)width)/((double)height);
  double  fovX = 90.0;
  glViewport(0, 0, width, height);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  //
  GLdouble fovY=fovX/aspect;
  GLdouble ymax = znear * tan( fovY * 3.14159265358 / 360.0 ), ymin = -ymax;
  GLdouble xmin = ymin * aspect, xmax = ymax * aspect;
  glFrustum(xmin, xmax, ymin, ymax, znear, zfar );
}

TGlData::~TGlData()
{ if (hgl!=NULL) {wglMakeCurrent(NULL, NULL);  wglDeleteContext(hgl); hgl=NULL;}
  if (hdc!=NULL) ReleaseDC(hwnd,hdc); hdc=NULL;
}


void TGlData::Timer()
{ RotX += 5; RotY += 5; RotZ += 5;
  Paint();
}


void TGlData::Paint()
{ glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT); // Wipe the color and Depth Buffers
  glMatrixMode(GL_MODELVIEW); glPushMatrix(); glLoadIdentity();
  xglLookAt(0,0,-10,  0,0,0,  0,1,0);  // put eye at (0,0,-10) looking at (0,0,0)
  glRotatef(RotX, 1.0f, 0.0f, 0.0f); // rotate about the x unit vector
  glRotatef(RotY, 0.0f, 1.0f, 0.0f); // rotate about the y unit vector
  glRotatef(RotZ, 0.0f, 0.0f, 1.0f); // rotate about the z unit vector
  glBegin(GL_QUADS);
    // Six faces of a cube to draw
    glColor3f(0.0f, 0.0f, 0.0f); glVertex3f(-1.0f, -1.0f, -1.0f);
    glColor3f(0.0f, 1.0f, 0.0f); glVertex3f(-1.0f,  1.0f, -1.0f);
    glColor3f(1.0f, 1.0f, 0.0f); glVertex3f( 1.0f,  1.0f, -1.0f);
    glColor3f(1.0f, 0.0f, 0.0f); glVertex3f( 1.0f, -1.0f, -1.0f);
    //
    glColor3f(1.0f, 0.0f, 1.0f); glVertex3f( 1.0f, -1.0f,  1.0f);
    glColor3f(1.0f, 1.0f, 1.0f); glVertex3f( 1.0f,  1.0f,  1.0f);
    glColor3f(0.0f, 1.0f, 1.0f); glVertex3f(-1.0f,  1.0f,  1.0f);
    glColor3f(0.0f, 0.0f, 1.0f); glVertex3f(-1.0f, -1.0f,  1.0f);
    //
    glColor3f(0.0f, 1.0f, 0.0f); glVertex3f(-1.0f,  1.0f, -1.0f);
    glColor3f(0.0f, 1.0f, 1.0f); glVertex3f(-1.0f,  1.0f,  1.0f);
    glColor3f(1.0f, 1.0f, 1.0f); glVertex3f( 1.0f,  1.0f,  1.0f);
    glColor3f(1.0f, 1.0f, 0.0f); glVertex3f( 1.0f,  1.0f, -1.0f);
    //
    glColor3f(1.0f, 0.0f, 0.0f); glVertex3f( 1.0f, -1.0f, -1.0f);
    glColor3f(1.0f, 0.0f, 1.0f); glVertex3f( 1.0f, -1.0f,  1.0f);
    glColor3f(0.0f, 0.0f, 1.0f); glVertex3f(-1.0f, -1.0f,  1.0f);
    glColor3f(0.0f, 0.0f, 0.0f); glVertex3f(-1.0f, -1.0f, -1.0f);
    //
    glColor3f(0.0f, 0.0f, 1.0f); glVertex3f(-1.0f, -1.0f,  1.0f);
    glColor3f(0.0f, 1.0f, 1.0f); glVertex3f(-1.0f,  1.0f,  1.0f);
    glColor3f(0.0f, 1.0f, 0.0f); glVertex3f(-1.0f,  1.0f, -1.0f);
    glColor3f(0.0f, 0.0f, 0.0f); glVertex3f(-1.0f, -1.0f, -1.0f);
    //
    glColor3f(1.0f, 0.0f, 0.0f); glVertex3f( 1.0f, -1.0f, -1.0f);
    glColor3f(1.0f, 1.0f, 0.0f); glVertex3f( 1.0f,  1.0f, -1.0f);
    glColor3f(1.0f, 1.0f, 1.0f); glVertex3f( 1.0f,  1.0f,  1.0f);
    glColor3f(1.0f, 0.0f, 1.0f); glVertex3f( 1.0f, -1.0f,  1.0f);
  glEnd();
  glPopMatrix();
  SwapBuffers(hdc);
}



LRESULT CALLBACK MyScreenSaverProc(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam)
{ TGlData *gd=(TGlData *)GetWindowLong(hwnd,GWL_USERDATA);
  if (gd==NULL && msg!=WM_CREATE) return DefScreenSaverProc(hwnd,msg,wParam,lParam);
  switch (msg) {
    case (WM_CREATE):
      { TGlData *gd=new TGlData(hwnd);
        gd->uTimer=SetTimer(hwnd,1,20,NULL);
		gd->HasInited=FALSE;
        SetWindowLong(hwnd,GWL_USERDATA,(LONG)gd);
      } break;
    case (WM_ERASEBKGND):
      { RECT rc; HDC hdc=GetDC(hwnd); GetClientRect(hwnd,&rc);
        FillRect(hdc,&rc,(HBRUSH)GetStockObject(BLACK_BRUSH)); ReleaseDC(hwnd,hdc);
      } break;
    case (WM_TIMER):
      { if (!gd->HasInited) gd->Init(); gd->HasInited=TRUE;
	    gd->Timer();
      } break; 
    case (WM_DESTROY):
      { if (gd->uTimer) KillTimer(hwnd,gd->uTimer); delete gd;
        SetWindowLong(hwnd,GWL_USERDATA,0);
      } break;
  }
  return DefScreenSaverProc(hwnd,msg,wParam,lParam);
}


BOOL HaveChangedMode=FALSE;
void ChangeMode(BOOL b)
{ // changing into mode...
  if (b)
  { DEVMODE dm; ZeroMemory(&dm,sizeof(dm)); dm.dmSize=sizeof(dm);
    dm.dmFields=DM_PELSWIDTH | DM_PELSHEIGHT;
	dm.dmPelsWidth=640; dm.dmPelsHeight=480;
    LONG res=::ChangeDisplaySettings(&dm,CDS_TEST);
    if (res==DISP_CHANGE_SUCCESSFUL)
	{ ChangeDisplaySettings(&dm,CDS_FULLSCREEN);
	  HaveChangedMode=TRUE;
	}
	return;
  }
  // changing out of mode...
  if (!b)
  { if (HaveChangedMode) ChangeDisplaySettings(NULL,NULL); HaveChangedMode=FALSE;
    return;
  }
}


HWND CALLBACK LuCreateSaver(BOOL ispreview,CREATESTRUCT &cs)
{ if (!ispreview)
  { ChangeMode(TRUE);
    cs.cx=GetSystemMetrics(SM_CXSCREEN); cs.cy=GetSystemMetrics(SM_CYSCREEN);
    // Normally, ScrPlus supplies us with suggest cx and cy.
    // But because we've changed mode, their suggestions will not be valid
  }
  return CreateWindowEx(cs.dwExStyle,"glsaverclass","SaverWindow",cs.style,
								cs.x,cs.y,cs.cx,cs.cy,cs.hwndParent,NULL,
								hInstance,cs.lpCreateParams);
}


int  CALLBACK LuConfig(HWND,CONFIGSHEETHEADER &psh,BOOL)
{ return PropertySheetA(&psh);
}



int WINAPI WinMain(HINSTANCE h,HINSTANCE,LPSTR,int)
{ // The following few lines are to set up global preferences
  hInstance=h;
  WNDCLASS wcs;
  wcs.style=CS_VREDRAW | CS_HREDRAW | CS_OWNDC; // Must have OWNDC for OpenGL
  wcs.lpfnWndProc=MyScreenSaverProc;
  wcs.cbClsExtra=0;
  wcs.cbWndExtra=0;
  wcs.hInstance=hInstance; // instance of the procedure
  wcs.hIcon=NULL;
  wcs.hCursor=NULL;
  wcs.hbrBackground=(HBRUSH)GetStockObject(NULL_BRUSH);
  wcs.lpszMenuName=NULL;
  wcs.lpszClassName="glsaverclass";
  RegisterClass(&wcs);
                      
  TScrPrefs *p=ScrCreatePrefs();
  p->AllowActiveConfig=TRUE;
  p->AllowMute=FALSE;
  p->IsTrace=FALSE;
  p->IsDebug=FALSE;
  ScrSetPrefs(p);        
  int res=ScrExecuteEx(hInstance,LuCreateSaver,LuConfig);
  ChangeMode(FALSE);
  return res;
}


