//
// Puzzle screensaver
// -------------------
// (c) 1998 Lucian Wischik
//


#define WIN32_LEAN_AND_MEAN
#define STRICT
#include <windows.h>
#include <commctrl.h>
#include <shlobj.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "..\..\lib\scrplus.h"
#pragma hdrstop
#include "resource.h"



// GLOBAL VARIABLES and prototypes
HINSTANCE hInstance;
HINSTANCE hNViewLib=NULL;
typedef HBITMAP (_stdcall *NVIEWLIBLOAD) (char *FileName, BOOL ShowProgress);
NVIEWLIBLOAD pNViewLibLoad=NULL;
void SaveBitmap(char *fn,const HBITMAP hbm,const HPALETTE pal);
BOOL CALLBACK TheEndDialogProc(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam);



// CLASSES
// ---------
// There are two classes.
//
// TPuzzleImage: this stores the actual bitmap of the image and how it's scrambled,
// and is in charge of reading the registry settings, extracting jpegs that
// are embedded as resources, and so on.
//
// TPuzzleWindow: handles user interraction.
//

class TPuzzleImage
{
public:
  TPuzzleImage();
  virtual ~TPuzzleImage();
  void Clear();
  void CopyDIB(HWND hwnd,HBITMAP hbm);
  void Mix();
  void Swap(int *srci,int *dsti);
  void RememberState();
  void PickAnImage(HWND hwnd);
  void NextImage(HWND hwnd);
  BOOL IsLastPicture(int i);
  void AddCount(DWORD time,DWORD clicks);
  void HaveGotToEnd(BOOL Increment,BOOL CurrentImages); // if increment, add to count. If not currentimages, then refer to RegMainKey.
  void UseOwnPics(int uop); // -1 ignore, 0no, 1yes.
  DWORD GameTime, GameClicks;
  DWORD TimesAtEnd; // how many times we've reached the end
  BOOL bUseOwnPics;
  BOOL FindPictureFile(int i,char *c);
  void SetNextHard(int nh);
  void Hard(int newhard,BOOL AlsoSetNext);
  void OwnDir(char *newowndir);
  char cOwnDir[MAX_PATH];
  void ResetToFirstPic();
  int TopHard;
  int hard,InitHard;
  int *index;
  int numtilesx,numtilesy;
  int tilesize;
  int width,height;
  int numright;
  int currentpic;
  char RegMainKey[MAX_PATH], RegImageKey[MAX_PATH];
  //
  HBITMAP hbm;
  HDC hdc;
  HGDIOBJ hbmOldImage;
};


class TPuzzleWindow
{
public:
  TPuzzleWindow(HWND hwnd);
  ~TPuzzleWindow();
  void WMPaint(HDC hdc);
  void DrawTiles(HDC hdc,int minx,int miny,int w,int h);
  void DrawBanner(HDC hdc);
  void DrawPick(HDC hdc);
  void DrawUnplaced(HDC hdc);
  void PerformClunk(HDC hdc,int clunk1,int clunk2);
  void PerformWholeClunk(HDC hdc);
  void CancelPick();
  void EngagePuzzleMode();
  void PictureInitialization();
  void WMSetCursor();
  void WMTimer(UINT id);
  void WMKeyDown(int vk);
  void WMKeyUp(int vk);
  void WMLButtonDown(int x,int y);
  void WMMouseMove();
  DWORD IdleBeforeBanner, BannerAfterCompleting, IdleAfterNewPicture, TimeToLetFade;
  DWORD StartOfIdleTime;
  UINT idTimer, idRedrawTimer; // idRedrawTimer is a grotty hack to a problem:
  // when the mini-preview-window started, sometimes it didn't draw all the tiles.
  // So after 100ms I issue a WM_PAINT.
  DWORD LastIdleTime, StartPictureTime;  // moves within a second of StartPictureTime are ignored
  TPuzzleImage *pi;
  int offx,offy;
  int screenx,screeny;
  int ssx(int x); int ssy(int y);
  int ssw(int x,int w); int ssh(int y,int h);
  HWND hwnd;
  HBITMAP hbmOrig;
  HDC hdc;
  int mode; // 0=unengaged, 1=engaged, 2=primed with piece
  BOOL IsBlue; // whether or not the screen has been blued out
  void EngageBlue(); void ReleaseBlue(); void CheckBlue();
  BOOL banneractive;
  int bannerx,bannery;
  int bannerwidth,bannerheight,bannerheader;
  HFONT hFont;
  int pickactive;
  int pickx,picky;
  double screenscalex,screenscaley; // for previews
  //
  DWORD SessionTotalTime, PhaseStartTime, SessionTotalClicks, PhaseCurrentClicks;
  BOOL CountingPhase;
  void CountClicks(); void DontCountClicks();
  void DisplayProgress(BOOL IsVictory);
  BOOL DialogUp;
};






TPuzzleImage::TPuzzleImage()
{ int res=LoadString(hInstance,2,RegMainKey,MAX_PATH);
  if (res==0) strcpy(RegMainKey,"SOFTWARE\\Lu\\Puzzle");
  char c[10]; res=LoadString(hInstance,3,c,10); if (res==0) strcpy(c,"10");
  int i=sscanf(c,"%i",&InitHard); if (i==0) InitHard=10;
  hbm=NULL; hdc=NULL; hbmOldImage=NULL; hard=InitHard; index=NULL; currentpic=1;
  TimesAtEnd=0; bUseOwnPics=FALSE; OwnDir(NULL); UseOwnPics(-1); 
  strcpy(RegImageKey,RegMainKey); if (bUseOwnPics) strcat(RegImageKey,"\\OwnPics");
  HaveGotToEnd(FALSE,TRUE); Hard(0,FALSE); 
}
TPuzzleImage::~TPuzzleImage()
{ Clear();
}
void TPuzzleImage::Clear()
{ if (hdc!=NULL)
  { if (hbmOldImage!=NULL) SelectObject(hdc,hbmOldImage); hbmOldImage=NULL;
    DeleteDC(hdc); hdc=NULL;
  }
  if (hbm!=NULL) DeleteObject(hbm); hbm=NULL;
  if (index!=NULL) delete[] index; index=NULL;
}

void TPuzzleImage::CopyDIB(HWND hwnd,HBITMAP hbmOrig)
{ Clear();
  if (hbmOrig==NULL) return;
  BITMAP bmp;
  GetObject(hbmOrig,sizeof(bmp),&bmp);
  int cx=bmp.bmWidth;
  int cy=bmp.bmHeight;
  if (cx>1000) cx=1000; if (cy>1000) cy=1000; if (cx<=0) cx=16; if (cy<=0) cy=16;
  //RECT wrc; GetClientRect(hwnd,&wrc); int screenx=wrc.right-wrc.left, screeny=wrc.bottom-wrc.top;
  int screenx=GetSystemMetrics(SM_CXSCREEN),screeny=GetSystemMetrics(SM_CYSCREEN);
  double xscale=((double)screenx)/((double)cx);
  double yscale=((double)screeny)/((double)cy);
  double scale=xscale; if (yscale<scale) scale=yscale;
  int scaledx=(int)(((double)cx)*scale);
  int scaledy=(int)(((double)cy)*scale);
  if (scaledx>scaledy) tilesize=scaledx/hard; else tilesize=scaledy/hard;
  numtilesx=scaledx/tilesize; numtilesy=scaledy/tilesize;
  if (numtilesy==0) numtilesy=1; if (numtilesx==0) numtilesx=1;
  int finalx=numtilesx*tilesize;
  int finaly=numtilesy*tilesize;
  width=finalx; height=finaly;
  int srcwidth=(int)(((double)finalx)/((double)scaledx)*((double)cx));
  int srcheight=(int)(((double)finaly)/((double)scaledy)*((double)cy));
  if (srcwidth>cx) srcwidth=cx; if (srcheight>cy) srcheight=cy;
  int srcleft=(cx-srcwidth)/2;
  int srctop=(cy-srcheight)/2;
  // finalx, finaly are the size of the bitmap we'll be working with
  HDC hdcScreen=GetDC(NULL);
  HDC hdcOrig=CreateCompatibleDC(hdcScreen);
  HGDIOBJ hbmOldOrig=SelectObject(hdcOrig,hbmOrig);
  hdc=CreateCompatibleDC(hdcScreen);
  hbm=CreateCompatibleBitmap(hdcScreen,finalx,finaly);
  hbmOldImage=SelectObject(hdc,hbm);
  StretchBlt(hdc,0,0,finalx,finaly,hdcOrig,srcleft,srctop,srcwidth,srcheight,SRCCOPY);
  SelectObject(hdcOrig,hbmOldOrig);
  DeleteObject(hbmOrig);
  DeleteDC(hdcOrig);
  ReleaseDC(NULL,hdcScreen);
  // now to set the other things
  index=new int[hard*hard];
  for (int i=0; i<hard*hard; i++) index[i]=i;
  numright=numtilesx*numtilesy;
}

#define NEXTRAND(r) (1664525*r+1013904223)
void TPuzzleImage::Mix()
{ DWORD r=GetTickCount();
  r=NEXTRAND(r);
  for (int i=0; i<2000; i++)
  { int srcx=((r>>20) % numtilesx); r=NEXTRAND(r);
    int srcy=((r>>20) % numtilesy); r=NEXTRAND(r);
    int destx=((r>>20)% numtilesx); r=NEXTRAND(r);
    int desty=((r>>20)%numtilesy); r=NEXTRAND(r);
    int srci=srcy*hard+srcx;
    int desti=desty*hard+destx;
    int srctile=index[srci];
    int desttile=index[desti];
    // we'll ensure that at least one of them is different
    if (srci!=desttile)
    { index[srci]=desttile;
      index[desti]=srctile;
    }
  }
  numright=0;
  for (int x=0; x<numtilesx; x++)
  { for (int y=0; y<numtilesy; y++)
    { int i=y*hard+x;
      if (index[i]==i) numright++;
    }
  }
}
void TPuzzleImage::Swap(int *srci,int *dsti)
{ int srctile=index[*srci];
  int dsttile=index[*dsti];
  if (srctile==*srci) numright--;
  if (dsttile==*dsti) numright--;
  index[*srci]=dsttile;
  index[*dsti]=srctile;
  if (*srci==dsttile) numright++;
  if (*dsti==srctile) numright++;
  if (*srci!=dsttile) *srci=-1;
  if (*dsti!=srctile) *dsti=-1;
}
void TPuzzleImage::RememberState()
{ LONG res; HKEY skey; DWORD disp;
  res=RegCreateKeyEx(HKEY_CURRENT_USER,RegImageKey,0,NULL,REG_OPTION_NON_VOLATILE,KEY_ALL_ACCESS,NULL,&skey,&disp);
  if (res!=ERROR_SUCCESS) return;
  RegSetValueEx(skey,"index",0,REG_BINARY,(LPBYTE)index,hard*hard*sizeof(int));
  RegCloseKey(skey);
}
void TPuzzleImage::SetNextHard(int nh)
{ LONG res; HKEY skey; DWORD disp,val;
  res=RegCreateKeyEx(HKEY_CURRENT_USER,RegImageKey,0,NULL,REG_OPTION_NON_VOLATILE,KEY_ALL_ACCESS,NULL,&skey,&disp);
  if (res!=ERROR_SUCCESS) return;
  val=nh; RegSetValueEx(skey,"NextHard", 0,REG_DWORD,(CONST BYTE*)&val,sizeof(val));
  RegCloseKey(skey);
  // We obviously couldn't just set the regkey 'Hard'. If we did then, upon quitten
  // and restarting the image, we would find that the registry value of 'hard'
  // was the wrong size and the user would have to restart the final image all
  // over again!
}
void TPuzzleImage::NextImage(HWND hwnd)
{ currentpic++;
  LONG res; HKEY skey; DWORD val, disp;
  res=RegCreateKeyEx(HKEY_CURRENT_USER,RegImageKey,0,NULL,REG_OPTION_NON_VOLATILE,KEY_ALL_ACCESS,NULL,&skey,&disp);
  if (res!=ERROR_SUCCESS) return;
  val=currentpic; RegSetValueEx(skey,"CurrentPic",0,REG_DWORD,(CONST BYTE*)&val,sizeof(val));
  val=0;          RegSetValueEx(skey,"numtilesx", 0,REG_DWORD,(CONST BYTE*)&val,sizeof(val));
  val=0;          RegSetValueEx(skey,"numtilesy", 0,REG_DWORD,(CONST BYTE*)&val,sizeof(val));
  RegCloseKey(skey);
  PickAnImage(hwnd);
  // What we've done is select the next picture, but corrupt the information about it. So,
  // PickAnImage will load the next image but remix it itself.
}
BOOL TPuzzleImage::FindPictureFile(int i,char *ret)
{ OwnDir(NULL);
  int numleft=i;
  char c[MAX_PATH]; strcpy(c,cOwnDir); strcat(c,"\\*.*");
  WIN32_FIND_DATA wfd;
  HANDLE hFind=FindFirstFile(c,&wfd);
  if (hFind==NULL || hFind==INVALID_HANDLE_VALUE) return FALSE;
  while (TRUE)
  { char fn[MAX_PATH]; strcpy(fn,wfd.cFileName);
    CharLowerBuff(fn,strlen(fn));
    int len=strlen(fn);
    BOOL IsPicture=FALSE;
    if (len>=5 && strcmp(&(fn[len-5]),".jpeg")==0) IsPicture=TRUE;
    if (len>=4 && strcmp(&(fn[len-4]),".jpg")==0) IsPicture=TRUE;
    if (len>=4 && strcmp(&(fn[len-4]),".bmp")==0) IsPicture=TRUE;
    if (IsPicture)
    { numleft--;
      if (numleft==0)
      { if (ret!=NULL) wsprintf(ret,"%s\\%s",cOwnDir,fn);
        FindClose(hFind);
        return TRUE;
      }
    }
    BOOL res=FindNextFile(hFind,&wfd);
    if (!res) {FindClose(hFind);return FALSE;}
  }
}
BOOL TPuzzleImage::IsLastPicture(int i)
{ currentpic=1;
  LONG res; HKEY skey; DWORD valtype, valsize, val;
  res=RegOpenKeyEx(HKEY_CURRENT_USER,RegImageKey,0,KEY_ALL_ACCESS,&skey);
  if (res==ERROR_SUCCESS)
  { valsize=sizeof(val); res=RegQueryValueEx(skey,"CurrentPic",0,&valtype,(LPBYTE)&val,&valsize);
    if (res==ERROR_SUCCESS) currentpic=val;
    RegCloseKey(skey);
  }
  if (bUseOwnPics) return !FindPictureFile(currentpic+1,NULL);
  HRSRC hrsrc=FindResource(hInstance,MAKEINTRESOURCE(currentpic+1),RT_RCDATA);
  if (hrsrc==NULL) return TRUE; else return FALSE;
}
void TPuzzleImage::PickAnImage(HWND hwnd)
{ char path[MAX_PATH];
  char fn[MAX_PATH];
  GetTempPath(MAX_PATH,path);
  GetTempFileName(path,"scj",0,fn);
  currentpic=1;
  BOOL RegOpen=FALSE;
  BOOL WrappedAround=FALSE;
  //
  LONG res; HKEY skey; DWORD valtype, valsize, val, disp;
  res=RegOpenKeyEx(HKEY_CURRENT_USER,RegImageKey,0,KEY_ALL_ACCESS,&skey);
  if (res==ERROR_SUCCESS)
  { RegOpen=TRUE;
    valsize=sizeof(val); res=RegQueryValueEx(skey,"CurrentPic",0,&valtype,(LPBYTE)&val,&valsize);
    if (res==ERROR_SUCCESS) currentpic=val;
    valsize=sizeof(val); res=RegQueryValueEx(skey,"hard",0,&valtype,(LPBYTE)&val,&valsize);
    if (res==ERROR_SUCCESS) hard=val;
  }
  //
  BOOL WillUseOwnPics=bUseOwnPics;
  if (WillUseOwnPics)
  { BOOL foundfile=FindPictureFile(currentpic,fn);
    if (!foundfile) {WrappedAround=TRUE; currentpic=1;FindPictureFile(currentpic,fn);}
    if (WrappedAround && RegOpen)
    { valsize=sizeof(val); res=RegQueryValueEx(skey,"NextHard",0,&valtype,(LPBYTE)&val,&valsize);
      if (res==ERROR_SUCCESS) hard=val;
    }
    SendMessage(hwnd,SCRM_DIALOG,TRUE,0);
    HBITMAP hbmOrig=pNViewLibLoad(fn,FALSE);
    SendMessage(hwnd,SCRM_DIALOG,FALSE,0);
    if (hbmOrig==NULL) WillUseOwnPics=FALSE; else CopyDIB(hwnd,hbmOrig);
  }
  if (!WillUseOwnPics) // a cunning way of falling through in case couldn't load bitmap
  { HRSRC hrsrc=FindResource(hInstance,MAKEINTRESOURCE(currentpic),RT_RCDATA);
    if (hrsrc==NULL) {WrappedAround=TRUE;currentpic=1; hrsrc=FindResource(hInstance,MAKEINTRESOURCE(currentpic),RT_RCDATA);}
    if (WrappedAround && RegOpen)
    { valsize=sizeof(val); res=RegQueryValueEx(skey,"NextHard",0,&valtype,(LPBYTE)&val,&valsize);
      if (res==ERROR_SUCCESS) hard=val;
    }
    DWORD size=SizeofResource(hInstance,hrsrc);
    HGLOBAL hglob=LoadResource(hInstance,hrsrc);
    LPVOID rdata=LockResource(hglob);
    HANDLE hFile=CreateFile(fn,GENERIC_WRITE,0,NULL,CREATE_ALWAYS,FILE_ATTRIBUTE_NORMAL,NULL);
    DWORD writ; WriteFile(hFile,rdata,size,&writ,NULL);
    CloseHandle(hFile);
    SendMessage(hwnd,SCRM_DIALOG,TRUE,0);
    HBITMAP hbmOrig=pNViewLibLoad(fn,FALSE);
    SendMessage(hwnd,SCRM_DIALOG,FALSE,0);
    CopyDIB(hwnd,hbmOrig);
    DeleteObject(hbmOrig);
    DeleteFile(fn);
  }
  //
  BOOL ismixed=FALSE;
  if (RegOpen)
  { int Rnumtilesx=-1,Rnumtilesy=-1;
    valsize=sizeof(val); res=RegQueryValueEx(skey,"numtilesx",0,&valtype,(LPBYTE)&val,&valsize);
    if (res==ERROR_SUCCESS) Rnumtilesx=val;
    valsize=sizeof(val); res=RegQueryValueEx(skey,"numtilesy",0,&valtype,(LPBYTE)&val,&valsize);
    if (res==ERROR_SUCCESS) Rnumtilesy=val;
    if (Rnumtilesx==numtilesx && Rnumtilesy==numtilesy)
    { valsize=hard*hard*sizeof(int); res=RegQueryValueEx(skey,"index",0,&valtype,(LPBYTE)index,&valsize);
      if (res==ERROR_SUCCESS && valsize==hard*hard*sizeof(int))
      { ismixed=TRUE;
        numright=0;
        for (int x=0; x<numtilesx; x++)
        { for (int y=0; y<numtilesy; y++)
          { int i=y*hard+x;
            if (index[i]==i) numright++;
          }
        }
      }
    }
    RegCloseKey(skey);
  }
  // Note that the routine NextImage() will have corrupted the state stored in the
  // registry. hence, if we are being called from NextImage, then ismixed will certainly
  // be false.
  if (!ismixed) Mix();
  //
  if (RegOpen && ismixed) return; // ie. if we loaded the state in from the registry
  // otherwise, we'll write the entire state to the registry
  res=RegCreateKeyEx(HKEY_CURRENT_USER,RegImageKey,0,NULL,REG_OPTION_NON_VOLATILE,KEY_ALL_ACCESS,NULL,&skey,&disp);
  if (res!=ERROR_SUCCESS) return;
  val=currentpic; RegSetValueEx(skey,"CurrentPic",0,REG_DWORD,(CONST BYTE*)&val,sizeof(val));
  val=hard;       RegSetValueEx(skey,"hard",      0,REG_DWORD,(CONST BYTE*)&val,sizeof(val));
  val=numtilesx;  RegSetValueEx(skey,"numtilesx", 0,REG_DWORD,(CONST BYTE*)&val,sizeof(val));
  val=numtilesy;  RegSetValueEx(skey,"numtilesy", 0,REG_DWORD,(CONST BYTE*)&val,sizeof(val));
  if (WrappedAround)
  { val=0; RegSetValueEx(skey,"GameTime",0,REG_DWORD,(CONST BYTE*)&val,sizeof(val));
    val=0; RegSetValueEx(skey,"GameClicks",0,REG_DWORD,(CONST BYTE*)&val,sizeof(val));
  }
  RegSetValueEx(skey,"index",0,REG_BINARY,(LPBYTE)index,hard*hard*sizeof(int));
  RegCloseKey(skey);
}
void TPuzzleImage::ResetToFirstPic()
{ LONG res; HKEY skey; DWORD disp,val;
  res=RegCreateKeyEx(HKEY_CURRENT_USER,RegImageKey,0,NULL,REG_OPTION_NON_VOLATILE,KEY_ALL_ACCESS,NULL,&skey,&disp);
  currentpic=1; RegSetValueEx(skey,"CurrentPic",0,REG_DWORD,(CONST BYTE*)&val,sizeof(val));
  RegCloseKey(skey);
}
void TPuzzleImage::AddCount(DWORD SessionTime,DWORD SessionClicks)
{ GameTime=0; GameClicks=0;
  LONG res; HKEY skey; DWORD disp,valsize,val,valtype;
  res=RegCreateKeyEx(HKEY_CURRENT_USER,RegImageKey,0,NULL,REG_OPTION_NON_VOLATILE,KEY_ALL_ACCESS,NULL,&skey,&disp);
  if (res!=ERROR_SUCCESS) return;
  valsize=sizeof(val); res=RegQueryValueEx(skey,"GameTime",0,&valtype,(LPBYTE)&val,&valsize);
  if (res==ERROR_SUCCESS) GameTime=val;
  valsize=sizeof(val); res=RegQueryValueEx(skey,"GameClicks",0,&valtype,(LPBYTE)&val,&valsize);
  if (res==ERROR_SUCCESS) GameClicks=val;
  GameTime=GameTime+SessionTime;
  GameClicks=GameClicks+SessionClicks;
  val=GameTime;  RegSetValueEx(skey,"GameTime",0,REG_DWORD,(CONST BYTE*)&val,sizeof(val));
  val=GameClicks;RegSetValueEx(skey,"GameClicks",0,REG_DWORD,(CONST BYTE*)&val,sizeof(val));
  RegCloseKey(skey);
}
void TPuzzleImage::HaveGotToEnd(BOOL Increment,BOOL CurrentImages)
{ LONG res; HKEY skey; DWORD disp,valsize,val,valtype;
  res=RegCreateKeyEx(HKEY_CURRENT_USER,CurrentImages?RegImageKey:RegMainKey,0,NULL,REG_OPTION_NON_VOLATILE,KEY_ALL_ACCESS,NULL,&skey,&disp);
  if (res!=ERROR_SUCCESS) {TimesAtEnd=(Increment?1:0); return;}
  valsize=sizeof(val); res=RegQueryValueEx(skey,"TimesAtEnd",0,&valtype,(LPBYTE)&val,&valsize);
  if (res==ERROR_SUCCESS) TimesAtEnd=val; else TimesAtEnd=0;
  if (Increment) TimesAtEnd++;
  val=TimesAtEnd;  RegSetValueEx(skey,"TimesAtEnd",0,REG_DWORD,(CONST BYTE*)&val,sizeof(val));
  RegCloseKey(skey);
}
void TPuzzleImage::UseOwnPics(int uop)
{ LONG res; HKEY skey; DWORD disp,valsize,val,valtype;
  res=RegCreateKeyEx(HKEY_CURRENT_USER,RegMainKey,0,NULL,REG_OPTION_NON_VOLATILE,KEY_ALL_ACCESS,NULL,&skey,&disp);
  if (res!=ERROR_SUCCESS) {bUseOwnPics=FALSE; return;}
  if (uop==-1)
  { valsize=sizeof(val); res=RegQueryValueEx(skey,"UseOwnPics",0,&valtype,(LPBYTE)&val,&valsize);
    if (res==ERROR_SUCCESS) bUseOwnPics=val; else bUseOwnPics=FALSE;
  } else bUseOwnPics=(uop==1);
  val=bUseOwnPics;  RegSetValueEx(skey,"UseOwnPics",0,REG_DWORD,(CONST BYTE*)&val,sizeof(val));
  RegCloseKey(skey);
  if (!FindPictureFile(1,NULL)) bUseOwnPics=FALSE;
}
void TPuzzleImage::Hard(int newhard,BOOL AlsoNext)
{ LONG res; HKEY skey; DWORD disp,valsize,val,valtype;
  int deftophard=20; if (InitHard>deftophard) deftophard=InitHard+10;
  res=RegCreateKeyEx(HKEY_CURRENT_USER,RegImageKey,0,NULL,REG_OPTION_NON_VOLATILE,KEY_ALL_ACCESS,NULL,&skey,&disp);
  if (res!=ERROR_SUCCESS) {hard=InitHard; TopHard=deftophard; return;}
  valsize=sizeof(val); res=RegQueryValueEx(skey,"hard",0,&valtype,(LPBYTE)&val,&valsize);
  if (res==ERROR_SUCCESS) hard=val; else hard=InitHard;
  valsize=sizeof(val); res=RegQueryValueEx(skey,"TopHard",0,&valtype,(LPBYTE)&val,&valsize);
  if (res==ERROR_SUCCESS) TopHard=val; else TopHard=deftophard;
  if (newhard!=0) hard=newhard;
  if (hard>TopHard) TopHard=hard;
  val=hard;  RegSetValueEx(skey,"hard",0,REG_DWORD,(CONST BYTE*)&val,sizeof(val));
  val=TopHard;  RegSetValueEx(skey,"TopHard",0,REG_DWORD,(CONST BYTE*)&val,sizeof(val));
  if (AlsoNext) {val=hard; RegSetValueEx(skey,"NextHard",0,REG_DWORD,(CONST BYTE*)&val,sizeof(val));}
  RegCloseKey(skey);
}
void TPuzzleImage::OwnDir(char *newowndir)
{ LONG res; HKEY skey; DWORD disp,valsize,valtype;
  res=RegCreateKeyEx(HKEY_CURRENT_USER,RegMainKey,0,NULL,REG_OPTION_NON_VOLATILE,KEY_ALL_ACCESS,NULL,&skey,&disp);
  if (res!=ERROR_SUCCESS) {GetWindowsDirectory(cOwnDir,MAX_PATH); return;}
  valsize=MAX_PATH; res=RegQueryValueEx(skey,"OwnDir",0,&valtype,(LPBYTE)cOwnDir,&valsize);
  if (res!=ERROR_SUCCESS) GetWindowsDirectory(cOwnDir,MAX_PATH);
  if (newowndir!=NULL) strcpy(cOwnDir,newowndir);
  RegSetValueEx(skey,"OwnDir",0,REG_SZ,(LPBYTE)cOwnDir,strlen(cOwnDir));
  RegCloseKey(skey);
}

/*
3                       RCDATA  DISCARDABLE     "three.jpg"
4                       RCDATA  DISCARDABLE     "four.jpg"
5                       RCDATA  DISCARDABLE     "five.jpg"
6                       RCDATA  DISCARDABLE     "six.jpg"
7                       RCDATA  DISCARDABLE     "seven.jpg"
8                       RCDATA  DISCARDABLE     "eight.jpg"
*/


// -----------------------------------------------------------------




TPuzzleWindow::TPuzzleWindow(HWND h)
{ IdleBeforeBanner=12000; BannerAfterCompleting=6000; IdleAfterNewPicture=1500;
  TimeToLetFade=100000;
  DialogUp=FALSE;
  CountingPhase=FALSE;
  hwnd=h; pi=new TPuzzleImage();
  pi->PickAnImage(hwnd);
  mode=0; StartOfIdleTime=GetTickCount();
  idTimer=SetTimer(hwnd,1,1500,NULL);
  idRedrawTimer=SetTimer(hwnd,2,200,NULL);
  LastIdleTime=GetTickCount();
  srand(LastIdleTime);
  pickactive=FALSE;
  banneractive=FALSE;
  PictureInitialization();
  //
  LOGFONT cursfont;
  cursfont.lfHeight         =  18;
  cursfont.lfWidth          =  0;
  cursfont.lfEscapement     =  0;
  cursfont.lfOrientation    =  0;
  cursfont.lfWeight         =  FW_BOLD;
  cursfont.lfItalic         =  TRUE;
  cursfont.lfUnderline      =  FALSE;
  cursfont.lfStrikeOut      =  FALSE;
  cursfont.lfCharSet        =  ANSI_CHARSET;
  cursfont.lfOutPrecision   =  OUT_DEFAULT_PRECIS;
  cursfont.lfClipPrecision  =  CLIP_DEFAULT_PRECIS;
  cursfont.lfQuality        =  DEFAULT_QUALITY;
  cursfont.lfPitchAndFamily =  VARIABLE_PITCH | FF_SWISS;
  strcpy(cursfont.lfFaceName, "");
  hFont=CreateFontIndirect((LPLOGFONT) &cursfont);
  //
}
void TPuzzleWindow::PictureInitialization()
{ screenx=GetSystemMetrics(SM_CXSCREEN);
  screeny=GetSystemMetrics(SM_CYSCREEN);
  RECT wrc; GetClientRect(hwnd,&wrc);
  int clientx=wrc.right-wrc.left; int clienty=wrc.bottom-wrc.top;
  screenscalex=((double)clientx)/((double)screenx);
  screenscaley=((double)clienty)/((double)screeny);
  offx=(screenx-pi->width)/2;
  offy=(screeny-pi->height)/2;
  HDC hdcScreen=GetDC(NULL);
  HGDIOBJ hOldFont=SelectObject(hdcScreen,hFont);
  RECT rc; rc.top=0; rc.bottom=0; rc.right=0; rc.left=0;
  char *c[6]={"Screensaver Puzzle:","Left-click next image","Left click to move tiles","Right-click to exit","CTRL to set wallpaper","SHIFT to show bad tiles"};
  int bwidth=0;
  for (int i=0; i<6; i++)
  { char *d=c[i];
    DrawText(hdcScreen,d,strlen(d),&rc,DT_CALCRECT);
    if (rc.right>bwidth) bwidth=rc.right;
  }
  char *d="Screensaver Puzzle:\r\nLeft-click next image\r\nRight-click to exit\r\nCTRL to set wallpaper";
  rc.left=0; rc.right=bwidth; rc.top=0;
  DrawText(hdcScreen,d,strlen(d),&rc,DT_CALCRECT|DT_WORDBREAK);
  int bheight=rc.bottom; int textheight=bheight;
  bwidth=bwidth*12/10;
  bheight=bheight*12/10;
  bannerwidth=(bwidth+pi->tilesize-1)/pi->tilesize;
  bannerheight=(bheight+pi->tilesize-1)/pi->tilesize;
  int boxheight=bannerheight*pi->tilesize;
  bannerheader=(boxheight-textheight)/2;
  SelectObject(hdcScreen,hOldFont);
  ReleaseDC(NULL,hdcScreen);
  StartPictureTime=GetTickCount();
  SessionTotalTime=0; SessionTotalClicks=0; 
  CheckBlue();
}
TPuzzleWindow::~TPuzzleWindow()
{ DontCountClicks();
  pi->AddCount(SessionTotalTime,SessionTotalClicks); SessionTotalTime=0; SessionTotalClicks=0;
  //
  if (pi!=NULL) delete pi; pi=NULL;
  if (idTimer!=0) KillTimer(hwnd,idTimer); idTimer=0;
  if (idRedrawTimer!=0) KillTimer(hwnd,idRedrawTimer); idRedrawTimer=0;
  if (hFont!=NULL) DeleteObject(hFont); hFont=NULL;
}
void TPuzzleWindow::CountClicks()
{ if (CountingPhase) return;
  PhaseStartTime=GetTickCount();
  PhaseCurrentClicks=0;
  CountingPhase=TRUE;
}
void TPuzzleWindow::DontCountClicks()
{ if (!CountingPhase) return;
  SessionTotalTime=SessionTotalTime+(GetTickCount()-PhaseStartTime);
  SessionTotalClicks=SessionTotalClicks+PhaseCurrentClicks;
  CountingPhase=FALSE;
}
void TPuzzleWindow::EngageBlue()
{ if (IsBlue) return;
  IsBlue=TRUE;
  HDC hdc=GetDC(hwnd);
  DrawUnplaced(hdc);   // calls DrawTiles for incorrect tiles, hence filling them in blue
  // Damon doesn't like the pick shown when we're blue. if (mode==2) DrawPick(hdc);
  ReleaseDC(hwnd,hdc);
}
void TPuzzleWindow::ReleaseBlue()
{ if (!IsBlue) return;
  IsBlue=FALSE;
  HDC hdc=GetDC(hwnd);
  DrawUnplaced(hdc);
  if (mode==2) DrawPick(hdc);
  ReleaseDC(hwnd,hdc);
}
void TPuzzleWindow::CheckBlue()
{ SHORT res=GetKeyState(VK_SHIFT);
  if (!(res&0x8000)) {ReleaseBlue();return;}
  EngageBlue();
  LastIdleTime=GetTickCount();
}
void TPuzzleWindow::WMKeyDown(int vk)
{ LastIdleTime=GetTickCount();
  if (vk==VK_F1) DisplayProgress(FALSE);
  if (mode==0 && vk==VK_SHIFT)
  { EngagePuzzleMode(); LastIdleTime=GetTickCount();}
  if ((mode==1 || mode==2) && vk==VK_SHIFT)
  { EngageBlue(); return;
  }
  if (mode==2) return;
  if (pi->numright==pi->numtilesx*pi->numtilesy && vk==VK_CONTROL)
  { HCURSOR old=SetCursor(LoadCursor(NULL,IDC_WAIT));
    char fn[MAX_PATH];
    GetWindowsDirectory(fn,MAX_PATH);
    strcat(fn,"\\PuzzleSaver.bmp");
    SaveBitmap(fn,pi->hbm,NULL);
    SystemParametersInfo(SPI_SETDESKWALLPAPER,0,fn,TRUE);
    SetCursor(old);
  }
}
void TPuzzleWindow::WMKeyUp(int vk)
{ LastIdleTime=GetTickCount();
  ReleaseBlue();
}
void TPuzzleWindow::CancelPick()
{ if (!pickactive) return;
  pickactive=FALSE;
  mode=1;
  HDC hdc=GetDC(hwnd);
  DrawTiles(hdc,pickx,picky,1,1);
  ReleaseDC(hwnd,hdc);
}
void TPuzzleWindow::WMLButtonDown(int x,int y)
{ if (mode==0)
  { EngagePuzzleMode(); LastIdleTime=GetTickCount(); return; }
  LastIdleTime=GetTickCount();
  if (pi->numright==pi->numtilesx*pi->numtilesy)
  { HCURSOR old=SetCursor(LoadCursor(NULL,IDC_WAIT));
    pi->NextImage(hwnd);
    PictureInitialization();
    InvalidateRect(hwnd,NULL,TRUE);
    UpdateWindow(hwnd);
    SetCursor(old);
    if (mode!=0) CountClicks();
    return;
  }
  if (IsBlue) return;
  x=(int)(((double)x)/screenscalex); y=(int)(((double)y)/screenscaley);
  int tilex=(x-offx)/pi->tilesize;
  int tiley=(y-offy)/pi->tilesize;
  if (tilex>=pi->numtilesx || tiley>=pi->numtilesy || tilex<0 || tiley<0) return;
  if (mode==1)
  { int i=tiley*pi->hard+tilex;
    if (pi->index[i]==i) return; // can't pick it up if it's in the right place
    pickx=tilex; picky=tiley; pickactive=TRUE;
    HDC hdc=GetDC(hwnd);
    DrawPick(hdc);
    ReleaseDC(hwnd,hdc);
    mode=2;
    return;
  }
  if (mode==2)
  { int srci=picky*pi->hard+pickx;
    int dsti=tiley*pi->hard+tilex;
    if (pi->index[dsti]==dsti) return;
    pi->Swap(&srci,&dsti);
    PhaseCurrentClicks++;    
    pickactive=FALSE;
    mode=1;
    HDC hdc=GetDC(hwnd);
    int clunk1=srci, clunk2=dsti;
    BOOL solved=(pi->numright==pi->numtilesx*pi->numtilesy);
    if ((clunk1!=-1 || clunk2!=-1) && !solved)
    { MessageBeep(0);
      PerformClunk(hdc,clunk1,clunk2);
    }
    if (!solved)
    { DrawTiles(hdc,pickx,picky,1,1);
      DrawTiles(hdc,tilex,tiley,1,1);
    }
    if (solved)
    { MessageBeep(0);
      PerformWholeClunk(hdc);
      DrawTiles(hdc,0,0,pi->numtilesx,pi->numtilesy);
      LastIdleTime=GetTickCount()+BannerAfterCompleting-IdleBeforeBanner;
      DontCountClicks();
      pi->AddCount(SessionTotalTime,SessionTotalClicks); SessionTotalTime=0; SessionTotalClicks=0;
      if (pi->IsLastPicture(pi->currentpic))
      { pi->HaveGotToEnd(TRUE,TRUE);
        DisplayProgress(TRUE);
      }
    }
    ReleaseDC(hwnd,hdc);
  }
}
void TPuzzleWindow::DisplayProgress(BOOL IsVictory)
{ if (CountingPhase)
  { DontCountClicks();
    CountClicks();
  } // to add current phase
  pi->AddCount(SessionTotalTime,SessionTotalClicks); SessionTotalTime=0; SessionTotalClicks=0;
  if (IsVictory)
  { SendMessage(hwnd,SCRM_DIALOG,TRUE,0);
    DialogUp=TRUE;
    WMSetCursor();
    DialogBoxParam(hInstance,MAKEINTRESOURCE(DLG_THEEND),hwnd,TheEndDialogProc,(LPARAM)pi);
    if (pi->TimesAtEnd==1 && !pi->bUseOwnPics) MessageBox(hwnd,"PS. You can play with your own images! Configure them from the Settings "
                    "button in the Display Properties control panel.","A reward",MB_OK);
    SendMessage(hwnd,SCRM_DIALOG,FALSE,0);
    DialogUp=FALSE;
    WMSetCursor();
    return;
  }
  DWORD time=pi->GameTime/1000;
  DWORD sectot=time; DWORD sec=time%60; time=time/60;
  DWORD mintot=time; DWORD min=time%60; time=time/60;
  DWORD hourtot=time; DWORD hour=time%24; time=time/24;
  DWORD daystot=time; DWORD days=time;
  char ctime[2000],d[50]; strcpy(ctime,"");
  if (daystot>0) {wsprintf(d,"%i days, ",days); strcat(ctime,d); }
  if (hourtot>0) {wsprintf(d,"%i hours, ",hour); strcat(ctime,d); }
  if (mintot>0)  {wsprintf(d,"%i minutes and ",min); strcat(ctime,d);}
  wsprintf(d,"%i seconds",sec); strcat(ctime,d);
  char cclicks[500]; wsprintf(cclicks,"%i clicks",pi->GameClicks);
  char cimage[500]; wsprintf(cimage,"Currently on image %i",pi->currentpic);
  char e[3000]; wsprintf(e,"%s. %s in %s",cimage,cclicks,ctime);
  SendMessage(hwnd,SCRM_DIALOG,TRUE,0);
  DialogUp=TRUE;
  WMSetCursor();
  MessageBox(hwnd,e,"How you're doing:",MB_OK);
  SendMessage(hwnd,SCRM_DIALOG,FALSE,0);
  DialogUp=FALSE;
  WMSetCursor();
}
void TPuzzleWindow::EngagePuzzleMode()
{ mode=1;
  WMTimer(idTimer); // to turn off banner
  WMSetCursor(); // to show cursor
  HDC hdc=GetDC(hwnd);
  DrawTiles(hdc,0,0,pi->numtilesx,pi->numtilesy);
  ReleaseDC(hwnd,hdc);
  BOOL solved=(pi->numright==pi->numtilesx*pi->numtilesy);
  if (!solved) CountClicks();
}
void TPuzzleWindow::WMMouseMove()
{ DWORD t=GetTickCount();
  if (t<StartPictureTime+IdleAfterNewPicture) return;
  LastIdleTime=t;
  if (mode==0)
  { EngagePuzzleMode();
  }
}
void TPuzzleWindow::WMTimer(UINT id)
{ CheckBlue();
  BOOL isfull=FALSE; if (GetParent(hwnd)==NULL) isfull=TRUE; 
  if (id==idRedrawTimer)
  { KillTimer(hwnd,idRedrawTimer); idRedrawTimer=0;
    if (!isfull) InvalidateRect(hwnd,NULL,FALSE);
  }
  if (banneractive)
  { HDC hdc=GetDC(hwnd);
    BOOL restoretile=TRUE;
    if (mode==0)
    { DWORD t=(GetTickCount()-StartOfIdleTime)%(TimeToLetFade*2); restoretile=(t<(TimeToLetFade)); }
    if (restoretile) DrawTiles(hdc,bannerx,bannery,bannerwidth,bannerheight);
    else
    { RECT rc;
      rc.left=ssx(offx+bannerx*pi->tilesize);
      rc.right=ssx(offx+(bannerx+bannerwidth)*pi->tilesize);
      rc.top=ssy(offy+bannery*pi->tilesize);
      rc.bottom=ssy(offy+(bannery+bannerheight)*pi->tilesize);
      FillRect(hdc,&rc,(HBRUSH)GetStockObject(BLACK_BRUSH));
    }
    ReleaseDC(hwnd,hdc);
    banneractive=FALSE;
  }
  if (!isfull) return;
  if (mode!=0 && GetTickCount()>LastIdleTime+IdleBeforeBanner && !DialogUp)
  { if (pickactive) CancelPick();
    mode=0;
    WMSetCursor();
    DWORD CurrentTime=GetTickCount();
    StartOfIdleTime=CurrentTime;
    DontCountClicks();
  }
  if (mode!=0) return;
  if (pi->numtilesx<=bannerwidth) bannerx=0;
  else bannerx=((rand())>>3)%(pi->numtilesx-bannerwidth+1);
  if (pi->numtilesy<=bannerheight) bannery=0;
  else bannery=((rand())>>3)%(pi->numtilesy-bannerheight+1);
  banneractive=TRUE;
  HDC hdc=GetDC(hwnd);
  DrawBanner(hdc);
  ReleaseDC(hwnd,hdc);
}
void TPuzzleWindow::DrawUnplaced(HDC hdc)
{ for (int x=0; x<pi->numtilesx; x++)
  { for (int y=0; y<pi->numtilesy; y++)
    { int i=y*pi->hard+x;
      if (pi->index[i]!=i) DrawTiles(hdc,x,y,1,1);
    }
  }
}
void TPuzzleWindow::DrawPick(HDC hdc)
{ if (!pickactive) return;
  int scx=ssx(offx+pickx*pi->tilesize);
  int scy=ssy(offy+picky*pi->tilesize);
  int scw=ssw(offx+pickx*pi->tilesize,pi->tilesize);
  int sch=ssh(offy+picky*pi->tilesize,pi->tilesize);
  RECT rc; rc.left=scx; rc.top=scy;
  rc.right=scx+scw-1; rc.bottom=scy+sch-1;
  POINT pt[5];
  pt[0].x=rc.left; pt[0].y=rc.top; pt[1].x=rc.left; pt[1].y=rc.bottom;
  pt[2].x=rc.right;pt[2].y=rc.bottom; pt[3].x=rc.right;pt[3].y=rc.top; pt[4].x=rc.left; pt[4].y=rc.top;
  SelectObject(hdc,GetStockObject(WHITE_PEN));
  Polyline(hdc,pt,5);
  InflateRect(&rc,-1,-1);
  pt[0].x=rc.left; pt[0].y=rc.top; pt[1].x=rc.left; pt[1].y=rc.bottom;
  pt[2].x=rc.right;pt[2].y=rc.bottom; pt[3].x=rc.right;pt[3].y=rc.top; pt[4].x=rc.left; pt[4].y=rc.top;
  SelectObject(hdc,GetStockObject(BLACK_PEN));
  Polyline(hdc,pt,5);
  InflateRect(&rc,-1,-1);
  pt[0].x=rc.left; pt[0].y=rc.top; pt[1].x=rc.left; pt[1].y=rc.bottom;
  pt[2].x=rc.right;pt[2].y=rc.bottom; pt[3].x=rc.right;pt[3].y=rc.top; pt[4].x=rc.left; pt[4].y=rc.top;
  SelectObject(hdc,GetStockObject(WHITE_PEN));
  Polyline(hdc,pt,5);
  GdiFlush();
}
void TPuzzleWindow::DrawBanner(HDC hdc)
{ if (!banneractive) return;
  int scx=ssx(offx+bannerx*pi->tilesize);
  int scy=ssy(offy+bannery*pi->tilesize);
  int scw=ssw(offx+bannerx*pi->tilesize,bannerwidth*pi->tilesize);
  int sch=ssh(offy+bannery*pi->tilesize,bannerheight*pi->tilesize);
  RECT rc; rc.left=scx; rc.top=scy; rc.right=scx+scw; rc.bottom=scy+sch;
  FillRect(hdc,&rc,GetSysColorBrush(COLOR_ACTIVEBORDER));
  SetTextColor(hdc,GetSysColor(COLOR_CAPTIONTEXT));
  SetBkColor(hdc,GetSysColor(COLOR_ACTIVEBORDER));
  char *c;
  if (pi->numright==pi->numtilesx*pi->numtilesy) c="Screensaver Puzzle:\r\nLeft-click next image\r\nRight-click to exit\r\nCTRL to set wallpaper";
  else c="Screensaver Puzzle\r\nLeft-click to move tiles\r\nRight-click to exit\r\nSHIFT to show bad tiles";
  rc.top=rc.top+bannerheader;
  DrawText(hdc,c,strlen(c),&rc,DT_CENTER|DT_VCENTER|DT_WORDBREAK);
  SelectObject(hdc,GetStockObject(BLACK_PEN));
  MoveToEx(hdc,scx,scy,NULL);
  LineTo(hdc,scx,scy+sch-1);
  LineTo(hdc,scx+scw-1,scy+sch-1);
  LineTo(hdc,scx+scw-1,scy);
  LineTo(hdc,scx,scy);
  GdiFlush();
}
void TPuzzleWindow::PerformClunk(HDC hdc,int clunk1,int clunk2)
{ if (clunk1==-1 && clunk2==-1) return;
  DWORD starttime=GetTickCount();
  RECT rc1,rc2;
  if (clunk1!=-1)
  { int x=clunk1%pi->hard, y=clunk1/pi->hard;
    int scx=ssx(offx+x*pi->tilesize); int scy=ssy(offy+y*pi->tilesize);
    int scw=ssw(offx+x*pi->tilesize,pi->tilesize);
    int sch=ssh(offy+y*pi->tilesize,pi->tilesize);
    rc1.left=scx; rc1.top=scy; rc1.right=scx+scw; rc1.bottom=scy+sch;
  }
  if (clunk2!=-1)
  { int x=clunk2%pi->hard, y=clunk2/pi->hard;
    int scx=ssx(offx+x*pi->tilesize); int scy=ssy(offy+y*pi->tilesize);
    int scw=ssw(offx+x*pi->tilesize,pi->tilesize);
    int sch=ssh(offy+y*pi->tilesize,pi->tilesize);
    rc2.left=scx; rc2.top=scy; rc2.right=scx+scw; rc2.bottom=scy+sch;
  }
  SelectObject(hdc,GetStockObject(NULL_PEN));
  int bi[5]={BLACK_BRUSH,DKGRAY_BRUSH,GRAY_BRUSH,LTGRAY_BRUSH,WHITE_BRUSH};
  int time[5]={0,0,000,100,200};
  for (int i=3; i<5; i++)
  { DWORD endtime=starttime+time[i];
    while (GetTickCount()<endtime) Sleep(0);
    HBRUSH hbr=(HBRUSH)GetStockObject(bi[i]);
    if (clunk1!=-1) FillRect(hdc,&rc1,hbr);
    if (clunk2!=-1) FillRect(hdc,&rc2,hbr);
    GdiFlush();
  }
}  
void TPuzzleWindow::PerformWholeClunk(HDC hdc)
{ DWORD starttime=GetTickCount();
  RECT rc; rc.left=ssx(offx); rc.right=ssx(offx+pi->numtilesx*pi->tilesize);
  rc.top=ssy(offy); rc.bottom=ssy(offy+pi->numtilesy*pi->tilesize);
  SelectObject(hdc,GetStockObject(NULL_PEN));
  int bi[5]={BLACK_BRUSH,DKGRAY_BRUSH,GRAY_BRUSH,LTGRAY_BRUSH,WHITE_BRUSH};
  int time[5]={0,0,000,100,200};
  for (int i=3; i<5; i++)
  { DWORD endtime=starttime+time[i];
    while (GetTickCount()<endtime) Sleep(0);
    HBRUSH hbr=(HBRUSH)GetStockObject(bi[i]);
    FillRect(hdc,&rc,hbr);
    GdiFlush();
  }
}  
void TPuzzleWindow::WMSetCursor()
{ if (mode==0 && !DialogUp) SetCursor(NULL);
  else SetCursor(LoadCursor(NULL,IDC_ARROW));
}
int TPuzzleWindow::ssx(int x) {return (int)(floor(((double)x)*screenscalex+0.5));}
int TPuzzleWindow::ssy(int y) {return (int)(floor(((double)y)*screenscaley+0.5));}
int TPuzzleWindow::ssw(int x,int w) {return ssx(x+w)-ssx(x);}
int TPuzzleWindow::ssh(int y,int h) {return ssy(y+h)-ssy(y);}
void TPuzzleWindow::DrawTiles(HDC hdc,int minx,int miny,int w,int h)
{ int OldStretchMode=SetStretchBltMode(hdc,COLORONCOLOR);
  for (int y=miny; y<miny+h; y++)
  { for (int x=minx; x<minx+w; x++)
    { int destx=ssx(offx+x*pi->tilesize);
      int desty=ssy(offy+y*pi->tilesize);
      int destw=ssw(offx+x*pi->tilesize,pi->tilesize);
      int desth=ssh(offy+y*pi->tilesize,pi->tilesize);
      int pos=y*pi->hard+x;
      int index=pi->index[y*pi->hard+x];
      if (pos!=index && IsBlue)
      { RECT rc; rc.left=destx; rc.top=desty; rc.right=rc.left+destw; rc.bottom=rc.top+desth;
        FillRect(hdc,&rc,GetSysColorBrush(COLOR_HIGHLIGHT));
      }
      else
      { int srcy=(index/pi->hard)*pi->tilesize;
        int srcx=(index%pi->hard)*pi->tilesize;
        StretchBlt(hdc,destx,desty,destw,desth,pi->hdc,srcx,srcy,pi->tilesize,pi->tilesize,SRCCOPY);
      }
    }
  }
  SetStretchBltMode(hdc,OldStretchMode);
}
void TPuzzleWindow::WMPaint(HDC hdc)
{ DrawTiles(hdc,0,0,pi->numtilesx,pi->numtilesy);
  if (!IsBlue) DrawPick(hdc);
  HBRUSH hbr=(HBRUSH)GetStockObject(BLACK_BRUSH);
  if (banneractive) DrawBanner(hdc);
  RECT rc; rc.left=0; rc.top=0; rc.right=ssx(offx); rc.bottom=ssy(screeny);
  FillRect(hdc,&rc,hbr);
  rc.left=ssx(offx+pi->width); rc.right=ssx(screenx);
  FillRect(hdc,&rc,hbr);
  rc.left=ssx(offx); rc.right=ssx(offx+pi->width); rc.top=0; rc.bottom=ssy(offy);
  FillRect(hdc,&rc,hbr);
  rc.top=ssy(offy+pi->height); rc.bottom=ssy(screeny);
  FillRect(hdc,&rc,hbr);
}



// -----------------------------------------------------------------



BOOL reallyclose=FALSE;
LRESULT CALLBACK PuzzleWindowProc(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam)
{ TPuzzleWindow *pw;
  if (msg==WM_CREATE)
  { pw=new TPuzzleWindow(hwnd);
    SetWindowLong(hwnd,GWL_USERDATA,(LONG)pw);
  }
  else
  { pw=(TPuzzleWindow*)GetWindowLong(hwnd,GWL_USERDATA);
  }
  if (pw==NULL) return DefScreenSaverProc(hwnd,msg,wParam,lParam);
  //
  switch (msg)
  { case WM_KEYDOWN: pw->WMKeyDown(wParam); break;
    case WM_KEYUP: pw->WMKeyUp(wParam); break;
    case WM_LBUTTONDOWN: pw->WMLButtonDown(LOWORD(lParam),HIWORD(lParam)); return 0;
    // catch the up message, so it doesn't get sent to the desktop after we quit
    case WM_MBUTTONUP: case WM_RBUTTONUP:
    { BOOL isfull=FALSE; if (GetParent(hwnd)==NULL) isfull=TRUE;
      if (isfull) {reallyclose=TRUE; SendMessage(hwnd,WM_CLOSE,0,0);} return 0;
    }
    case WM_MOUSEMOVE: pw->WMMouseMove(); return 0;
    case WM_TIMER: pw->WMTimer(wParam); break;
    case WM_PAINT:
    { PAINTSTRUCT ps; BeginPaint(hwnd,&ps);
      pw->WMPaint(ps.hdc); EndPaint(hwnd,&ps);
    } break;
    case WM_SETCURSOR: pw->WMSetCursor(); return 0;
	case (WM_CLOSE):
    { // Read the MouseVC example saver for an explanation of this.
	  BOOL isfull=FALSE; if (GetParent(hwnd)==NULL) isfull=TRUE;
	  if (isfull && !reallyclose) return FALSE;
      pw->pi->RememberState();
	} break;
  }
  //
  if (msg==WM_DESTROY)
  { SetWindowLong(hwnd,GWL_USERDATA,0);
    delete pw;
    PostQuitMessage(0);
  }
  //
  return DefScreenSaverProc(hwnd,msg,wParam,lParam);
}
HWND CALLBACK LuCreateSaver(BOOL,CREATESTRUCT &cs)
{ return CreateWindowEx(cs.dwExStyle,"PuzzleWindowClass","SaverWindow",cs.style,
								cs.x,cs.y,cs.cx,cs.cy,cs.hwndParent,NULL,
								hInstance,cs.lpCreateParams);
}






// -----------------------------------------------------------------

BOOL FileExists(char *fn)
{ HANDLE hFile=CreateFile(fn,0,FILE_SHARE_READ|FILE_SHARE_WRITE,NULL,OPEN_EXISTING,0,NULL);
  if (hFile==INVALID_HANDLE_VALUE) return FALSE;
  CloseHandle(hFile); return TRUE;
}

void InstallNView(char *fn)
{ HRSRC hrsrc=FindResource(hInstance,"NVIEWDLLIMAGE",RT_RCDATA);
  if (hrsrc==NULL) return;
  DWORD size=SizeofResource(hInstance,hrsrc);
  HGLOBAL hglob=LoadResource(hInstance,hrsrc);
  LPVOID rdata=LockResource(hglob);
  HANDLE hFile=CreateFile(fn,GENERIC_WRITE,0,NULL,CREATE_ALWAYS,FILE_ATTRIBUTE_NORMAL,NULL);
  DWORD writ; WriteFile(hFile,rdata,size,&writ,NULL);
  CloseHandle(hFile);
}


DWORD GetVersion(char *fn)
{ DWORD hVersion;
  DWORD vis=GetFileVersionInfoSize(fn,&hVersion);
  if (vis==0) return 0;
  void *vData;
  vData=(void *)new char[(UINT)vis];
  if (!GetFileVersionInfo(fn,hVersion,vis,vData)) {delete vData;return 0;}
  char vn[100];
  strcpy(vn,"\\VarFileInfo\\Translation");
  LPVOID transblock; UINT vsize;
  BOOL res=VerQueryValue(vData,vn,&transblock,&vsize);
  if (!res) {delete vData;return 0;}
  // Swap the words so wsprintf will print the lang-charset in the correct format.
  *(DWORD *)transblock = MAKELONG(HIWORD(*(DWORD *)transblock), LOWORD(*(DWORD *)transblock));
  wsprintf(vn,"\\StringFileInfo\\%08lx\\%s",*(DWORD *)transblock,"FileVersion");
  char *ver;
  res=VerQueryValue(vData,vn,(LPVOID*)&ver,&vsize);
  if (!res) {delete vData; return 0;}
  // The following code converts a string like 1.0 or 1.1.5 into
  // a DWORD where the first digit is the leading byte, and so on.
  // It discards version numbers greater than 255, and discards
  // anything after the fourth version sub-part.
  DWORD v=0;
  BOOL atend=FALSE;
  char *first=ver;
  char *next=first;
  int places=0;
  while (!atend)
  { while (*next!=0 && *next!='.') next++;
    if (*next==0) atend=TRUE; *next=0;
    int i; int res=sscanf(first,"%i",&i);
    if (res==0) atend=TRUE; else {v=(v<<8)|(i&255); places++;}
    next++; first=next;
    if (places==4) atend=TRUE;
  }
  for (int i=places; i<4; i++) v=v<<8;
  delete vData;
  return v;
}


BOOL TryToLoadNView()
{ // First, we'll just try in plain simplicity to load it 
  hNViewLib=LoadLibrary("nviewlib.dll");
  if (hNViewLib!=NULL)
  { pNViewLibLoad=(NVIEWLIBLOAD)GetProcAddress(hNViewLib,"NViewLibLoad");
    if (pNViewLibLoad!=NULL) return TRUE;
    FreeLibrary(hNViewLib);
  }
  // If that didn't work--either because it wasn't there or because it was an
  // earlier version--then we'll have to install it ourselves perhaps.
  char syspath[MAX_PATH];
  GetSystemDirectory(syspath,MAX_PATH);
  char nviewsyspath[MAX_PATH];
  strcpy(nviewsyspath,syspath); strcat(nviewsyspath,"\\nviewlib.dll");
  if (!FileExists(nviewsyspath))
  { InstallNView(nviewsyspath);
  }
  else
  { // Otherwise, we'll check versions first
    char path[MAX_PATH];
    char fn[MAX_PATH];
    GetTempPath(MAX_PATH,path);
    GetTempFileName(path,"scj",0,fn);
    InstallNView(fn);
    DWORD origv=GetVersion(nviewsyspath);
    DWORD newv=GetVersion(fn);
    DeleteFile(fn);
    if (origv<newv)
    { int res=MessageBox(NULL,"You have an out-of-date version of NViewLib.dll on your system. I need a more recent version. May I try to upgrade, please?","Puzzle Saver",MB_YESNOCANCEL);
      if (res!=IDYES) return FALSE;
      BOOL br=DeleteFile(nviewsyspath);
      if (!br)
      { MessageBox(NULL,"Error. Unable to delete old version of NViewLib.dll. Maybe it was in use by another application. Close all applications and try again.","Puzzle Saver Error",MB_ICONEXCLAMATION|MB_OK);
        return FALSE;
      }
      InstallNView(nviewsyspath);
    }
  }
  hNViewLib=LoadLibrary(nviewsyspath);
  if (hNViewLib!=NULL)
  { pNViewLibLoad=(NVIEWLIBLOAD)GetProcAddress(hNViewLib,"NViewLibLoad");
    if (pNViewLibLoad!=NULL) return TRUE;
    FreeLibrary(hNViewLib);
  }
  MessageBox(NULL,"Error. The version of NVIEWLIB.DLL is wrong somehow. Don't know why.","Puzzle Saver Error",MB_ICONEXCLAMATION|MB_OK);
  return FALSE;
}




BOOL CALLBACK TheEndDialogProc(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam)
{ switch (msg)
  { case WM_INITDIALOG:
    { TPuzzleImage *pi=(TPuzzleImage *)lParam;
      SetWindowLong(hwnd,DWL_USER,(LONG)pi);
      DWORD time=pi->GameTime/1000;
      DWORD sectot=time; DWORD sec=time%60; time=time/60;
      DWORD mintot=time; DWORD min=time%60; time=time/60;
      DWORD hourtot=time; DWORD hour=time%24; time=time/24;
      DWORD daystot=time; DWORD days=time;
      char ctime[1000],d[50]; strcpy(ctime,"");
      if (daystot>0) {wsprintf(d,"%i days, ",days); strcat(ctime,d); }
      if (hourtot>0) {wsprintf(d,"%i hours, ",hour); strcat(ctime,d); }
      if (mintot>0)  {wsprintf(d,"%i minutes and ",min); strcat(ctime,d);}
      wsprintf(d,"%i seconds",sec); strcat(ctime,d);
      char cclicks[500]; wsprintf(cclicks,"%i clicks",pi->GameClicks);
      char e[1000]; wsprintf(e,"It took you %s:",cclicks); SetDlgItemText(hwnd,ID_CLICKS,e);
      SetDlgItemText(hwnd,ID_TIME,ctime);
      wsprintf(e,"Well done. You've completed all %i images",pi->currentpic);
      SetDlgItemText(hwnd,ID_PICTURES,e);
      CheckRadioButton(hwnd,ID_HARD1,ID_HARD3,ID_HARD3);
      return TRUE;
    }
    case WM_COMMAND:
    { int id=LOWORD(wParam);
      if (id==IDOK)
      { TPuzzleImage *pi=(TPuzzleImage *)GetWindowLong(hwnd,DWL_USER);
        int h=pi->hard;
        if (IsDlgButtonChecked(hwnd,ID_HARD1)) h=pi->InitHard;
        if (IsDlgButtonChecked(hwnd,ID_HARD3)) h=h+2;
        pi->SetNextHard(h);
      }
      if (id==IDOK || id==IDCANCEL) EndDialog(hwnd,id);
      return TRUE;
    }
  }
  return FALSE;
}


typedef struct
{ BOOL IsBig;
  TPuzzleImage *pi;
  BOOL HasChangedDir;
  int bigheight,smallheight;
  int vdiff;
} TConfigDialogData;

#define OWNM_MAKEBIG WM_APP
#define OWNM_MAKESMALL WM_APP+1
BOOL CALLBACK OwnPicturesDialogProc(HWND hwnd, UINT msg,WPARAM wParam,LPARAM lParam)
{ if (msg==WM_INITDIALOG) SetWindowLong(hwnd,DWL_USER,lParam);
  TConfigDialogData *cdd=(TConfigDialogData*)GetWindowLong(hwnd,DWL_USER);
  if (cdd==NULL) return FALSE;
  TPuzzleImage *pi=cdd->pi;
  switch (msg)
  { case WM_INITDIALOG:
    { pi->Hard(0,FALSE); // get it to read in 'hard'
      pi->OwnDir(NULL); // get it to read in owndir
      SetDlgItemText(hwnd,ID_PICDIR,pi->cOwnDir);
      int min=3, max=20; if (pi->TopHard>max) max=pi->TopHard;
      SendDlgItemMessage(hwnd,ID_PICDIFF,TBM_SETRANGE,FALSE,MAKELONG(min,max));
      SendDlgItemMessage(hwnd,ID_PICDIFF,TBM_SETPOS,TRUE,pi->hard);
      SendMessage(hwnd,WM_HSCROLL,0,0); // display the right (5x5) thing
      pi->HaveGotToEnd(FALSE,FALSE); // to make it load in the TimesAtEnd variable for the main inbuilt images
      BOOL candopics=(pi->TimesAtEnd!=0);
      EnableWindow(GetDlgItem(hwnd,ID_MOREORLESS),candopics);
      ShowWindow(GetDlgItem(hwnd,ID_MOREORLESS),candopics?SW_SHOW:SW_HIDE);
      pi->UseOwnPics(-1); // to read in state
      CheckDlgButton(hwnd,ID_ORIGPICS,!pi->bUseOwnPics);
      CheckDlgButton(hwnd,ID_OWNPICS,pi->bUseOwnPics);
      SendMessage(hwnd,WM_COMMAND,MAKEWPARAM(ID_OWNPICS,0),0);
      RECT rc; GetWindowRect(hwnd,&rc); cdd->bigheight=rc.bottom-rc.top;
      RECT cc; GetClientRect(hwnd,&cc);
      RECT sc; GetWindowRect(GetDlgItem(hwnd,ID_MOREORLESS),&sc);
      RECT oc; GetWindowRect(GetDlgItem(hwnd,IDOK),&oc);
      cdd->vdiff=oc.bottom-sc.bottom;
      cdd->smallheight=rc.bottom-rc.top-cdd->vdiff;
      SendMessage(hwnd,OWNM_MAKESMALL,0,0);
      return TRUE;
    }
    case OWNM_MAKEBIG:
    { if (cdd->IsBig) return FALSE;
      RECT rc; GetWindowRect(hwnd,&rc);
      MoveWindow(hwnd,rc.left,rc.top,rc.right-rc.left,cdd->bigheight,TRUE);
      SetDlgItemText(hwnd,ID_MOREORLESS,"&Less <<");
      EnableWindow(GetDlgItem(hwnd,ID_MOREORLESS),FALSE);
      ShowWindow(GetDlgItem(hwnd,ID_MOREORLESS),SW_HIDE);
      GetWindowRect(GetDlgItem(hwnd,IDOK),&rc);
      SendMessage(hwnd,DM_SETDEFID,IDOK,0);
      SendDlgItemMessage(hwnd,IDOK,BM_SETSTYLE,BS_DEFPUSHBUTTON,TRUE);
      SetFocus(GetDlgItem(hwnd,IDOK));
      POINT pt; pt.x=rc.left; pt.y=rc.top; ScreenToClient(hwnd,&pt);
      MoveWindow(GetDlgItem(hwnd,IDOK),pt.x,pt.y+cdd->vdiff,rc.right-rc.left,rc.bottom-rc.top,TRUE);
      cdd->IsBig=TRUE;
    } break;
    case OWNM_MAKESMALL:
    { if (!cdd->IsBig) return FALSE;
      RECT rc; GetWindowRect(hwnd,&rc);
      MoveWindow(hwnd,rc.left,rc.top,rc.right-rc.left,cdd->smallheight,TRUE);
      SetDlgItemText(hwnd,ID_MOREORLESS,"&More >>");
      GetWindowRect(GetDlgItem(hwnd,IDOK),&rc);
      POINT pt; pt.x=rc.left; pt.y=rc.top; ScreenToClient(hwnd,&pt);
      MoveWindow(GetDlgItem(hwnd,IDOK),pt.x,pt.y-cdd->vdiff,rc.right-rc.left,rc.bottom-rc.top,TRUE);
      cdd->IsBig=FALSE;
    } break;
    case WM_HSCROLL:
    { int i=SendDlgItemMessage(hwnd,ID_PICDIFF,TBM_GETPOS,0,0);
      char c[20]; wsprintf(c,"(%ix%i)",i,i);
      SetDlgItemText(hwnd,ID_PICDIFFTEXT,c);
    } break;
    case WM_COMMAND:
    { int id=LOWORD(wParam);
      switch (id)
      { case ID_ORIGPICS: case ID_OWNPICS:
        { BOOL pics=IsDlgButtonChecked(hwnd,ID_OWNPICS);
          EnableWindow(GetDlgItem(hwnd,ID_PICDIR),pics);
          EnableWindow(GetDlgItem(hwnd,ID_CHANGEPICDIR),pics);
          EnableWindow(GetDlgItem(hwnd,ID_PICDIFF),pics);
          EnableWindow(GetDlgItem(hwnd,ID_PICDIFFTEXT),pics);
        } break;
        case ID_MOREORLESS:
        { SendMessage(hwnd,cdd->IsBig?OWNM_MAKESMALL:OWNM_MAKEBIG,0,0);
        } break;
        case IDOK:
        { BOOL pics=IsDlgButtonChecked(hwnd,ID_OWNPICS);
          pi->UseOwnPics(pics);
          if (pics)
          { char c[MAX_PATH]; GetDlgItemText(hwnd,ID_PICDIR,c,MAX_PATH);
            pi->OwnDir(c); // to write OwnDir
            if (cdd->HasChangedDir) pi->ResetToFirstPic();          
            int i=SendDlgItemMessage(hwnd,ID_PICDIFF,TBM_GETPOS,0,0);
            pi->Hard(i,TRUE); // to write Hard
          }
          EndDialog(hwnd,id);
        } break;
        case IDCANCEL: EndDialog(hwnd,id); break;
        case ID_CHANGEPICDIR:
        { char c[MAX_PATH];
          BROWSEINFO bi;
          bi.hwndOwner=hwnd;
          bi.pidlRoot=NULL; //pidl;
          bi.pszDisplayName=c;
          bi.lpszTitle="Directory with your own images:";
          bi.ulFlags=BIF_RETURNONLYFSDIRS;
		      bi.lpfn=NULL;
          LPITEMIDLIST pidl=SHBrowseForFolder(&bi);
          if (pidl!=NULL)
          {	SHGetPathFromIDList(pidl,c);
            SetDlgItemText(hwnd,ID_PICDIR,c);
            cdd->HasChangedDir=TRUE;
          }
        }
      }
    } break;
  }
  return FALSE;
}
int  CALLBACK LuConfig(HWND hwnd,CONFIGSHEETHEADER &,BOOL)
{ TConfigDialogData cdd;
  cdd.pi=new TPuzzleImage();
  cdd.HasChangedDir=FALSE; cdd.IsBig=TRUE;
  int res=DialogBoxParam(hInstance,MAKEINTRESOURCE(DLG_OWNPICTURES),hwnd,OwnPicturesDialogProc,(LPARAM)&cdd);
  delete cdd.pi;
  return res;
}








void RegisterClass()
{ WNDCLASS wcs;
  wcs.style=CS_VREDRAW | CS_HREDRAW;
  wcs.lpfnWndProc=PuzzleWindowProc;
  wcs.cbClsExtra=0;
  wcs.cbWndExtra=0;
  wcs.hInstance=hInstance;
  wcs.hIcon=NULL;
  wcs.hCursor=NULL;
  wcs.hbrBackground=(HBRUSH)GetStockObject(NULL_BRUSH);
  wcs.lpszMenuName=NULL;
  wcs.lpszClassName="PuzzleWindowClass";
  RegisterClass(&wcs);
}
                      



int WINAPI WinMain(HINSTANCE h,HINSTANCE,LPSTR,int)
{ hInstance=h;
  if (!TryToLoadNView()) return 0;
  RegisterClass();
  // The following few lines are to set up global preferences
  TScrPrefs *p=ScrCreatePrefs();
  p->AllowMute=FALSE;
  p->IsTrace=FALSE;
  p->IsDebug=FALSE;
  ScrSetPrefs(p);
  // And this next line runs the screen saver!
  //int res=ScrExecute(hInstance,PuzzleWindowProc,NULL);
  int res=ScrExecuteEx(hInstance,LuCreateSaver,LuConfig);
  FreeLibrary(hNViewLib);
  return res;
}













// -----------------------------------------------------------------
// SaveBitmap:
// given an hbm, saves it to a .bmp file. This is so it can be
// selected as the wallpaper.
void SaveBitmap(char *fn,const HBITMAP hbm,const HPALETTE pal)
{ int byte_per_pixel=1; HDC screen=GetDC(NULL);
  HDC dibDC=CreateCompatibleDC(screen);
  ReleaseDC(NULL,screen);
  BITMAP bm; GetObject(hbm, sizeof(BITMAP), &bm);
  if (bm.bmBitsPixel>8) byte_per_pixel=3;
  int scanline=(bm.bmWidth+3)&0xFFFFFFFC;
  int bitssize=scanline*bm.bmPlanes*byte_per_pixel*bm.bmHeight;
  int dibsize=sizeof(BITMAPFILEHEADER)+sizeof(BITMAPINFO)+bitssize;
  LPBYTE dib=new BYTE[dibsize];
  BITMAPFILEHEADER *bfh=(BITMAPFILEHEADER *)dib;
  bfh->bfType=0x4D42;
  bfh->bfSize=dibsize;
  bfh->bfReserved1=0;
  bfh->bfReserved2=0;
  bfh->bfOffBits=sizeof(BITMAPFILEHEADER)+sizeof(BITMAPINFO);
  BITMAPINFO *bi=(BITMAPINFO *)(&dib[sizeof(BITMAPFILEHEADER)]);
  BITMAPINFOHEADER *bih=(BITMAPINFOHEADER *)(bi);
  bih->biSize      = sizeof(BITMAPINFOHEADER);
  bih->biWidth     = bm.bmWidth;
  bih->biHeight    = bm.bmHeight;
  bih->biPlanes    = 1;
  if (bm.bmPlanes==1) bih->biBitCount=bm.bmBitsPixel;
  else bih->biBitCount = bm.bmPlanes;
  if (bih->biBitCount>8) bih->biBitCount = 24;
  bih->biSizeImage=0;
  bih->biXPelsPerMeter=0; bih->biYPelsPerMeter= 0;
  bih->biCompression=BI_RGB;
  bih->biClrUsed=0; bih->biClrImportant=0;
  int bitsoffset=sizeof(BITMAPFILEHEADER)+sizeof(BITMAPINFO);
  int res=GetDIBits(dibDC,hbm,0,bm.bmHeight,&dib[bitsoffset],bi,DIB_RGB_COLORS);
  if (res!=0)
  { HANDLE hFile=CreateFile(fn,GENERIC_WRITE,0,NULL,CREATE_ALWAYS,FILE_ATTRIBUTE_NORMAL,NULL);
    DWORD writ; WriteFile(hFile,dib,dibsize,&writ,NULL);
    CloseHandle(hFile);
  }
  DeleteDC(dibDC);
  delete[] dib;
}
