//
// Bits
// ------
// (c) 1996 Lucian Wischik
// see 'bits.h' for details.
//

#define STRICT
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <ddraw.h>
#pragma hdrstop
#include <stdio.h>
#include <float.h>
#include "bits.h"

HINSTANCE hBitsInstance; // set by BitsInit

HINSTANCE hDDraw=NULL;
DIRECTDRAWCREATE pDirectDrawCreate=NULL;
DDSURFACEDESC *DDModes=NULL;
unsigned int DDNumModes=0;
DDCAPS DDDriverCaps;
DDCAPS DDHelCaps;

#define NumSysColours (sizeof(SysPalIndex)/sizeof(SysPalIndex[1]))
#define rgbBlack RGB(0,0,0)
#define rgbWhite RGB(255,255,255)

#define DEFAULT_PROMISE 10

// These are the GetSysColor display element identifiers
const int SysPalIndex[] = {
  COLOR_ACTIVEBORDER,  COLOR_ACTIVECAPTION,  COLOR_APPWORKSPACE,
  COLOR_BACKGROUND,  COLOR_BTNFACE,  COLOR_BTNSHADOW,
  COLOR_BTNTEXT,  COLOR_CAPTIONTEXT,  COLOR_GRAYTEXT,
  COLOR_HIGHLIGHT,  COLOR_HIGHLIGHTTEXT,  COLOR_INACTIVEBORDER,
  COLOR_INACTIVECAPTION,  COLOR_MENU,  COLOR_MENUTEXT,
  COLOR_SCROLLBAR,  COLOR_WINDOW,  COLOR_WINDOWFRAME,  COLOR_WINDOWTEXT
};

// This array translates the display elements to black and white
static COLORREF MonoColours[] = {
  rgbBlack,  rgbWhite,  rgbWhite,  rgbWhite,  rgbWhite,  rgbBlack,
  rgbBlack,  rgbBlack,  rgbBlack,  rgbBlack,  rgbWhite,  rgbWhite,
  rgbWhite,  rgbWhite,  rgbBlack,  rgbWhite,  rgbWhite,  rgbBlack,  rgbBlack
};



BOOL isqueried;
BOOL trace=FALSE;

void BitsClose();
void LoadDDrawLib();
void CloseDDrawLib();

void BitsDebug(const char *);


void BitsInit(HINSTANCE h,BOOL a_trace)
{ hBitsInstance=h;
  trace=a_trace;
  if (trace)
  { FILE *f=fopen("\\ddx.log","wt");
    fprintf(f,".\n");
    fclose(f);
    SYSTEMTIME si;
    GetSystemTime(&si);
    char c[100];
    wsprintf(c,"Starting run at %i:%i:%i",si.wHour,si.wMinute,si.wSecond);
    BitsDebug(c);
    BitsDebug("---------------");
  }
}
void BitsClose()
{ BitsDebug("Ending run.");
  if (DDModes!=NULL) delete DDModes; DDModes=NULL;
  DDNumModes=0;
  if (hDDraw!=NULL) CloseDDrawLib(); hDDraw=NULL;
}

void LoadDDrawLib()
{ if (hDDraw!=NULL) return;
  hDDraw=LoadLibrary("DDRAW.DLL");
  if (hDDraw==NULL)
  { ::MessageBox(NULL,"Could not load DDRAW.DLL: is it installed?","Error",MB_OK);
    return;
  }
  pDirectDrawCreate=(DIRECTDRAWCREATE) GetProcAddress(hDDraw,"DirectDrawCreate");
  if (pDirectDrawCreate==NULL)
  { FreeLibrary(hDDraw); hDDraw=NULL;
    ::MessageBox(NULL,"DDRAW.DLL didn't exprt the right function. Is it a correct version?","Error",MB_OK);
    return;
  }
  // otherwise, success!
}

void CloseDDrawLib()
{ pDirectDrawCreate=NULL;
  if (hDDraw!=NULL) FreeLibrary(hDDraw); hDDraw=NULL;
}



typedef struct tagBitsInitCallbackData { BOOL enumfirst; unsigned int nummodes; } TBitsInitCallbackData;
HRESULT CALLBACK EnumModeCallback(LPDDSURFACEDESC pdds,LPVOID lParam)
{ TBitsInitCallbackData *bicd=(TBitsInitCallbackData *)lParam;
  if (bicd->enumfirst) { bicd->nummodes++; return S_FALSE; } // return S_FALSE to keep going
  memcpy(&(DDModes[bicd->nummodes]),pdds,sizeof(DDSURFACEDESC));
  bicd->nummodes++; return S_FALSE;
}
HRESULT BitsQueryDD(HWND)
{ if (isqueried) return BITS_OK;
  if (hDDraw==NULL) LoadDDrawLib(); if (hDDraw==NULL) return BITSERR_NODDRAW;
  HRESULT err;
  IDirectDraw *ddtemp=NULL;
  err=pDirectDrawCreate(NULL,&ddtemp,NULL);
  if (err!=DD_OK) return BITSERR_BADDDRAW;
  // Set exclusive level, so that it lists all of them
  //err=ddtemp->SetCooperativeLevel(hwnd,DDSCL_EXCLUSIVE | DDSCL_FULLSCREEN | DDSCL_ALLOWMODEX);
  //if (err!=DD_OK) {ddtemp->Release(); return BITSERR_BADDDRAW;}
  // Now we're going to get the device caps
  DDDriverCaps.dwSize=sizeof(DDDriverCaps);
  DDHelCaps.dwSize=sizeof(DDHelCaps);
  ddtemp->GetCaps(&DDDriverCaps,&DDHelCaps);
  // Now we're going to use a clalback to enumerate the modes
  // First, just count them
  TBitsInitCallbackData bicd;
  bicd.enumfirst=TRUE; bicd.nummodes=0;
  ddtemp->EnumDisplayModes(0,NULL,&bicd,EnumModeCallback);
  // that should have armed us with 'nummodes' preoperly
  if (DDModes!=NULL) delete[] DDModes;
  DDModes=new DDSURFACEDESC[bicd.nummodes];
  bicd.enumfirst=FALSE;
  bicd.nummodes=0;
  ddtemp->EnumDisplayModes(0,NULL,&bicd,EnumModeCallback);
  DDNumModes=bicd.nummodes;
  ddtemp->Release();
  isqueried=TRUE;
  return TRUE;
}


TBits::TBits()
{ isddx=FALSE; isdib=FALSE; surface=NULL; hwnd=NULL; hDibPal=NULL;
  dd=NULL; ddpal=NULL;
  havecopiedoldcolours=FALSE;  haveremovedstatics=FALSE;
  enablesubclass=FALSE;
  issubclassed=FALSE;
  autopaint=FALSE;
  reverse=FALSE;
  hassetpalette=FALSE;
  hbrBackground=NULL;
  isgettingscreen=FALSE;
}

TBits::~TBits()
{ Reset();
}

HRESULT TBits::Reset()
{
  if (surface!=NULL) delete[] surface; surface=NULL;
  NumSurfaces=0;
  if (issubclassed)
  { if (fullsubclass)
    { SetWindowLong(hwnd,GWL_STYLE,oldstyle);
      MoveWindow(hwnd,oldsize.left,oldsize.top,oldsize.right,oldsize.bottom,TRUE);
    }
    RelinquishFullPalette();
    SubclassClose();
  }
  if (hDibPal!=NULL) DeleteObject(hDibPal); hDibPal=NULL;
  if (ddpal!=NULL) ddpal->Release(); ddpal=NULL;
  if (dd!=NULL) dd->Release(); dd=NULL;
  isdib=FALSE;
  isddx=FALSE;
  hassetpalette=FALSE;
  cx=0;
  cy=0;
  bpp=0;
  hwnd=NULL;
  return BITS_OK;
}




HRESULT TBits::Create(HWND a_hwnd,BOOL useddx,int a_NumSurfaces,unsigned int a_cx,unsigned int a_cy,unsigned int a_bpp, BOOL dufp)
{ if (isddx || isdib) return BITSERR_ALREADYCREATED;
  DibUseFullPalette=dufp;
  hwnd=a_hwnd;
  NumSurfaces=a_NumSurfaces;
  cx=a_cx; cy=a_cy; bpp=a_bpp;
  return Init(useddx);
}



HRESULT TBits::CreateCompatible(HWND a_hwnd,BOOL useddx,TBits &c,BOOL dufp)
{ if (isddx || isdib) return BITSERR_ALREADYCREATED;
  hwnd=a_hwnd;
  DibUseFullPalette=dufp;
  NumSurfaces=c.NumSurfaces;
  cx=c.cx; cy=c.cy; bpp=c.bpp;
  return Init(useddx);
}

HRESULT TBits::CreateDibCopy(HWND a_hwnd,TBitsSurface &copyfrom)
{ if (isddx || isdib) return BITSERR_ALREADYCREATED;
  hwnd=a_hwnd;
  TBits *b=copyfrom.bits;
  DibUseFullPalette=b->DibUseFullPalette;
  NumSurfaces=1;
  cx=b->cx; cy=b->cy; bpp=b->bpp;
  HRESULT res=Init(FALSE);
  if (res!=BITS_OK) return res;
  if (!hassetpalette)
  { int min=0, max=256;
    if (!DibUseFullPalette) {min=10; max=246;}
    for (int i=min; i<max; i++)
    { LogPal.Entries[i].peRed=b->LogPal.Entries[i].peRed;
      LogPal.Entries[i].peGreen=b->LogPal.Entries[i].peGreen;
      LogPal.Entries[i].peBlue=b->LogPal.Entries[i].peBlue;
    }
  }
  ChangePalette(TRUE); // lazy
  surface[0].CopyBitsFrom(copyfrom);
  return BITS_OK;
}

HRESULT TBits::CreateDibResample(HWND a_hwnd,TBitsSurface &copyfrom,unsigned int a_cx,unsigned int a_cy,unsigned int a_bpp,BOOL dufp)
{ if (isddx || isdib) return BITSERR_ALREADYCREATED;
  hwnd=a_hwnd;
  TBits *b=copyfrom.bits;
  DibUseFullPalette=dufp;
  if (a_cx==0) cx=b->cx; else cx=a_cx;
  if (a_cy==0) cy=b->cy; else cy=a_cy;
  if (a_bpp==0) bpp=b->bpp; else bpp=a_bpp;
  if (bpp==0) bpp=8;
  BOOL mustdeletebits=FALSE;
  TBits *scrbits=b;
  TBitsSurface *copybits=&copyfrom;
  if (b->isddx)
  { mustdeletebits=TRUE;
    scrbits=new TBits();
    scrbits->CreateDibCopy(NULL,copyfrom);
    copybits=&scrbits->surface[0];
  }
  NumSurfaces=1;
  HRESULT res=Init(FALSE);
  if (res!=BITS_OK)
  { if (mustdeletebits) delete scrbits; return res; }
  if (!hassetpalette)
  { int min=0, max=256;
    if (!DibUseFullPalette) {min=10; max=246;}
    for (int i=min; i<max; i++)
    { LogPal.Entries[i].peRed=scrbits->LogPal.Entries[i].peRed;
      LogPal.Entries[i].peGreen=scrbits->LogPal.Entries[i].peGreen;
      LogPal.Entries[i].peBlue=scrbits->LogPal.Entries[i].peBlue;
    }
  }
  ChangePalette(TRUE); // lazy
  if (scrbits->cx==cx && scrbits->cy==cy) BitBlt(surface[0].Dibhdc,0,0,cx,cy,copybits->Dibhdc,0,0,SRCCOPY);
  else StretchBlt(surface[0].Dibhdc,0,0,cx,cy,copybits->Dibhdc,0,0,scrbits->cx,scrbits->cy,SRCCOPY);
  if (mustdeletebits) delete scrbits;
  return BITS_OK;
}

HRESULT TBits::CreateDibFromScreen(HWND a_hwnd,unsigned int a_cx,unsigned int a_cy)
{ if (isddx || isdib) return BITSERR_ALREADYCREATED;
  // Normally, when a user calls it, it will have 'isgettingscreen' set to FALSE
  if (!isgettingscreen)
  { TBits *screenbits=new TBits();
    screenbits->isgettingscreen=TRUE;
    screenbits->CreateDibFromScreen(NULL,0,0);
    screenbits->isgettingscreen=FALSE;
    int b_cx=a_cx, b_cy=a_cy;
    if (b_cx==0) b_cx=GetSystemMetrics(SM_CXSCREEN);
    if (b_cy==0) b_cy=GetSystemMetrics(SM_CYSCREEN);
    HRESULT res=CreateDibResample(a_hwnd,screenbits->surface[0],b_cx,b_cy,screenbits->bpp,TRUE);
    delete screenbits;
    return res;
  }
  // Otherwise, we take the size of the screen
  hwnd=a_hwnd;
  DibUseFullPalette=TRUE;
  NumSurfaces=1;
  cx=GetSystemMetrics(SM_CXSCREEN);
  cy=GetSystemMetrics(SM_CYSCREEN);
  bpp=8; // MMM
  HRESULT res=Init(FALSE);
  if (res!=BITS_OK) return res;
  HDC hdc=GetDC(HWND_DESKTOP);
  if (!hassetpalette) GetSystemPaletteEntries(hdc,0,256,LogPal.Entries);
  ChangePalette(TRUE); // lazy
  BitBlt(surface[0].Dibhdc,0,0,cx,cy,hdc,0,0,SRCCOPY);
  ReleaseDC(HWND_DESKTOP,hdc);
  return BITS_OK;
}



void TBits::CopyPaletteFrom(TBits &cb)
{ for (int i=0; i<256; i++)
  { LogPal.Entries[i].peRed=cb.LogPal.Entries[i].peRed;
    LogPal.Entries[i].peGreen=cb.LogPal.Entries[i].peGreen;
    LogPal.Entries[i].peBlue=cb.LogPal.Entries[i].peBlue;
  }
  hassetpalette=TRUE;
}

HRESULT TBits::Init(BOOL useddx)
{ if (NumSurfaces==0) return BITSERR_INVALIDPARAMS;
  if (cx==0 || cy==0) return BITSERR_INVALIDPARAMS;
  if (surface!=NULL) delete[] surface;
  surface=new TBitsSurface[NumSurfaces];
  isddx=FALSE; isdib=FALSE;
  if (useddx) return InitDdx();
  else return InitDib();
}


typedef struct tagEnumSurfaceData { int currentsurface; TBits *bits; } TEnumSurfaceData;
HRESULT WINAPI EnumSurfaceProc(LPDIRECTDRAWSURFACE dds,LPDDSURFACEDESC,TEnumSurfaceData *esd)
{ int i=esd->currentsurface;
  esd->bits->surface[i].bits=esd->bits;
  esd->bits->surface[i].dds=dds;
  char errstr[100];
  wsprintf(errstr,"got surface %i = %lx",i,dds);
  BitsDebug(errstr);
  esd->currentsurface=esd->currentsurface+1;
  return DDENUMRET_OK;
}


HRESULT TBits::InitDdx()
{ BitsDebug("--------------- InitDdx");
  if (hwnd==NULL)
  { BitsDebug("hwnd==NULL: cannot init Ddx");
    return BITSERR_NOHWND;
  }
  if (NumSurfaces==0)
  { BitsDebug("numsurfaces set at zero. Error.");
    return BITSERR_INVALIDPARAMS;
  }
  if (bpp!=8)
  { BitsDebug("bpp not set at eight. Cannot cope.");
    return BITSERR_INVALIDPARAMS;
  }
  if (hDDraw==NULL)
  { BitsDebug("Loading Ddraw.dll");
    LoadDDrawLib();
    if (hDDraw==NULL)
    { BitsDebug("Ddraw.dll couldn't be loaded");
      return BITSERR_NODDRAW;
    }
  }
  BitsDebug("Closing down any existing Bits, just to be on the safe side");
  if (ddpal!=NULL) ddpal->Release(); ddpal=NULL;
  if (dd!=NULL) dd->Release(); dd=NULL;
  HRESULT err;
  BitsDebug("Creating DD object...");
  err=pDirectDrawCreate(NULL,&dd,NULL);
  if (err!=DD_OK)
  { BitsDebug("Failed. Returning.");
    Reset();
    return BITSERR_BADDDRAW;
  }
  char errstr[100];
  wsprintf(errstr,"desired resolution: %ix%ix%i",cx,cy,bpp);
  BitsDebug(errstr);
  DWORD modexflag=0;
  if ((cx==320 && cy==200 && bpp==8) || (cx==320 && cy==240 && bpp==8))
  { BitsDebug("We will allow Modex");
    modexflag=DDSCL_ALLOWMODEX;
  }
  wsprintf(errstr,"Setting coop level for window %lx...",hwnd);
  BitsDebug(errstr);
  err=dd->SetCooperativeLevel(hwnd,DDSCL_EXCLUSIVE | DDSCL_FULLSCREEN | modexflag);
  if (err!=DD_OK)
  { switch (err)
    { case DDERR_HWNDALREADYSET: BitsDebug("HWNDALREADYSET."); break;
      case DDERR_HWNDSUBCLASSED: BitsDebug("HWNDSUBCLASSED."); break;
      case DDERR_EXCLUSIVEMODEALREADYSET: BitsDebug("ExclusiveModeAlreadySet."); break;
      case DDERR_INVALIDPARAMS: BitsDebug("Invalid params."); break;
      case DDERR_INVALIDOBJECT: BitsDebug("Invalid object."); break;
      case DDERR_OUTOFMEMORY: BitsDebug("Out of memory."); break;
      default: BitsDebug("Unknown set-coop-level error.");
    }
    BitsDebug("Failed to set cooperative level: releasing DD, and aborting");
    Reset(); return BITSERR_BADDDRAW;
  }
  BitsDebug("Setting display mode...");
  err=dd->SetDisplayMode(cx,cy,bpp);
  if (err!=DD_OK)
  { BitsDebug("Failed to set display mode: releasing DD, and aborting");
    Reset(); return BITSERR_BADDDRAW;
  }
  numbytes=cx*cy; numdwords=numbytes/4; orientation=1;
  wsprintf(errstr,"numbytes=%i numdwords=%i orientation=%i",numbytes,numdwords,orientation);
  BitsDebug(errstr);
  // Now we have to create the surface. Remember that the call
  // to 'SetDisplayMode' will have destroyed any surfaces that were
  // previously attached.
  BitsDebug("About to set surface to null");
  if (surface[0].dds!=NULL)
  { BitsDebug("releasing it first");
    surface[0].dds->Release();
    surface[0].dds=NULL;
  }
  BitsDebug("Setting up ddsd");
  DDSURFACEDESC ddsd;
  wsprintf(errstr,"Number of back buffers: %i",NumSurfaces-1);
  BitsDebug(errstr);
  ddsd.dwSize = sizeof( ddsd );
  ddsd.dwFlags = DDSD_CAPS | DDSD_BACKBUFFERCOUNT;
  ddsd.ddsCaps.dwCaps = DDSCAPS_COMPLEX | DDSCAPS_FLIP | DDSCAPS_PRIMARYSURFACE;
  ddsd.dwBackBufferCount = NumSurfaces-1;
  BitsDebug("About to create primary surface...");
  err=dd->CreateSurface(&ddsd,&surface[0].dds,NULL);
  if (err!=DD_OK)
  { switch (err)
    { case DDERR_INCOMPATIBLEPRIMARY: BitsDebug("IncompatiblePrimary."); break;
      case DDERR_INVALIDCAPS: BitsDebug("InvalidCaps."); break;
      case DDERR_INVALIDOBJECT: BitsDebug("InvalidObject."); break;
      case DDERR_INVALIDPARAMS: BitsDebug("InvalidParams."); break;
      case DDERR_INVALIDPIXELFORMAT: BitsDebug("InvalidPixelFormat."); break;
      case DDERR_NOALPHAHW: BitsDebug("No alpha hw."); break;
      case DDERR_NOCOOPERATIVELEVELSET: BitsDebug("No coop level set."); break;
      case DDERR_NODIRECTDRAWHW: BitsDebug("No direct draw hardware."); break;
      case DDERR_NOEMULATION: BitsDebug("No emulation."); break;
      case DDERR_NOEXCLUSIVEMODE: BitsDebug("No exclusive mode."); break;
      case DDERR_NOFLIPHW: BitsDebug("No flip hw."); break;
      case DDERR_NOMIPMAPHW: BitsDebug("No mipmap hw."); break;
      case DDERR_NOZBUFFERHW: BitsDebug("No zbuffer hw."); break;
      case DDERR_OUTOFMEMORY: BitsDebug("Out of memory."); break;
      case DDERR_OUTOFVIDEOMEMORY: BitsDebug("Out of video memory."); break;
      case DDERR_PRIMARYSURFACEALREADYEXISTS: BitsDebug("PrimarySurfaceAlreadyExists."); break;
      case DDERR_UNSUPPORTEDMODE: BitsDebug("Unsupported mode."); break;
      default: BitsDebug("Unknown error.");
    }
    BitsDebug("Failed to create primary surface. Releasing DD, and aborting.");
    Reset(); return BITSERR_BADDDRAW;
  }
  BitsDebug("Succesfully created surface!");
  // Now try to get pointers to the back buffers
  for (int i=1; i<NumSurfaces; i++)
  { surface[i].dds=NULL;
    surface[i].bits=this;
    DDSCAPS ddsc;
    if (i==0) ddsc.dwCaps=DDSCAPS_FLIP; else ddsc.dwCaps=DDSCAPS_FLIP;
    HRESULT err=surface[i-1].dds->GetAttachedSurface(&ddsc,&surface[i].dds);
    if (err!=DD_OK)
    { BitsDebug("Error getting surface. Aborting.");
      Reset();
      return BITSERR_DDERROR;
    }
    char c[100];
    wsprintf(c,"Got attached surface (%i) at %lx",i,surface[i].dds);
    BitsDebug(c);
  }
  BitsDebug("Got all surfaces!");

  // Palettes are the one exception. They re not destroyed when we cahnge mode.
  // (I don't think. I may be wrong).
  // I'll assume we're in a 256 colour mode.
  BitsDebug("Getting screen palette...");
  HDC hdc=GetDC(NULL);
  //if (GetDeviceCaps(hdc,RASTERCAPS) & RC_PALETTE)
  { if (!hassetpalette) GetSystemPaletteEntries(hdc,0,256,LogPal.Entries);
    BitsDebug("Got screen palette");
    if (ddpal!=NULL)
    { BitsDebug("Copying palette onto existing dibpal");
      ddpal->SetEntries(0,0,256,LogPal.Entries);
    }
    else
    { BitsDebug("Creating new dibpal");
      dd->CreatePalette(DDPCAPS_8BIT,LogPal.Entries,&ddpal,NULL);
    }
    // It's not a major calamity if the palette doesn't work. So, don't check error.
    BitsDebug("Settings palette for surface");
    surface[0].dds->SetPalette(ddpal);
  }
  ReleaseDC(NULL,hdc);
  isddx=TRUE;
  BitsDebug("InitDdx completed successfully!");
  return BITS_OK;
}

HRESULT TBits::InitDib()
{ if (bpp!=8) return BITSERR_UNSUPPORTED;
  if (hwnd==NULL) enablesubclass=FALSE;
  // So, here we create DIBs
  isddx=FALSE; isdib=TRUE;
  orientation=-1;
  LogPal.Version=0x300;
  LogPal.NumberOfEntries=256;
  HDC hdc=GetDC(NULL);
  // We're going to get syspal entries even if the screen doesn't support
  // palettes, because we might be going to create 8bpp things anyway
  // Note: the PC_RESERVED stuff is so that we create a
  // 'system identity palette', which makes blts faster on 8bpp displays
  if (!hassetpalette) GetSystemPaletteEntries(hdc,0,256,LogPal.Entries);
  for (int i=0; i<256; i++)
  { DibCols[i].rgbBlue =LogPal.Entries[i].peBlue;
    DibCols[i].rgbRed  =LogPal.Entries[i].peRed;
    DibCols[i].rgbGreen=LogPal.Entries[i].peGreen;
    DibCols[i].rgbReserved=0;
    LogPal.Entries[i].peFlags=PC_RESERVED;
  }
  // These two have to have their flags at zero, for some reason. Don't
  // ask me why.
  LogPal.Entries[0].peFlags=0;
  LogPal.Entries[255].peFlags=0;
  hDibPal=CreatePalette((LOGPALETTE *)&LogPal);
  ReleaseDC(0,hdc);
  for (i=0; i<NumSurfaces; i++)
  { BITMAPINFOHEADER bi;
    bi.biSize=sizeof(BITMAPINFOHEADER);
    bi.biPlanes=1;
    bi.biBitCount=8;
    bi.biCompression=BI_RGB;
    bi.biSizeImage=0;
    bi.biClrUsed=0;
    bi.biClrImportant=0;
    bi.biWidth=cx;
    bi.biHeight=cy;
    TBitsSurface &bs=surface[i];
    bs.bits=this;
    bs.Dibhdc=CreateCompatibleDC(NULL);
    HGDIOBJ hbm;
    hbm=CreateDIBSection(bs.Dibhdc,(BITMAPINFO *)&bi,DIB_PAL_COLORS,&bs.internaldata.lpv,NULL,0);
    bi.biSizeImage=bi.biWidth*bi.biHeight;
    cx=bi.biWidth; cy=bi.biHeight;
    if (hbm!=NULL) bs.DibhOldBitmap=SelectObject(bs.Dibhdc,hbm);
    else bs.DibhOldBitmap=NULL;
    SetDIBColorTable(bs.Dibhdc,0,256,DibCols);
    numbytes=bi.biSizeImage;
    numdwords=numbytes/4;
    if (numbytes!=0) FillMemory(bs.internaldata.lpv,numbytes,0);
    bs.DibPresent=TRUE;
  }
  isdib=TRUE;
  if (enablesubclass && fullsubclass)
  { int scrx=GetSystemMetrics(SM_CXSCREEN);
    int scry=GetSystemMetrics(SM_CYSCREEN);
    HWND hparwnd=GetParent(hwnd);
    LONG style;
    if (hparwnd==NULL) style=WS_POPUP | WS_VISIBLE;
    else style=WS_CHILD | WS_VISIBLE;
    oldstyle=SetWindowLong(hwnd,GWL_STYLE,style);
    GetWindowRect(hwnd,&oldsize);
    MoveWindow(hwnd,0,0,scrx,scry,FALSE);
  }
  if (enablesubclass) TakeFullPalette();
  DibPainter.Setup(hwnd,cx,cy);
  SubclassInit();
  return BITS_OK;
}


HRESULT TBits::CreateDibFromFile(HWND a_hwnd,const char *file)
{ if (isddx || isdib) return BITSERR_ALREADYCREATED;
  DibUseFullPalette=TRUE;
  hwnd=a_hwnd;
  NumSurfaces=1;
  isddx=FALSE; isdib=TRUE;
  orientation=-1;
  if (surface!=NULL) delete[] surface;
  surface=new TBitsSurface[NumSurfaces];
  TBitsSurface &bs=surface[0];
  bs.bits=this;
  bs.Dibhdc=CreateCompatibleDC(NULL);
  BitsDebug("About to load file image");
  HGDIOBJ hbm=LoadImage(NULL,file,IMAGE_BITMAP,0,0,LR_CREATEDIBSECTION | LR_LOADFROMFILE);
  BitsDebug("Loaded file image");
  if (hbm==NULL)
  { bs.DibhOldBitmap=NULL;
    cx=0; cy=0; numbytes=0; numdwords=0;
    LogPal.Version=0x300;
    LogPal.NumberOfEntries=256;
    HDC hdc=GetDC(NULL);
    if (!hassetpalette) GetSystemPaletteEntries(hdc,0,256,LogPal.Entries);
    for (int i=0; i<256; i++)
    { DibCols[i].rgbBlue =LogPal.Entries[i].peBlue;
      DibCols[i].rgbRed  =LogPal.Entries[i].peRed;
      DibCols[i].rgbGreen=LogPal.Entries[i].peGreen;
      DibCols[i].rgbReserved=0;
      LogPal.Entries[i].peFlags=PC_RESERVED;
    }
    LogPal.Entries[0].peFlags=0;
    LogPal.Entries[255].peFlags=0;
  }
  else
  { bs.DibhOldBitmap=SelectObject(bs.Dibhdc,hbm);
    DIBSECTION ds;
    GetObject(hbm,sizeof(ds),&ds);
    cx=ds.dsBmih.biWidth; cy=ds.dsBmih.biHeight;
    numbytes=ds.dsBmih.biSizeImage; numdwords=numbytes/4;
    bpp=ds.dsBmih.biBitCount;
    LogPal.Version=0x300;
    LogPal.NumberOfEntries=256;
    GetDIBColorTable(bs.Dibhdc,0,256,DibCols);
    for (int i=0; i<256; i++)
    { LogPal.Entries[i].peBlue =DibCols[i].rgbBlue;
      LogPal.Entries[i].peGreen=DibCols[i].rgbGreen;
      LogPal.Entries[i].peRed  =DibCols[i].rgbRed;
      DibCols[i].rgbReserved=0;
      LogPal.Entries[i].peFlags=PC_RESERVED;
    }
    LogPal.Entries[0].peFlags=0;
    LogPal.Entries[255].peFlags=0;
  }
  BitsDebug("About to create palette");
  hDibPal=CreatePalette((LOGPALETTE *)&LogPal);
  if (hbm==NULL) bs.DibPresent=FALSE;
  else
  { bs.DibPresent=TRUE;
    if (enablesubclass) TakeFullPalette();
    DibPainter.Setup(hwnd,cx,cy);
    SubclassInit();
  }
  BitsDebug("Completed create");
  return BITS_OK;
}



HRESULT TBits::CreateDibFromResource(HWND a_hwnd,int idRes)
{ if (isddx || isdib) return BITSERR_ALREADYCREATED;
  DibUseFullPalette=TRUE;
  hwnd=a_hwnd;
  NumSurfaces=1;
  isddx=FALSE; isdib=TRUE;
  orientation=-1;
  if (surface!=NULL) delete[] surface;
  surface=new TBitsSurface[NumSurfaces];
  TBitsSurface &bs=surface[0];
  bs.bits=this;
  bs.Dibhdc=CreateCompatibleDC(NULL);
  HGDIOBJ hbm=LoadImage(hBitsInstance,MAKEINTRESOURCE(idRes),IMAGE_BITMAP,0,0,LR_CREATEDIBSECTION);
  if (hbm==NULL)
  { bs.DibhOldBitmap=NULL;
    cx=0; cy=0; numbytes=0; numdwords=0;
    LogPal.Version=0x300;
    LogPal.NumberOfEntries=256;
    HDC hdc=GetDC(NULL);
    if (!hassetpalette) GetSystemPaletteEntries(hdc,0,256,LogPal.Entries);
    for (int i=0; i<256; i++)
    { DibCols[i].rgbBlue =LogPal.Entries[i].peBlue;
      DibCols[i].rgbRed  =LogPal.Entries[i].peRed;
      DibCols[i].rgbGreen=LogPal.Entries[i].peGreen;
      DibCols[i].rgbReserved=0;
      LogPal.Entries[i].peFlags=PC_RESERVED;
    }
    LogPal.Entries[0].peFlags=0;
    LogPal.Entries[255].peFlags=0;
  }
  else
  { bs.DibhOldBitmap=SelectObject(bs.Dibhdc,hbm);
    DIBSECTION ds;
    GetObject(hbm,sizeof(ds),&ds);
    cx=ds.dsBmih.biWidth; cy=ds.dsBmih.biHeight;
    numbytes=ds.dsBmih.biSizeImage; numdwords=numbytes/4;
    bpp=ds.dsBmih.biBitCount;
    GetDIBColorTable(bs.Dibhdc,0,256,DibCols);
    for (int i=0; i<256; i++)
    { LogPal.Entries[i].peBlue =DibCols[i].rgbBlue;
      LogPal.Entries[i].peGreen=DibCols[i].rgbGreen;
      LogPal.Entries[i].peRed  =DibCols[i].rgbRed;
      DibCols[i].rgbReserved=0;
      LogPal.Entries[i].peFlags=PC_RESERVED;
    }
    LogPal.Entries[0].peFlags=0;
    LogPal.Entries[255].peFlags=0;
  }
  hDibPal=CreatePalette((LOGPALETTE *)&LogPal);
  if (hbm==NULL) bs.DibPresent=FALSE;
  else
  { bs.DibPresent=TRUE;
    if (enablesubclass) TakeFullPalette();
    DibPainter.Setup(hwnd,cx,cy);
    SubclassInit();
  }
  return BITS_OK;
}




TBitsSurface *TBits::GetSurface(int i)
{ if (!isddx && !isdib) return NULL;
  if (i>=NumSurfaces) return NULL;
  return &surface[i];
}

TBitsSurface *TBits::LockSurface(int i,BOOL toread,BOOL towrite)
{ BitsDebug("locking surface...");
  TBitsSurface *s=GetSurface(i);
  if (s==NULL)
  { BitsDebug("couldn't find surface.");
    return NULL;
  }
  HRESULT err=s->Lock(toread,towrite);
  if (err!=BITS_OK) return NULL;
  return s;
}
void TBits::UnlockSurface(TBitsSurface *bs)
{ if (bs!=NULL) bs->Unlock();
}

HRESULT TBits::Flip(DWORD dwFlipFlags)
{ BitsDebug("----------- Flip");
  if (hwnd==NULL)
  { BitsDebug("hwnd is null: cannot flip");
    return BITSERR_NOHWND;
  }
  if (!isddx && !isdib)
  { InvalidateRect(hwnd,NULL,TRUE);
    return BITSERR_NOTCREATED;
  }
  BOOL anylocked=FALSE;
  for (int i=0; i<NumSurfaces; i++)
  { if (surface[i].islocked) anylocked=TRUE; }
  if (anylocked)
  { BitsDebug("some still locked. cannot flip.");
    return BITSERR_LOCKED;
  }
  if (isddx)
  { BitsDebug("About to flip ddx");
    int tries=0;
    BOOL keeptrying=TRUE;
    while (keeptrying)
    { BitsDebug("trying to flip");
      HRESULT err=surface[0].dds->Flip(NULL,dwFlipFlags);
      if (err==DD_OK) {BitsDebug("Flip succesfull!"); keeptrying=FALSE;}
      if (err==DDERR_SURFACELOST) { BitsDebug("Flip surface lost. aborting flip."); keeptrying=FALSE; }
      if (err==DDERR_WASSTILLDRAWING)
      { BitsDebug("Flip was still drawing.");
        if (tries>100) {BitsDebug("100 tries with no luck. Giving up."); keeptrying=FALSE;}
        else {BitsDebug("Sleeping then trying again.");Sleep(1);}
      }
    }
    // surface[0].dds->Flip(NULL,DDFLIP_WAIT);
    BitsDebug("Returning ddx");
    return BITS_OK;
  }
  // A DIB flip means: we rotate them all, so surface[1] becomes
  // surface[0], and then we InvalidateWindow and UpdateWindow
  if (NumSurfaces>1)
  { BitsDebug("About to flip dib");
    TBitsSurface *frontsurface=new TBitsSurface();
    memcpy(frontsurface,&surface[0],sizeof(TBitsSurface));
    for (int i=0; i<NumSurfaces-1; i++)
    { int from=(i+1)%NumSurfaces;
      int to=i;
      memcpy(&surface[to],&surface[from],sizeof(TBitsSurface));
    }
    memcpy(&surface[NumSurfaces-1],frontsurface,sizeof(TBitsSurface));
    frontsurface->DibPresent=FALSE; // otherwise it would try to delete it
    delete frontsurface;
    BitsDebug("Flipped dib");
  }
  if (surface[0].dirtypalette)
  { BitsDebug("Palette was dirty. Fixing it.");
    SetDIBColorTable(surface[0].Dibhdc,0,256,DibCols);
    surface[0].dirtypalette=FALSE;
  }
  InvalidateRect(hwnd,NULL,DibPainter.dirty);
  DibPainter.dirty=FALSE;
  BitsDebug("redrawing dib.");
  UpdateWindow(hwnd);
  BitsDebug("done dib flip");
  return BITS_OK;
}

// Before changing palette, you should first set up your new palette in LogPal
HRESULT TBits::ChangePalette(BOOL lazy)
{ LogPal.Entries[  0].peRed=0;  LogPal.Entries[  0].peGreen=0;  LogPal.Entries[  0].peBlue=0;
  LogPal.Entries[255].peRed=255;LogPal.Entries[255].peGreen=255;LogPal.Entries[255].peBlue=255;
  if (!isddx && !isdib) {hassetpalette=TRUE; return BITS_OK;}
  if (bpp!=8) return BITSERR_NOTPALETTE;
  if (isddx)
  { ddpal->SetEntries( 0, 0, 256, LogPal.Entries); // ignore error code
  }
  if (isdib)
  { // We trust them not to have overwritten bad stuff
    int min=0;
    //if (!DibUseFullPalette) min=10;
    int max=256;
    //if (!DibUseFullPalette) max=246;
    for (int i=min; i<max; i++)
    { DibCols[i].rgbRed  =LogPal.Entries[i].peRed;
      DibCols[i].rgbGreen=LogPal.Entries[i].peGreen;
      DibCols[i].rgbBlue =LogPal.Entries[i].peBlue;
    }
    SetDIBColorTable(surface[0].Dibhdc,min,max,DibCols);
    for (int s=1; s<NumSurfaces; s++) surface[s].dirtypalette=TRUE;
    if (hwnd!=NULL)
    { HDC hdc=GetDC(hwnd);
      SelectPalette(hdc,hDibPal,FALSE);
      //if (DibUseFullPalette)
        AnimatePalette(hDibPal,0,256,&LogPal.Entries[0]);
      //else AnimatePalette(hDibPal,10,236,&LogPal.Entries[NumSysColours]);
      ReleaseDC(hwnd,hdc);
    }
    if (!lazy && hwnd!=NULL)
    { InvalidateRect(hwnd,NULL,FALSE);
      UpdateWindow(hwnd);
    }
  }
  return BITS_OK;
}



HRESULT TBits::EnableSubclass(BOOL ap, HBRUSH hb)
{ if (isdib || isddx) return BITSERR_ALREADYCREATED;
  if (issubclassed) return BITSERR_ALREADYCREATED;
  enablesubclass=TRUE;
  autopaint=ap;
  hbrBackground=hb;
  reverse=FALSE;
  fullsubclass=FALSE;
  return BITS_OK;
}

HRESULT TBits::EnableFullSubclass(BOOL f)
{ if (isdib || isddx) return BITSERR_ALREADYCREATED;
  if (issubclassed) return BITSERR_ALREADYCREATED;
  enablesubclass=TRUE;
  autopaint=TRUE;
  hbrBackground=NULL;
  reverse=f;
  fullsubclass=TRUE;
  if (DibPainter.promise==DEFAULT_PROMISE) DibPainter.StretchToWindow();
  return BITS_OK;
}

HRESULT TBits::DisableSubclass()
{ if (isdib || isddx) return BITSERR_ALREADYCREATED;
  if (issubclassed) return BITSERR_ALREADYCREATED;
  enablesubclass=FALSE;
  fullsubclass=FALSE;
  return BITS_OK;
}


void TBits::WmActivateApp(WPARAM wParam,LPARAM)
{ if (!isdib) return;
  BOOL act=(BOOL)wParam;
  if (act) TakeFullPalette();
  else RelinquishFullPalette();
  HDC hdc=GetDC(hwnd);
  UnrealizeObject(hDibPal);
  SelectPalette(hdc,hDibPal,FALSE);
  RealizePalette(hdc);
  ReleaseDC(hwnd,hdc);
}

void TBits::TakeFullPalette()
{ if (!isdib) return;
  if (!DibUseFullPalette) return;
  if (haveremovedstatics) return;
  HDC hdc=GetDC(NULL);
  int palres=GetDeviceCaps(hdc,RASTERCAPS);
  if ((palres & RC_PALETTE)==0) {ReleaseDC(NULL,hdc); return;}
  // save colour mapping & switch to SYSPAL_NOSTATIC
  if (GetSystemPaletteUse(hdc)==SYSPAL_STATIC)
  { // store the current mapping
    for (int i=0; i<NumSysColours; i++) OldColours[i]=GetSysColor(SysPalIndex[i]);
    // switch to SYSPAL_NOSTATIC and remap the colours
    SetSystemPaletteUse(hdc,SYSPAL_NOSTATIC);
    SetSysColors(NumSysColours,SysPalIndex,MonoColours);
    havecopiedoldcolours=TRUE;
    haveremovedstatics=TRUE;
  }
  ReleaseDC(NULL,hdc);
}

LRESULT CALLBACK BitsWindowProc(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam)
{ TBits *bits=(TBits *)GetProp(hwnd,"PROP_BITSDATA");
  if (bits==NULL || !bits->issubclassed)
  { MessageBeep(0);
    return DefWindowProc(hwnd,msg,wParam,lParam);
  }
  switch (msg)
  { case WM_ACTIVATEAPP: bits->WmActivateApp(wParam,lParam); break;
    case WM_PAINT:
    { if (bits->autopaint) return bits->WmPaint(wParam,lParam);} break;
    case WM_PALETTECHANGED: bits->WmPaletteChanged(wParam,lParam); break;
    case WM_QUERYNEWPALETTE: return bits->WmQueryNewPalette(wParam,lParam);
    case WM_ERASEBKGND:
    { if (bits->hbrBackground!=NULL) {bits->WmEraseBkgnd(wParam,lParam); return 1;}} break;
  }
  return CallWindowProc(bits->oldwndproc,hwnd,msg,wParam,lParam);
}

void TBits::SubclassInit()
{ if (!enablesubclass) return;
  if (issubclassed) return;
  SetProp(hwnd,"PROP_BITSDATA",this);
  issubclassed=TRUE;
  oldwndproc=(WNDPROC)SetWindowLong(hwnd,GWL_WNDPROC,(LONG)BitsWindowProc);
}

void TBits::SubclassClose()
{ if (!issubclassed) return;
  SetWindowLong(hwnd,GWL_WNDPROC,(LONG)oldwndproc);
  issubclassed=FALSE;
  RemoveProp(hwnd,"PROP_BITSDATA");
}

void TBits::RelinquishFullPalette()
{ HDC hdc=GetDC(NULL);
  if (!havecopiedoldcolours || !haveremovedstatics) return;
  // if either is FALSE, there's not much we can do
  SetSystemPaletteUse(hdc,SYSPAL_STATIC);
  SetSysColors(NumSysColours,SysPalIndex,OldColours);
  haveremovedstatics=FALSE;
  ReleaseDC(NULL,hdc);
}

void TBits::WmEraseBkgnd(WPARAM wParam,LPARAM)
{ HBRUSH hbr;
  if (!issubclassed) hbr=(HBRUSH)GetStockObject(BLACK_BRUSH);
  else hbr=hbrBackground;
  // because if they've called us manually, then presumably
  // they want us to do something
  HDC hdc=(HDC)wParam;
  RECT rc; GetClientRect(hwnd,&rc);
  FillRect(hdc,&rc,hbr);
}


LRESULT TBits::WmPaint(WPARAM,LPARAM)
{ PAINTSTRUCT ps;
  HDC hdc=BeginPaint(hwnd,&ps);
  Paint(hdc,ps);
  EndPaint(hwnd,&ps);
  return FALSE;
}
void TBits::Paint(HDC hdc,PAINTSTRUCT &ps)
{ if (!isdib) return;
  if (bpp==8)
  { SelectPalette(hdc,hDibPal,FALSE);
    RealizePalette(hdc);
  }
  BOOL flip=FALSE;
  if (issubclassed && reverse) flip=TRUE;
  DibPainter.Paint(hdc,ps,surface[0].Dibhdc,flip);
}

void TBits::WmPaletteChanged(WPARAM wParam,LPARAM)
{ HWND hPalChg=(HWND)wParam;
  if (hPalChg==hwnd) return;
  if (!isdib) return;
  WmQueryNewPalette(0,0);
}

LRESULT TBits::WmQueryNewPalette(WPARAM,LPARAM)
{ if (!isdib) return FALSE;
  if (hDibPal==NULL) return FALSE;
  HDC hdc=GetDC(hwnd);
  SelectPalette(hdc,hDibPal,FALSE);
  int ncolsmapped=RealizePalette(hdc);
  ReleaseDC(hwnd,hdc);
  if (ncolsmapped>0)
  { InvalidateRect(hwnd,NULL,TRUE); return TRUE; }
  else
  { return FALSE; }
}



TBitsSurface::TBitsSurface()
{ data.lpv=NULL;
  internaldata.lpv=NULL;
  DibPresent=FALSE;
  Dibhdc=NULL;
  DibhOldBitmap=NULL;
  bits=NULL;
  islocked=FALSE;
  dds=NULL;
  dirtypalette=FALSE;
}

TBitsSurface::~TBitsSurface()
{ if (DibPresent)
  { if (DibhOldBitmap!=NULL)
    { HGDIOBJ hbm=SelectObject(Dibhdc,DibhOldBitmap);
      DeleteObject(hbm);
    }
    DeleteDC(Dibhdc);
    DibPresent=FALSE;
  }
  if (dds!=NULL) dds->Release(); dds=NULL;
}

HRESULT TBitsSurface::InvertBits()
{ BOOL dolock=FALSE;
  if (!islocked) dolock=TRUE;
  HRESULT err;
  if (dolock)
  { err=Lock();
    if (err!=BITS_OK) return err;
  }
  BYTE *line;
  line=new BYTE[bits->numbytes];
  for (int y=0; y<bits->cy; y++)
  { LPBYTE srcaddr,destaddr;
    srcaddr=data.lpb+y*bits->cx;
    destaddr=line+(bits->cy-1-y)*bits->cx;
    for (int x=0; x<bits->cx; x++) *destaddr++=*srcaddr++;
  }
  LPBYTE srcaddr=line, destaddr=data.lpb;
  for (int i=0; i<bits->numbytes; i++) *destaddr++=*srcaddr++;
  delete[] line;
  if (dolock) Unlock();
  return BITS_OK;
}

HRESULT TBitsSurface::CopyBitsFrom(TBitsSurface &src,BOOL respectorientation)
{ BOOL dodestlock=FALSE;
  if (!islocked) dodestlock=TRUE;
  BOOL dosrclock=FALSE;
  if (!src.islocked) dosrclock=TRUE;
  HRESULT err;
  if (dosrclock)
  { err=src.Lock();
    if (err!=BITS_OK) return err;
  }
  if (dodestlock)
  { err=Lock();
    if (err!=BITS_OK) return err;
  }
  // We have to figure out how to preserve orientation
  int srcorient=src.bits->orientation;
  int destorient=bits->orientation;
  if (!respectorientation) destorient=srcorient;
  unsigned int maxcx=bits->cx; if (src.bits->cx<maxcx) maxcx=src.bits->cx;
  unsigned int maxcy=bits->cy; if (src.bits->cy<maxcy) maxcy=src.bits->cy;
  // This y counts down from the top
  for (int y=0; y<maxcy; y++)
  { LPBYTE srcaddr,destaddr;
    if (srcorient==1) srcaddr=src.data.lpb+y*src.bits->cx; else srcaddr=src.data.lpb+(src.bits->cy-1-y)*src.bits->cx;
    if (destorient==1)destaddr=data.lpb+y*bits->cx;    else destaddr=data.lpb   +(bits->cy-1-y)*bits->cx;
    for (int x=0; x<maxcx; x++) *destaddr++=*srcaddr++;
    for (x=maxcx; x<bits->cx; x++) *destaddr++=0;
  }
  for (y=maxcy; y<bits->cy; y++)
  { LPBYTE destaddr;
    if (destorient==1)destaddr=data.lpb+y*bits->cx;    else destaddr=data.lpb   +(bits->cy-1-y)*bits->cx;
    for (int x=0; x<bits->cx; x++) *destaddr++=0;
  }
  if (dodestlock) Unlock();
  if (dosrclock) src.Unlock();
  return BITS_OK;
}


HRESULT TBitsSurface::Lock(BOOL toread,BOOL towrite)
{ int thissurface=999;
  for (int i=0; i<bits->NumSurfaces; i++)
  { if (&(bits->surface[i])==this) thissurface=i;
  }
  char errstr[100];
  wsprintf(errstr,"------- Surface %i:Lock",thissurface);
  BitsDebug(errstr);
  if (islocked)
  { BitsDebug("was already locked. returning");
    return BITSERR_LOCKED;
  }
  if (bits->isdib)
  { data.lpv=internaldata.lpv;
    BitsDebug("Dib lock: done it.");
  }
  if (bits->isddx)
  { // Possibly must restore surface, if we were switched away from
    BitsDebug("Ddx lock: checking if is lost.");
    if (dds->IsLost()==DDERR_SURFACELOST)
    { BitsDebug("it was lost. restoring now.");
      dds->Restore();
    }
    ZeroMemory(&ddsd,sizeof(ddsd));
    ddsd.dwSize=sizeof(ddsd);
    BitsDebug("About to lock...");
    DWORD flags=0;
    if (!toread && !towrite)
    { BitsDebug("Requested neither to read nor to write. No point locking!");
      return BITSERR_INVALIDPARAMS;
    }
    if (toread && !towrite) { BitsDebug("Read-only."); flags=DDLOCK_READONLY;}
    if (!toread && towrite) { BitsDebug("Write-only.");flags=DDLOCK_WRITEONLY;}
    if (toread && towrite ) { BitsDebug("read&write.");flags=0;}
    //flags |= DDLOCK_WAIT;
    HRESULT err=dds->Lock(NULL,&ddsd,flags,NULL);
    if (err!=DD_OK)
    { switch (err)
      { case DDERR_INVALIDOBJECT: BitsDebug("Invalid object."); break;
        case DDERR_OUTOFMEMORY:   BitsDebug("Out of memory."); break;
        case DDERR_SURFACELOST:   BitsDebug("Surface list."); break;
        case DDERR_INVALIDPARAMS: BitsDebug("Invalid params."); break;
        case DDERR_SURFACEBUSY:   BitsDebug("Surface busy."); break;
        case DDERR_WASSTILLDRAWING: BitsDebug("Was still drawing."); break;
        default: BitsDebug("Unknown DD error on lock.");
      }
      return BITSERR_DDERROR;
    }
    data.lpv=ddsd.lpSurface;
  }
  islocked=TRUE;
  wsprintf(errstr,"Lock successfull! at %lx",data.lpv);
  BitsDebug(errstr);
  return BITS_OK;
}



HRESULT TBitsSurface::Unlock()
{ int thissurface=999;
  for (int i=0; i<bits->NumSurfaces; i++)
  { if (&(bits->surface[i])==this) thissurface=i;
  }
  char errstr[100];
  wsprintf(errstr,"------- Surface %i:Unlock",thissurface);
  BitsDebug(errstr);
  if (!islocked)
  { BitsDebug("Surface was not locked! aborting.");
    return BITSERR_NOTLOCKED;
  }
  islocked=FALSE;
  data.lpv=NULL;
  if (bits->isddx)
  { BitsDebug("Unlocking DDX surface");
    dds->Unlock(ddsd.lpSurface);
  }
  return BITS_OK;
}


TDibPainter::TDibPainter()
{ hwnd=NULL; cx=10; cy=10; dirty=FALSE; paint=FALSE; promise=DEFAULT_PROMISE;}

void TDibPainter::Setup(HWND a_hwnd,int a_cx,int a_cy)
{ paint=FALSE; dirty=FALSE; hwnd=a_hwnd; cx=a_cx; cy=a_cy;
  switch (promise)
  { case 1: DontPaint(); break;
    case 2: StretchToWindow(); break;
    case 3: StretchToRect(promrc); break;
    case 4: CenterInWindow(); break;
    case 5: CenterAtPoint(prompt); break;
    case 6: PositionAtPoint(prompt); break;
    case 7: PositionInRect(promrc); break;
    case 8: ProportionStretchToWindow(); break;
    default: {prompt.x=0; prompt.y=0; PositionAtPoint(prompt);}
  }
}

void TDibPainter::DontPaint()
{ paint=FALSE; dirty=TRUE; promise=0; }

void TDibPainter::StretchToWindow()
{ if (hwnd==NULL) {promise=2; return;}
  GetClientRect(hwnd,&rc); stretch=TRUE;paint=TRUE;dirty=TRUE;
}
void TDibPainter::ProportionStretchToWindow()
{ if (hwnd==NULL) {promise=8; return;}
  GetClientRect(hwnd,&rc);
  int wcx=rc.right-rc.left;
  int wcy=rc.bottom-rc.top;
  int wwidth=wcx;
  int wheight=wcy;
  _fpreset();
  double fwidth,fheight;
  if (wcx==0) fwidth=1.0; else fwidth=((double)cx)/((double)wcx);
  if (wcy==0) fheight=1.0; else fheight=((double)cy)/((double)wcy);
  double f;
  if (fheight<fwidth)
  { f=fwidth; if (f==0.0) f=1.0;
    wcy=((double)cy)/f;
  }
  else
  { f=fheight;
    wcx=((double)cx)/f;
  }
  rc.left=(wwidth-wcx)/2;
  rc.right=(wwidth+wcx)/2;
  rc.top=(wheight-wcy)/2;
  rc.bottom=(wheight+wcy)/2;
  stretch=TRUE; paint=TRUE; dirty=TRUE;
}
void TDibPainter::StretchToRect(RECT &a_rc)
{ rc=a_rc; stretch=TRUE;paint=TRUE;dirty=TRUE; promise=0;
}
void TDibPainter::StretchToRect(int left,int top,int right,int bottom)
{ rc.left=left; rc.top=top; rc.right=right; rc.bottom=bottom;
  stretch=TRUE; paint=TRUE; dirty=TRUE; promise=0;
}
void TDibPainter::CenterInWindow()
{ if (hwnd==NULL) {promise=4; return;}
  RECT wrc;
  GetClientRect(hwnd,&wrc);
  rc.left=   (wrc.right+wrc.left-cx)/2;
  rc.right=  (wrc.right+wrc.left+cx)/2;
  rc.top=    (wrc.top+wrc.bottom-cy)/2;
  rc.bottom= (wrc.top+wrc.bottom+cy)/2;
  stretch=FALSE;paint=TRUE;dirty=TRUE;
}
void TDibPainter::CenterAtPoint(POINT &pt)
{ rc.left=pt.x-(cx/2);
  rc.right=pt.x+(cx/2);
  rc.top=pt.y-(cy/2);
  rc.bottom=pt.y+(cy/2);
  stretch=FALSE;paint=TRUE;dirty=TRUE; promise=0;
}
void TDibPainter::CenterAtPoint(int x,int y)
{ POINT apt; apt.x=x; apt.y=y; CenterAtPoint(apt); }
void TDibPainter::PositionAtPoint(POINT &pt)
{ rc.left=pt.x; rc.top=pt.y;
  rc.right=pt.x+cx; rc.bottom=pt.y+cy;
  stretch=FALSE;paint=TRUE;dirty=TRUE; promise=0;
}
void TDibPainter::PositionAtPoint(int x,int y)
{ POINT apt; apt.x=x; apt.y=y; PositionAtPoint(apt); }
void TDibPainter::PositionInRect(RECT &a_rc)
{ rc=a_rc; stretch=FALSE;paint=TRUE;dirty=TRUE;
  if (rc.right-rc.left>cx) rc.right=rc.left+cx;
  if (rc.bottom-rc.top>cy) rc.bottom=rc.top+cy;
  promise=0;
}
void TDibPainter::PositionInRect(int left,int top,int right,int bottom)
{ RECT arc; arc.left=left; arc.top=top; arc.right=right;
  arc.bottom=bottom;
  PositionInRect(arc);
}

void TDibPainter::Paint(HDC hdc,PAINTSTRUCT &,HDC src,BOOL flip)
{ if (stretch)
  { if (flip) StretchBlt(hdc,rc.left,rc.bottom-1,rc.right-rc.left,rc.top-rc.bottom,src,0,0,cx,cy,SRCCOPY);
    else StretchBlt(hdc,rc.left,rc.top,rc.right-rc.left,rc.bottom-rc.top,src,0,0,cx,cy,SRCCOPY);
  }
  else
  { if (flip) BitBlt(hdc,rc.left,rc.bottom-1,rc.right-rc.left,rc.top-rc.bottom,src,0,cy,SRCCOPY);
    else BitBlt(hdc,rc.left,rc.top,rc.right-rc.left,rc.bottom-rc.top,src,0,0,SRCCOPY);
  }
}

void BitsDebug(const char *c)
{
  if (!trace) return;
  char d[300];
  strcpy(d,c);
  strcat(d,"      ______     \n");
  OutputDebugString(d);
  strcpy(d,c);
  strcat(d,"\n");
  FILE *f=fopen("\\ddx.log","a+t");
  fprintf(f,d);
  fflush(f);
  fclose(f);
}


