//
// Scotch screensaver
// -------------------
// (c) 1996 Lucian Wischik
//
// The code is quite convoluted and not very well documented.
// You'd be better off first looking at the 'static' example
// to get an idea of the basics of using DirectDraw in a
// screensaver.

#define WIN32_LEAN_AND_MEAN
#define STRICT
#include <windows.h>
#include "ddraw.h"
#pragma hdrstop
#include "mmsystem.h"

#include "../../lib/scrplus.h"
#include "../../lib/scrplus.rh"
#include "bits.h"
#include "resource.h"


HINSTANCE hInstance;

#define SCRCHANGE_SCOTCH 2


class TScotchData
{ public:
  TBits bits;          // for the actual animation
  TBits screenbits;    // to store the initial snapshot of the screen
  TBits suspendbits;   // to remember the bitmap, while popping up a dialog
  TBits backdropbits;  // to display as backdrop while dialog is up
  BOOL issuspended;    // whether a dialog is up
  BOOL easysuspend;    // if it was just a DIB in the first place
  BOOL ischangingmode; // SetDisplayMode() causes spurious WM_ACTIVATE, which we must ignore
  BOOL isfull;         // Are wee in full-screen mode?
  BOOL wantddx;        // and therefore, do we want to use ddx?
  BOOL isstarted;      // Has the sim yet started?
  int frame;           // current frame number
  int cx,cy;           // screen sizes to use
  int period;          // timer delay between cycles
  BOOL ignorechanges;  // Do we ignore dialog changes at the moment?

  int Screen,markScreen; // Dialog/registry selection.

  TScotchData(BOOL);
  unsigned int Timer(HWND);
  void Change(TScotchData *);
  void Start(HWND);
  void Stop(HWND);
  void Suspend(HWND);
  void Resume(HWND);

  void ReadRegistry();
  void WriteRegistry();
  void NotifyChanges(HWND hwnd,WPARAM code=SCRCHANGE_SCOTCH);
  void Mark();
  BOOL Reset();
  void SetControls(HWND);
  BOOL GetControls(HWND);
};


TScotchData::TScotchData(BOOL a_isfull)
{ isfull=a_isfull;
  wantddx=FALSE;
  isstarted=FALSE;
  ischangingmode=FALSE; issuspended=FALSE;
  easysuspend=FALSE;
  Screen=0; markScreen=-1;
  frame=0;
  period=1;
  cx=0; cy=0;
}


void TScotchData::ReadRegistry()
{ LONG res; HKEY skey; DWORD val,valsize,valtype;
  res=RegOpenKeyEx(HKEY_CURRENT_USER,"Software\\Lu\\Scotch",0,KEY_ALL_ACCESS,&skey);
  if (res!=ERROR_SUCCESS) return;
  // 1. Screen
  valsize=sizeof(val); res=RegQueryValueEx(skey,"Screen",0,&valtype,(LPBYTE)&val,&valsize);
  if (res==ERROR_SUCCESS) Screen=(BYTE)val;
  if (Screen<0) Screen=0; if (Screen>5) Screen=5;
  RegCloseKey(skey);
}


void TScotchData::WriteRegistry()
{ LONG res; HKEY skey; DWORD val; DWORD disp;
  res=RegCreateKeyEx(HKEY_CURRENT_USER,"Software\\Lu\\Scotch",0,NULL,
                     REG_OPTION_NON_VOLATILE,KEY_ALL_ACCESS,NULL,&skey,&disp);
  if (res!=ERROR_SUCCESS) return;
  // 1. Screen
  val=Screen;
  RegSetValueEx(skey,"Screen",0,REG_DWORD,(CONST BYTE*)&val,sizeof(val));
  RegCloseKey(skey);
}

void TScotchData::NotifyChanges(HWND hsrcwnd,WPARAM defcode)
{ WPARAM code=defcode | SCRCHANGE_FROMAPPLY   | SCRCHANGE_TOPREVIEW;
  TScrChangeRouter cr;
  cr.hsrcwnd=hsrcwnd; cr.lParam=(LPARAM)this; cr.code=code;
  HWND hdlg=GetParent(hsrcwnd);
  SendMessage(hdlg,SCRM_CHANGE,0,(LPARAM)&cr);
}

void TScotchData::Mark()
{ markScreen=Screen;
}

BOOL TScotchData::Reset()
{ BOOL ischanged=FALSE;
  if (markScreen!=Screen) ischanged=TRUE; Screen=markScreen;
  return ischanged;
}

BOOL TScotchData::GetControls(HWND hwnd)
{ if (ignorechanges) return FALSE;
  BOOL ischanged=FALSE;
  LRESULT res=SendDlgItemMessage(hwnd,ID_SCREEN,CB_GETCURSEL,0,0);
  if (res!=Screen) ischanged=TRUE; Screen=res;
  if (Screen<0) Screen=0;
  if (Screen>5) Screen=5;
  return ischanged;
}

void TScotchData::SetControls(HWND hwnd)
{ ignorechanges=TRUE;
  SendDlgItemMessage(hwnd,ID_SCREEN,CB_ADDSTRING,0,(LPARAM)"Normal screen");
  SendDlgItemMessage(hwnd,ID_SCREEN,CB_ADDSTRING,0,(LPARAM)"Fast 320x200");
  SendDlgItemMessage(hwnd,ID_SCREEN,CB_ADDSTRING,0,(LPARAM)"Fast 320x240");
  SendDlgItemMessage(hwnd,ID_SCREEN,CB_ADDSTRING,0,(LPARAM)"Fast 640x480");
  SendDlgItemMessage(hwnd,ID_SCREEN,CB_ADDSTRING,0,(LPARAM)"Fast 800x600");
  SendDlgItemMessage(hwnd,ID_SCREEN,CB_ADDSTRING,0,(LPARAM)"Fast 1024x768");
  SendDlgItemMessage(hwnd,ID_SCREEN,CB_SETCURSEL,Screen,0);
  ScrCreatePreview(GetDlgItem(hwnd,ID_MONITOR));
  ignorechanges=FALSE;
}





void TScotchData::Stop(HWND)
{ BitsDebug("Stop...");
  bits.Reset();
  screenbits.Reset();
  BitsDebug("Stopped.");
}

void TScotchData::Start(HWND hwnd)
{ BitsDebug("Starting...");
  ischangingmode=TRUE;
  // First, set up a palette. This one uses 128 entries as a 4*4*8
  // colour cube, and the next 128 entries as two 64-grey-scales.
  for (int i=0; i<256; i++)
  { BYTE r,g,b;
    if (i<128)
    { b=(BYTE)((i&3)*255/3);
      r=(BYTE)(((i>>2)&3)*255/3);
      g=(BYTE)(((i>>4)&7)*255/7);
      screenbits.LogPal.Entries[i].peRed=r;
      screenbits.LogPal.Entries[i].peGreen=g;
      screenbits.LogPal.Entries[i].peBlue=b;
    }
    else
    { if (i<192) b=(BYTE)((i-128)*2); else b=(BYTE)((i-192)*2);
      r=b; g=b;
      screenbits.LogPal.Entries[i].peRed=255;
      screenbits.LogPal.Entries[i].peGreen=255;
      screenbits.LogPal.Entries[i].peBlue=255;
    }
    bits.LogPal.Entries[i].peRed=r;
    bits.LogPal.Entries[i].peGreen=g;
    bits.LogPal.Entries[i].peBlue=b;
  }
  screenbits.ChangePalette();
  bits.ChangePalette();
  // So we set the palette for both of them. That means that,
  // when we do CreateSingleDibFromScreen on screenbits,
  // it will automatically be colour-matched to the palette we
  // just created

  switch (Screen)
  { case 1: { cx=320; cy=200; wantddx=TRUE; } break;
    case 2: { cx=320; cy=240; wantddx=TRUE; } break;
    case 3: { cx=640; cy=480; wantddx=TRUE; } break;
    case 4: { cx=800; cy=600; wantddx=TRUE; } break;
    case 5: { cx=1024;cy=768; wantddx=TRUE; } break;
    default:
    { cx=GetSystemMetrics(SM_CXSCREEN); cy=GetSystemMetrics(SM_CYSCREEN);
      wantddx=FALSE;
    }
  }
  if (!isfull) wantddx=FALSE;
  screenbits.CreateDibFromScreen(NULL,cx,cy);
  if (wantddx) screenbits.surface[0].InvertBits();

  // and start proper. We set DibPainter and EnableSubclass
  // just in case: if it's DDX then they won't do anything,
  // but if it's DIB then they will.
  if (isfull) bits.DibPainter.PositionAtPoint(0,0);
  else bits.DibPainter.StretchToWindow();
  bits.EnableSubclass(TRUE,NULL);
  BitsDebug("Creating new TBits image...");
  int numsurfaces; if (wantddx) numsurfaces=2; else numsurfaces=1;
  // That line was a workaround a bug in the Bits interface, whereby
  // multiple flippable DDX surfaces don't work.
  HRESULT err=bits.Create(hwnd,wantddx,numsurfaces,cx,cy,8,TRUE);
  if (err!=BITS_OK) {BitsDebug("Error creating image."); ischangingmode=FALSE; return;}

  // Now we're going to tidy up the screen DIB: we want to reserve
  // the top 128 colours for our own stuff. So, we'll map any pixel
  // that happened to use a high colour, back down.
  TBitsSurface *ss=screenbits.LockSurface(0);
  if (ss!=NULL)
  { LPBYTE addr=ss->data.lpb;
    for (int i=0; i<screenbits.numbytes; i++)
    { if (*addr>128) *addr=127; addr++; }
  }
  screenbits.UnlockSurface(ss);
  // And copy this screen onto our actual image.
  // the 'frametochange' stuff is a workaround a bug in the Bits module
  int frametochange; if (wantddx) frametochange=1; else frametochange=0;
  bits.surface[frametochange].CopyBitsFrom(screenbits.surface[0],FALSE);
  bits.Flip();
  ischangingmode=FALSE;
  isstarted=TRUE;
}


void TScotchData::Change(TScotchData *sd)
{ Screen=sd->Screen;
}

unsigned int TScotchData::Timer(HWND)
{ if (issuspended) return period;
  BitsDebug("Timer...");
  // 'frametochange' is a workaround a bug in the Bits module
  int frametochange; if (wantddx) frametochange=1; else frametochange=0;
  TBitsSurface *rs=bits.LockSurface(frametochange);
  TBitsSurface *ss=screenbits.LockSurface(0);
  if (rs==NULL || ss==NULL)
  { bits.UnlockSurface(rs); screenbits.UnlockSurface(ss); return period; }

  // First we might seed a new whatdyamacallit
  frame=frame+1;
  if ((frame%10)==0)
  { int r=rand()*rand();
    LPBYTE addr=rs->data.lpb+(r%bits.numbytes);
    *addr=254;
  }

  // This next bit is the actual animation. It's too complicated
  // for me to be bothered to explain, and it's not really that
  // important
  LPBYTE addr=rs->data.lpb+bits.cx+1;
  LPBYTE saddr=ss->data.lpb+screenbits.cx+1;
  LPBYTE last=rs->data.lpb+bits.numbytes-bits.cx-1;
  BYTE thisturn=0x80, lastturn=0xC0;
  if (frame & 1) {thisturn=0xC0; lastturn=0x80;}
  while (addr<=last)
  { BYTE val=*addr;
    if ((val&0xC0)==thisturn)
    { if (val==thisturn) *addr=*saddr; else *addr=(BYTE)(val-1);
      LPBYTE na; BYTE nv;
      if ((val&0x3F)==0x3E)
      { nv=*saddr;
        int greyapprox=64*((nv&3)+((nv>>2)&3))/9;
        greyapprox=greyapprox + 64*((nv>>4)&7)/21;
        greyapprox=greyapprox%64;
        if (greyapprox>=0x3D) greyapprox=0x3D;
        if (greyapprox<=3) greyapprox=3;
        nv=(BYTE)(greyapprox+lastturn);
        *addr=nv;
        na=addr-bits.cx;
        nv=*na; if ((nv&0x80)==0) *na=(BYTE)(lastturn | 0x3E);
        na=addr+1;
        nv=*na; if ((nv&0x80)==0) *na=(BYTE)(lastturn | 0x3E);
        na=addr+bits.cx;
        nv=*na; if ((nv&0x80)==0) *na=(BYTE)(lastturn | 0x3E);
        na=addr-1;
        nv=*na; if ((nv&0x80)==0) *na=(BYTE)(lastturn | 0x3E);
      }
    }
    addr++; saddr++;
  }

  bits.UnlockSurface(rs);
  screenbits.UnlockSurface(ss);
  bits.Flip();

  BitsDebug("Done period.");
  return period;
}

void TScotchData::Suspend(HWND hwnd)
{ if (issuspended) return;
  if (!isfull) return;
  if (!wantddx) return;
  issuspended=TRUE;
  easysuspend=FALSE;
  suspendbits.Reset();
  int copyfrom=0; if (wantddx) copyfrom=1;
  suspendbits.CreateDibCopy(hwnd,bits.surface[copyfrom]);
  ischangingmode=TRUE;
  bits.Reset();
  UINT oldval;
  SystemParametersInfo(SPI_SCREENSAVERRUNNING,1,&oldval,0);
  // because leaving DirectDraw would have suddenly enabled
  // ctrl-alt-delete and task switching. We want to disable it again.
  int scx=GetSystemMetrics(SM_CXSCREEN);
  int scy=GetSystemMetrics(SM_CYSCREEN);
  MoveWindow(hwnd,0,0,scx,scy,FALSE);
  backdropbits.Reset();
  backdropbits.EnableFullSubclass(FALSE);
  backdropbits.CreateDibCopy(hwnd,suspendbits.surface[0]);
  ischangingmode=FALSE;
}

void TScotchData::Resume(HWND hwnd)
{ if (!issuspended) return;
  if (easysuspend)
  { issuspended=FALSE;
    easysuspend=FALSE;
    return;
  }
  bits.Reset();
  if (isfull) bits.DibPainter.PositionAtPoint(0,0);
  else bits.DibPainter.StretchToWindow();
  bits.EnableSubclass(TRUE,NULL);
  bits.CopyPaletteFrom(suspendbits);
  int numsurfaces; if (wantddx) numsurfaces=2; else numsurfaces=1;
  ischangingmode=TRUE;
  backdropbits.Reset();
  HRESULT err=bits.Create(hwnd,wantddx,numsurfaces,cx,cy,8,TRUE);
  if (err!=BITS_OK) {ischangingmode=FALSE; return;}
  int frametochange; if (wantddx) frametochange=1; else frametochange=0;
  bits.surface[frametochange].CopyBitsFrom(suspendbits.surface[0]);
  suspendbits.Reset();
  bits.Flip();
  ischangingmode=FALSE;
  issuspended=FALSE;
}


LRESULT CALLBACK MyScreenSaverProc(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam)
{ TScotchData *sd=(TScotchData *)GetWindowLong(hwnd,GWL_USERDATA);
  if (sd==NULL && msg!=WM_CREATE) return DefScreenSaverProc(hwnd,msg,wParam,lParam);
  switch (msg)
  { case WM_CREATE:
    { BOOL isfull=FALSE; if (GetParent(hwnd)==NULL) isfull=TRUE;
      TScotchData *sd=new TScotchData(isfull);
      sd->ReadRegistry();
      SetWindowLong(hwnd,GWL_USERDATA,(LONG)sd);
    } break;
    case WM_ACTIVATE: case WM_NCACTIVATE:
    { if (sd->ischangingmode) return DefWindowProc(hwnd,msg,wParam,lParam);
    } break;
    case SCRM_TIMER: {if (!sd->isstarted) sd->Start(hwnd); return sd->Timer(hwnd); }
    case SCRM_CHANGE:
    { if ((wParam&SCRCHANGE_GETCODE)==SCRCHANGE_SCOTCH)
      { sd->Stop(hwnd);
        sd->ReadRegistry();
        sd->Start(hwnd);
      }
    } break;
    case SCRM_DIALOG:
    { if ((BOOL)wParam) sd->Suspend(hwnd); else sd->Resume(hwnd); } break;
    case WM_LBUTTONDOWN: sd->Stop(hwnd); break;
    case WM_DESTROY:
	 { sd->Stop(hwnd);
      SetWindowLong(hwnd,GWL_USERDATA,0);
      delete sd;
    } break;
  }
  return DefScreenSaverProc(hwnd,msg,wParam,lParam);
}



BOOL WINAPI MyScreenSaverConfigureDialog(HWND hwnd, UINT msg,WPARAM wParam,LPARAM lParam)
{ TScotchData *sd=(TScotchData *)GetWindowLong(hwnd,DWL_USER);
  switch (msg)
  { case WM_INITDIALOG:
    { TScotchData *sd=new TScotchData(FALSE);
      sd->ReadRegistry();
      sd->Mark();
      SetWindowLong(hwnd,DWL_USER,(LONG)sd);
      sd->SetControls(hwnd);
      return TRUE;
    }
    case WM_DESTROY:
    { delete sd;
      SetWindowLong(hwnd,DWL_USER,0);
      return TRUE;
    }
    case WM_COMMAND:
    { int wID=LOWORD(wParam);
      if (wID==ID_SCREEN)
      { BOOL ischanged=sd->GetControls(hwnd);
        if (ischanged)
        { sd->WriteRegistry();
          sd->NotifyChanges(GetParent(hwnd));
        }
      }
    } break;
    case WM_NOTIFY:
    { LPNMHDR nmh=(LPNMHDR) lParam;
      UINT code=nmh->code;
      switch (code) {
        case (PSN_APPLY):
        { BOOL ischanged=sd->GetControls(hwnd);
          if (ischanged)
          { sd->Mark();
            sd->WriteRegistry();
          }
          return TRUE;
        }
        case PSN_RESET:
        { BOOL ischanged=sd->Reset();
          if (ischanged)
          { sd->WriteRegistry();
          }
          return TRUE;
        }
        default: return FALSE;
      }
    }
  }
  return FALSE;
}



int WINAPI WinMain(HINSTANCE h,HINSTANCE,LPSTR,int)
{ hInstance=h;
  BitsInit(hInstance);
  TScrPrefs *p=ScrCreatePrefs();
  p->AllowActiveConfig=FALSE;
  p->AllowMute=FALSE;
  p->IsTrace=FALSE;
  p->IsDebug=FALSE;
  ScrSetPrefs(p);

  int res=ScrExecute(hInstance,MyScreenSaverProc,MyScreenSaverConfigureDialog);

  BitsClose();
  return res;
}



