//
// ScrPlus v1.51
// (c) 1996 Lucian Wischik. All rights reserved
//

#ifndef __scrplus_h
#define __scrplus_h


// The following few lines are so that it will still compile even if
// you don't have the latest Win32 headers.
#if (WINVER >= 0x0400)
#ifndef _PRSHT_H_
#include <prsht.h>
#endif
typedef PROPSHEETHEADER CONFIGSHEETHEADER;
#else
typedef struct _CONFIGSHEETHEADER { DWORD dwSize; } CONFIGSHEETHEADER;
#endif

#include <pshpack1.h> // to turn on 1-byte packing of structures
//typedef struct {int a;} PROPSHEETHEADER;


#define ScrCompany               "Lucian Wischik Software"
#define ScrCopyright             "(c) 1996 Lucian Wischik"
#define ScrProduct               "ScrPlus Screensavers"
#define ScrProductVersion        "1.51"
#define ScrProductVersionNumber  1,5,1,0


#define ICON_APP 100
// If you want to use the property sheet, you must put your icon here
// If you don't, then none will be used.
#define ICON_SCRAPP 199
// ScrApp icon is just the official scrapp icon
#define STR_SETTINGSDLGTITLE 3501
// If you use a property dialog and wish to supply a different title for
// it, put it here. Otherwise it will just use "Screen Saver Properties"
#define DLG_ABOUTSCRPLUS	3602
// This is here just in case you happen to want to put up the ScrPlus
// about dialog anywhere in your program.
#define ID_VERSION 10
// If you're using just ScrExecute with the standard dialog procedure,
// then any control with this ID will have the version appended to it.


#define STR_SCRHELPFILE 3560
// This string has the name of the help file for
// additional context help. If you don't supply one yourself, a default
// is taken. This help file must include the topic SCRHELP_ACTIVECONFIG.
// Both of these are only needed if you actually allow activeconfig.
#define SCRHELP_CORNERS       3100
#define SCRHELP_THRESHOLD     3101
#define SCRHELP_PASSWORDDELAY 3102
#define SCRHELP_COPYRIGHT     3103
#define SCRHELP_PREVIEW       3104
#define SCRHELP_MUTE          3105
#define SCRHELP_ACTIVECONFIG  3106


#define action_default 0
#define action_now     1
#define action_never   2
#define corner_nw 0
#define corner_ne 1
#define corner_se 2
#define corner_sw 3
typedef BYTE TCorners[4];



// There are three ways of executing a screensaver.
// 1. 'Normal': ExecuteScreensaverNormal(...)
//     this corresponds exactly to the standard Scrnsave.lib with Windows
// 2. 'Extended': ExecuteScreensaverEx(...)
//     this gives you access to all the Plus! functionality, via functions
// 3. 'Window': ExecuteScreensaverWin(HWND)
//     the same functionality as 'extended', but everything is routed via
//     your invisible main application window. This is so that the DLL
//     is easier to use under different development environments.

// Ancilliary definitions for the various bits are given first, followed
// by the definitions of the actual ExecuteX procedures.

extern "C" int WINAPI ScrGetVersionString(char *destbuf,int maxlen);


typedef struct tagScrPrefs{
  UINT cbSize;
  // That's the size of the structure
  BOOL AllowActiveConfig;
  BOOL AllowMute;
  // These two above say whether the 'activeconfig' and 'mute' checkboxes on
  // the general page should be displayed. Don't display them if they're not
  // relevant. By default, ActiveConfig is false and AllowMute is true.
  BOOL IsTrace;
  BOOL IsDebug;
  // IsTrace traces things to ODS. IsDebug makes it appear in a smaller window
  // at the top left that is more robust (ie. will not close as easily).
  // Both default false.
  BOOL DisableAbout;
  // Whether or not 'about' is added to the system menu. Default false
  BOOL DisablePreviewBroadcast;
  // This is an internal thing. It controls whether or not SCRM_CHANGE messages
  // that get sent to the parent of the dialog, should automatically seek out
  // every preview control in the config. dialog and pass the SCRM_CHANGE message
  // on to them as well. False by default.
  BOOL DisableThreadTimer;
  // This is an internal thing. It allows for SCRM_TIMER messages to get sent.
  // Whether or not you do anything with them is up to you. You need disable
  // them only if you are having trouble with multi-threaded stuff. Default false
  CHAR DynamicCode[256];
  CHAR StaticCode[256];
  // Set either of these fields to your private codeword, if you are
  // a registered user and want to get rid of the nag screens.
  BOOL AlwaysRequirePassword;         // default false
  BOOL NeverRequirePassword;          // default false
  BOOL CheckPasswordForActiveConfig;  // default true
  // require-password is to do with the whether the time-delay for passwords
  // is enabled, and controls the appearance of the password section in the
  // general-properties tab, and for the 'chance password' button.
  // CheckPFAConfig is to sto pcasual passers-by from altering your screensaver.
  // Both are available only in the registered version
} TScrPrefs, *LPTScrPrefs;
extern "C" LPTScrPrefs WINAPI ScrCreatePrefs();
extern "C" void WINAPI ScrSetPrefs(TScrPrefs *);



// Screensaver window
// --------------------
extern "C" LRESULT WINAPI  DefScreenSaverProc (HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam);
#ifndef WM_APP
#define SCRM_FIRST 0x800
#else
#define SCRM_FIRST WM_APP
#endif
#define SCRM_VERIFYPW      SCRM_FIRST
#define SCRM_ACTIVECONFIG  SCRM_FIRST+1
#define SCRM_CHANGE        SCRM_FIRST+2
#define SCRM_TIMER         SCRM_FIRST+3
#define SCRM_DIALOG        SCRM_FIRST+4     // wParam is true for starting, false for ending
#define SCRM_LAST          SCRM_FIRST+10

extern "C" void WINAPI ScrStartTimer(HWND);
extern "C" void WINAPI ScrStopTimer(HWND);


#define SCRCREATE_FULL          0
#define SCRCREATE_MAINPREVIEW   1
#define SCRCREATE_CONFIGPREVIEW 2
typedef struct tagScrCreateStruct { LPVOID reserved; UINT style; } TScrCreateStruct;
// A pointer to one of these structures, with a SCRCREATE_ constant in the
// 'style', is passed to WM_CREATE in the lParam of the creation data.
// The default behaviour frees the memory.

#define SCRCHANGE_FROMEITHER  0x0000
#define SCRCHANGE_FROMCONTROL 0x1000
#define SCRCHANGE_FROMAPPLY   0x2000
#define SCRCHANGE_FROMBOTH    0x3000
#define SCRCHANGE_TOEITHER    0x0000
#define SCRCHANGE_TOPREVIEW   0x4000
#define SCRCHANGE_TOFULL      0x8000
#define SCRCHANGE_TOBOTH      0xC000
#define SCRCHANGE_GETCODE     0x0FFF
#define SCRCHANGE_GETFROM     0x3000
#define SCRCHANGE_GETTO       0xC000
// SCRM_CHANGE takes a change code in wParam and additional information in lParam
// When you're sending a SCRM_CHANGE from your dialog, you send it to the
// parent of your dialog, which will do the routing according to how you've
// set the TO bits. The FROM bits are informational, and are currently used
// by the general page so you can decide how you want to behave when, say,
// the user clicks on 'mute': immediate effect, or only on OK? Up to you.
#define SCRCHANGE_GENERAL 1
typedef struct tagScrGeneralSettings {
  UINT cbSize; TCorners corner; BOOL mute; unsigned int mousethreshold;
  unsigned int passworddelay; BOOL activeconfig;
  } TScrGeneralSettings;
// The only predefined code is SCRCHANGE_GENERAL. It gets routed to both
// previews and full-screen whenever something is changed. lParam points
// to a TScrChangeGeneralInfo just as a convenience to you.
// All other codes are application specific.

extern "C" void WINAPI ScrGetGeneralSettings(TScrGeneralSettings *);

typedef struct tagScrChangeRouter { HWND hsrcwnd; WPARAM code; LPARAM lParam; } TScrChangeRouter;
// If you want the parent of your dialog box to do routing of change messages,
// you must send it an SCRM_CHANGE. wparam will be ignored; lparam must be
// a pointer to one of these.

// Configuration dialog
// ----------------------
#define MonitorClassName       "ScrplusMonitorClass"
#define MS_NOTHING       0x0000
#define MS_PREVIEW       0x0001
#define MS_CORNERS       0x0002
#define MS_CTOPLEFTDEF   0x0000
#define MS_CTOPLEFTNOW   0x0010
#define MS_CTOPLEFTNEV   0x0020
#define MS_CTOPRIGHTDEF  0x0000
#define MS_CTOPRIGHTNOW  0x0040
#define MS_CTOPRIGHTNEV  0x0080
#define MS_CBOTRIGHTDEF  0x0000
#define MS_CBOTRIGHTNOW  0x0100
#define MS_CBOTRIGHTNEV  0x0200
#define MS_CBOTLEFTDEF   0x0000
#define MS_CBOTLEFTNOW   0x0400
#define MS_CBOTLEFTNEV   0x0800
#define MS_MONITORMASK   0x0003
#define MS_CTOPLEFTMASK  0x0030
#define MS_CTOPRIGHTMASK 0x00C0
#define MS_CBOTRIGHTMASK 0x0300
#define MS_CBOTLEFTMASK  0x0C00


extern "C" HWND WINAPI ScrCreatePreview(HWND hparwnd);
// Use this to create a preview window that is a child of hparwnd.

typedef struct tagDialogInfo { BOOL IsAnActiveDialog; } TScrDialogInfo;
// When your dialog is running, then do GetWindowLong(GWL_USERDATA),
// you'll get a pointer to this structure

// Normal anciliary definitions:
// -------------------------------
#define DLG_SCRNSAVECONFIGURE   2003
// If you want to use your own configure stuff, your dialog must go here.
#define DLG_ABOUT 2004
// If this is present, help|about will be turned on.
typedef LRESULT (CALLBACK SCREENSAVERPROC) (HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam);
typedef BOOL (CALLBACK SCREENSAVERCONFIGUREDIALOG) (HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam);
typedef BOOL (CALLBACK REGISTERDIALOGCLASSES) (HANDLE hInst);
typedef BOOL (CALLBACK SCREENSAVERCHANGEPASSWORD) (HWND hParent);

//
extern "C" int  WINAPI ScrExecute(HINSTANCE hScreenSaverInstance,
                                            SCREENSAVERPROC *ScreenSaverProc,
                                            SCREENSAVERCONFIGUREDIALOG *ScreenSaverConfigureDialogProc=NULL,
                                            REGISTERDIALOGCLASSES *RegisterDialogClassesProc=NULL,
                                            SCREENSAVERCHANGEPASSWORD *ScreenSaverChangePasswordProc=NULL);


// Ancilliary definitions for both extended things
// -------------------------------------------------


// Extended anciliary definitions:
// ---------------------------------
// Note: if the person has old headers, then they won't have access to PROPSHEETHEADER
typedef HWND (CALLBACK LUCREATESAVER) (BOOL ispreview,CREATESTRUCT &);
typedef int (CALLBACK LUCONFIG) (HWND hParent, CONFIGSHEETHEADER &,BOOL isactive);
typedef int (CALLBACK LUABOUT) (HWND hParent);
typedef int (CALLBACK LUCHANGEPASSWORD) (HWND hParent);
#define DLGCODE_DOSCRDEFAULT  9
// The return from config, about, changepw can be IDOK or IDCANCEL.
// But if it's DLG_DOSCRDEFUALT then the default is done
//
extern "C" int WINAPI ScrExecuteEx(HINSTANCE hScreenSaverInstance,
											  LUCREATESAVER *LuCreateSaverProc,
											  LUCONFIG *LuConfigProc=NULL,
											  LUABOUT *LuAboutProc=NULL,
											  LUCHANGEPASSWORD *LuChangePasswordProc=NULL);

// Window ancilliary definitions:
// --------------------------------
#define SCRAPP_FIRST                  WM_APP+100
#define SCRAPP_LUCREATEMAINPREVIEW    WM_APP+100
#define SCRAPP_LUCREATECONFIGPREVIEW  WM_APP+101
#define SCRAPP_LUCREATEPREVIEW        WM_APP+102
#define SCRAPP_LUCREATEFULL           WM_APP+103
#define SCRAPP_LUCREATESAVER          WM_APP+104
#define SCRAPP_LUCONFIG               WM_APP+105
#define SCRAPP_LUABOUT                WM_APP+106
#define SCRAPP_LUCHANGEPASSWORD       WM_APP+107
#define SCRAPP_LAST                   WM_APP+109
extern "C" LRESULT WINAPI  DefScreenSaverAppProc (HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam);
//
extern "C" int WINAPI ScrExecuteWin(HINSTANCE hScreenSaverInstance,
													  HWND hScreenSaverAppWindow);



// And finally a function to make it easier for you to run
// the standard ScrPlus help. The second parameter will be
// etierh HELP_CONTEXT_MENU or HELP_WM_HELP.
extern "C" void WINAPI ScrHelp(HWND,DWORD helpcommand,DWORD helptopic);


// Now two calls which are only important if you statically link, if you
// wish to use the monitor class, and if you do not wish to use ScrExecute.
// ScrExecute normally initialises the monitor class, and tidies it up at
// the end: but since you don't use these, you will have to do it yourself.
extern "C" void WINAPI ScrRegister();
extern "C" void WINAPI ScrUnregister();

// The following definitions are just in case you compiled it
// on a system that didn't include the new win32 headers
#ifndef DS_3DLOOK
#define DS_3DLOOK 0x0004L
#endif
#ifndef DS_CONTEXTHELP
#define DS_CONTEXTHELP 0x2000L
#endif



#include <poppack.h>
#endif


